VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cBurnAudio"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Public Event WriteProgress(ByVal percent As Integer, ByVal TrackNum As Integer)
Public Event Finalizing()

'**************************************
'** Function: WriteAudioCD
'** Returns: Long
'**          -9: couldn't finalize disc
'**          -8: unknown error (maybe read error)
'**          -7: too much tracks
'**          -6: size of tracks to big for CD-R(W)'s capacity
'**          -5: CD-R(W) isn't empty
'**          -4: couldn't set write speed to 4x (obsolete)
'**          -3: couldn't send WMP
'**          -2: couldn't read track information
'**          -1: write error (Buffer Over-/Underrun?)
'**           0: couldn't close the current track
'**           1: finished!!
'** Usage: writes 0<X<100 wave files (44.1kHz 16bit stereo) to CD-R(W)
'**************************************
Public Function WriteAudioCD(ByVal DrvID As String, ByRef tracks() As String) As Long
    'On Error GoTo ErrH

    Dim CDSize As Long
    Dim i As Integer, j As Long
    Dim rti As t_RTI
    Dim StartLBA As Long
    Dim FF As Integer: FF = FreeFile
    Dim JOH As String: JOH = Space(44)
    Dim buffer(8 * 2352 - 1) As Byte
    Dim writes As Long, written As Long

    'not more than 99 tracks allowed
    If UBound(tracks) > 98 Then _
        WriteAudioCD = -7: Exit Function

    'get the total size of data to burn
    For i = 0 To UBound(tracks)
        CDSize = CDSize + FileLen(tracks(i))
    Next

    With CDInfo(DrvID)

        'capacity of the disc reached?
        If CDSize \ 2352 > .Capacity Then
            WriteAudioCD = -6
            Exit Function
        End If

        'empty disc?
        If .DiscStatus <> STAT_EMPTY Then
            WriteAudioCD = -5
            Exit Function
        End If

    End With

    'perform optimal power calibration
    SendOPCInformation DrvID, True, 0, 0

    'go!!
    For i = 0 To UBound(tracks)

        'set the writing speed to 4x (704 kb/s)
'        If Not SetCDSpeed(DrvID, 0, 4 * 176, True) Then _
'            WriteAudioCD = -4: Exit Function

        'set the WMP to 150 frames Pre-Gap, 2352 bytes/block, audio, TAO
        If Not WriteParams(DrvID, False, False, 150, 1, 0, 0, False) Then _
            WriteAudioCD = -3: Exit Function

        'read information about the current track
        If Not ReadTrackInformation(DrvID, 1, i + 1, VarPtr(rti), Len(rti)) Then _
            WriteAudioCD = -2: Exit Function

        'get the start LBA of the new track
        StartLBA = LShift(rti.Track_Next_Writable(0), 24) Or _
                   LShift(rti.Track_Next_Writable(1), 16) Or _
                   LShift(rti.Track_Next_Writable(2), 8) Or _
                   rti.Track_Next_Writable(3)

        'get the number of sectors to write
        writes = Fix(FileLen(tracks(i)) / (UBound(buffer) + 1))

        'open the wave file
        Open tracks(i) For Binary As #FF

            'jump over the wave header
            Get #FF, , JOH

            For j = StartLBA To StartLBA + writes

                DoEvents

'                If bCancel Then
'                    'cancel writing?
'                    bCancel = False
'                    'finalize the disc
'                    CloseCD DrvID, 2, 0
'                    WriteAudioCD = 2
'                    Exit Function
'                End If

                'read 8 sectors from the file...
                Get #FF, , buffer

                '...and write them
                If Not Write10(DrvID, j, (UBound(buffer) + 1) \ 2352, VarPtr(buffer(0)), UBound(buffer) + 1) Then
                    Close #FF
                    WriteAudioCD = -1
                    Exit Function
                End If

                'number of written sectors
                written = written + 8

                'update the progress
                RaiseEvent WriteProgress(Fix((written / (CDSize \ 2352)) * 100), i + 1)

            'next bunch of sectors
            Next j

        Close #FF

        'wait for the cd writer to burn all the data from the buffer to the disc
        RaiseEvent Finalizing
        SyncCache DrvID, j

        'close the track
        If Not CloseCD(DrvID, 1, i + 1) Then
            WriteAudioCD = 0: Exit Function
        End If

    'next track
    Next

    'finalize the disc
    If Not CloseCD(DrvID, 2, 0) Then
        WriteAudioCD = -9
        Exit Function
    End If

    'god, yeah, we made it!!!
    WriteAudioCD = 1
    Exit Function

ErrH:
    WriteAudioCD = -8
End Function
