VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cBurnISO"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Public Event WriteProgress(ByVal percent As Long)
Public Event Finalizing()
Public Event Canceling()

'**************************************
'** Function: WriteISO
'** Returns: Long
'**          -7: disc type not supported
'**          -6: disc not empty
'**          -5: disc's capacity too small
'**          -4: image not ISO9660 valid
'**          -3: couldn't set CD Speed (obsolete)
'**          -2: couldn't set WMP
'**          -1: write error
'**           0: couldn't close disc
'**           1: success
'**           2: canceled by user
'** Usage: writes a ISO9660 image to CD-R(W), if supported by drive
'**************************************
Public Function WriteISO(ByVal DrvID As String, ByVal sFile As String, Optional ByVal TestMode As Boolean, Optional ByRef bCancel As Boolean) As Long
    On Error Resume Next

    Dim FF As Integer: FF = FreeFile
    Dim iso_hdr As String
    Dim sector(16383) As Byte
    Dim sectors As Long
    Dim i As Long
    Dim cd_nfo As New cCDInfo

    'get information about the inserted disc
    Set cd_nfo = CDInfo(DrvID)

    'CD Type isn't CD-R or CD-RW?
    If Not cd_nfo.CDType = TYPE_CDR Then
        If Not cd_nfo.CDType = TYPE_CDRW Then
            WriteISO = -7
            Exit Function
        End If
    End If

    'empty?
    If Not cd_nfo.DiscStatus = STAT_EMPTY Then _
        WriteISO = -6: Exit Function

    'capacity to small for the image?
    If (cd_nfo.Capacity * 2048) < FileLen(sFile) Then _
        WriteISO = -5: Exit Function

    'check for a valid ISO9660 image
    Open sFile For Binary As FF
        iso_hdr = Space(32774)
        Get FF, , iso_hdr
    Close FF
    If Not Right$(iso_hdr, 5) = "CD001" Then WriteISO = -4: Exit Function

    'open the image
    Open sFile For Binary As FF

    'get the number of sectors to write
    sectors = CLng(FileLen(sFile) / (UBound(sector) + 1))

    'send the WMP (Mode-1 TAO)
    If Not WriteParams(DrvID, TestMode, False, 150, 1, 4, 8, False) Then _
        WriteISO = -2: Exit Function

    'perform optimal power calibration
    SendOPCInformation DrvID, True, 0, 0

    'write sectors
    For i = 0 To sectors

        If bCancel Then
            RaiseEvent Canceling
            'cancel writing?
            bCancel = False
            'close the disc
            If Not TestMode Then
                CloseCD DrvID, 2, 0
            End If
            WriteISO = 2
            Exit Function
        End If

        'read 8 sectors
        Get FF, , sector

        'write it
        If Not Write10( _
                    DrvID, i, (UBound(sector) + 1) \ 2048, _
                    VarPtr(sector(0)), UBound(sector) + 1 _
        ) Then

            'write error!!
            WriteISO = -1: Close FF: Exit Function

        End If

        RaiseEvent WriteProgress(Fix((i / sectors) * 100))
        DoEvents

    Next i

    Close FF


    RaiseEvent Finalizing

    'synchronize the cache
    SyncCache DrvID, i + 1

    'if not in test-mode, close the disc
    If Not TestMode Then
        If Not CloseCD(DrvID, 2, 0) Then
            WriteISO = 0
            Exit Function
        End If
    End If

    'success!
    WriteISO = 1
End Function
