VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cEnumDrvs"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

'**************************************
'** Function: EnumerateDrives
'** Returns: Long
'** Usage: proxy function for ASPIEnumerateDrives and SPTIEnumerateDrives,
'**        depending on bUseASPI
'**************************************
Public Function EnumerateDrives(ByRef drives As cDrives) As Long
    Select Case bUseASPI
        Case True
            EnumerateDrives = ASPIEnumerateDrives(drives)
        Case False
            EnumerateDrives = SPTIEnumerateDrives(drives)
    End Select
End Function

'**************************************
'** Function: EnumerateDrives
'** Returns: Long
'** Usage: Search for CD-ROMs through the ASPI and fill a drive collection
'**************************************
Private Function ASPIEnumerateDrives(ByRef drives As cDrives) As Long
    Dim inquiry As SRB_HAInquiry        ' Host Adapter Inquiry
    Dim DevType As SRB_GetDevType       ' Device Type Inquiry
    Dim DrvNfo As New cDriveInfo        ' Drive Info Class
    Dim mmc As t_MMC                    ' Multimedia Capabilities
    Dim HAcnt As Integer                ' Host Adapter Count
    Dim i As Integer, j As Integer, k As Integer  ' counters

    If Not FileExists(AddSlash(GetSysDir) & "wnaspi32.dll") Then
        'no ASPI driver present
        ASPIEnumerateDrives = -1
        Exit Function
    End If

    HAcnt = GetHostAdapters             ' get the number of host adapters

    If HAcnt = -1 Then                  ' ASPI Error
        ASPIEnumerateDrives = -1
        Exit Function
    End If

    If HAcnt = 0 Then                   ' no Host Adapters present?
        ASPIEnumerateDrives = -1
        Exit Function
    End If

    'clear the drive collection
    For i = 1 To drives.Count - 1
        drives.Remove i
    Next

    'go through every Host Adapter
    For i = 0 To HAcnt

        inquiry.SRB_Cmd = SC_HA_INQUIRY     ' Host Adapter Inquiry
        inquiry.SRB_HaID = i                ' Host Adapter Identifier

        SendASPI32CommandEx inquiry         ' Execute the HA Inquiry
        If (inquiry.SRB_Status <> SS_COMP) Then GoTo SkipAdapter    ' on problems
                                                                    ' skip the HA

        'every HA can carry up to 8 targets
        For j = 0 To 7

            DevType.SRB_Cmd = SC_GET_DEV_TYPE   ' Device Type Inquiry
            DevType.SRB_HaID = i                ' Host Adapter Identifier
            DevType.SRB_Target = j              ' Target Identifier

            SendASPI32CommandEx DevType         ' execute the Device Type Inquiry
            If DevType.SRB_Status <> SS_COMP Then GoTo skipDevice   ' on problems
                                                                    ' skip the device

            'found CD-ROM?
            If DevType.DEV_DeviceType = DTYPE_CDROM Then

                'request the device's capabilities
                If Not ModeSense10(i & "0000" & j, &H2A, VarPtr(mmc), Len(mmc)) Then
                    'didn't work, wether the drive can't be accessed at the moment,
                    'it isn't a CD-ROM at all, or it isn't MMC compatible...
                    GoTo skipDevice
                End If

                With mmc
                    DrvNfo.ReadCDR = IsBitSet(.ReadSupported, 0)    ' Can read CD-R?
                    DrvNfo.ReadCDRW = IsBitSet(.ReadSupported, 1)   ' Can read CD-RW?
                    DrvNfo.ReadDVDROM = IsBitSet(.ReadSupported, 3) ' Can read DVD-ROM?
                    DrvNfo.ReadDVDR = IsBitSet(.ReadSupported, 4)   ' Can read DVD-R?
                    DrvNfo.ReadDVDRAM = IsBitSet(.ReadSupported, 5) ' Can read DVD-RAM?
                    DrvNfo.ReadDASectors = IsBitSet(.misc(1), 0)    ' Can read CDDA?
                    DrvNfo.Mode2Form1 = IsBitSet(.misc(0), 4)       ' Can read M2F1?
                    DrvNfo.Mode2Form2 = IsBitSet(.misc(0), 5)       ' Can read M2F2?
                    DrvNfo.MultiSession = IsBitSet(.misc(0), 6)     ' Can handle MM?
                    DrvNfo.ReadISRC = IsBitSet(.misc(1), 5)         ' Can read ISRC?
                    DrvNfo.ReadUPC = IsBitSet(.misc(1), 6)          ' Can read UPC?
                    DrvNfo.ReadBarCode = IsBitSet(.misc(1), 7)      ' Can read BC?

                    DrvNfo.ReadSubChannels = IsBitSet(.misc(1), 2)  ' Can read Subs?
                    DrvNfo.ReadSubChannelsCorrected = IsBitSet(.misc(1), 3) ' Can read Subs interleaved?
                    DrvNfo.ReadSubChannelsLeadIn = IsBitSet(.misc(3), 5)    ' Can read Subs from the Lead-In?

                    DrvNfo.WriteCDR = IsBitSet(.WriteSupported, 0)  ' Can write CD-R?
                    DrvNfo.WriteCDRW = IsBitSet(.WriteSupported, 1) ' Can write CD-RW?
                    DrvNfo.TestMode = IsBitSet(.WriteSupported, 2)  ' Can write in Test Mode?
                    DrvNfo.WriteDVDR = IsBitSet(.WriteSupported, 4) ' Can write DVD-R?
                    DrvNfo.WriteDVDRAM = IsBitSet(.WriteSupported, 5) ' Can write DVD-RAM?
                    DrvNfo.BURNProof = IsBitSet(.misc(0), 7)        ' Hast BURN-Proof technologie?
                    DrvNfo.BufferSize = LShift(.BufferSize(0), 8) Or .BufferSize(1) ' Read/Write Buffer Size

                    DrvNfo.MaxReadSpeed = LShift(.MaxReadSpeed(0), 8) Or .MaxReadSpeed(1)
                    DrvNfo.CurrReadSpeed = LShift(.CurrReadSpeed(0), 8) Or .CurrReadSpeed(1)
                    DrvNfo.MaxWriteSpeed = LShift(.MaxWriteSpeed(0), 8) Or .MaxWriteSpeed(1)
                    DrvNfo.CurrWriteSpeed = LShift(.CurrWriteSpeed(0), 8) Or .CurrWriteSpeed(1)

                    DrvNfo.AnalogAudio = IsBitSet(.misc(0), 0)      ' Can play CDDA analog?
                    DrvNfo.JitterCorrection = IsBitSet(.misc(1), 1) ' Has built in Jitter-Correction?
                    DrvNfo.ReadC2ErrorPointers = IsBitSet(.misc(1), 4)  ' Can read and handle C2 Error Pointers?
                    DrvNfo.LockMedia = IsBitSet(.misc(2), 0)        ' Can lock the media?

                    'determine the loading mechanism
                    If IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_CADDY
                    ElseIf IsBitSet(.misc(2), 5) And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_TRAY
                    ElseIf IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_POPUP
                    ElseIf IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) Then
                        DrvNfo.LoadingMechanism = LOAD_CHANGER
                    ElseIf IsBitSet(.misc(2), 5) And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) Then
                        DrvNfo.LoadingMechanism = LOAD_CHANGER
                    Else
                        DrvNfo.LoadingMechanism = LOAD_UNKNWN
                    End If

                    If WriteParams(i & "0000" & j, False, False, 150, 1, 0, 0, False) Then _
                        DrvNfo.WriteModeTao = True

                    If WriteParams(i & "0000" & j, True, False, 150, 1, 0, 0, False) Then _
                        DrvNfo.WriteModeTaoTest = True

                    If WriteParams(i & "0000" & j, False, False, 150, 2, 0, 8, False) Then _
                        DrvNfo.WriteModeSao = True

                    If WriteParams(i & "0000" & j, True, False, 150, 2, 0, 8, False) Then _
                        DrvNfo.WriteModeSaoTest = True

                    If WriteParams(i & "0000" & j, False, False, 150, 3, 0, 1, False) Then _
                        DrvNfo.WriteModeRaw16 = True

                    If WriteParams(i & "0000" & j, True, False, 150, 3, 0, 1, False) Then _
                        DrvNfo.WriteModeRaw16Test = True

                    If WriteParams(i & "0000" & j, False, False, 150, 3, 0, 3, False) Then _
                        DrvNfo.WriteModeRaw96 = True

                    If WriteParams(i & "0000" & j, True, False, 150, 3, 0, 3, False) Then _
                        DrvNfo.WriteModeRaw96Test = True

                    DrvNfo.Interface = GetInterface(i & "0000" & j) ' Has information about its phyiscal interface?
                End With

                'add the drive to the drive collection
                drives.Add i & "0000" & j, InquiryDrive(i & "0000" & j, INQ_VENDOR), _
                           InquiryDrive(i & "0000" & j, INQ_NAME), InquiryDrive(i & "0000" & j, INQ_FIRMWARE), _
                           i, j, IIf(IsNT, Chr(GetDrvChrASPINT(i & "0000" & j)), Chr(GetDrvChrASPI32(i & "0000" & j))), DrvNfo

            End If

skipDevice:
        Next j

SkipAdapter:
    Next i

    'no drives found?
    If drives.Count = 0 Then
        ASPIEnumerateDrives = 0
        Exit Function
    End If

    'success!
    ASPIEnumerateDrives = 1
End Function

'*********************************
'** Function: SPTIEnumerateDrives
'** Returns: Long
'**          0: No HAs found
'**          1: Success
'** Usage: enumerates all drives through the SPTI
'*********************************
Private Function SPTIEnumerateDrives(ByRef drives As cDrives) As Long
    Dim i As Integer                     ' counter
    Dim DrvNfo As New cDriveInfo         ' Drive Info Class
    Dim mmc As t_MMC                     ' Multimedia Capabilities
    Dim fh As Long                       ' drive handle
    Dim pscsiAddr As t_SCSI_ADDRESS      ' SCSI address
    Dim returned As Long, status As Long ' return buffers
    Dim OL As OVERLAPPED                 ' overlapped buffer

    'go through every possible drive location
    For i = 1 To 26
        'is there a CD-ROM...?
        If GetDriveType(ChrW$(i + 64) & ":") <> DTYPE_CDROM Then GoTo Nextdrv

        'If Inquiry fails, something can't be right...
        If Not InquiryDrive(ChrW$(i + 64), INQ_NAME) = "Error" Then

                'request the device's capabilities
                If Not ModeSense10(ChrW$(i + 64), &H2A, VarPtr(mmc), Len(mmc)) Then
                    'didn't work, the drive can't be accessed at the moment,
                    'it isn't a CD-ROM at all, or it isn't MMC compatible...
                    GoTo Nextdrv
                End If

                With mmc
                    DrvNfo.ReadCDR = IsBitSet(.ReadSupported, 0)    ' Can read CD-R?
                    DrvNfo.ReadCDRW = IsBitSet(.ReadSupported, 1)   ' Can read CD-RW?
                    DrvNfo.ReadDVDROM = IsBitSet(.ReadSupported, 3) ' Can read DVD-ROM?
                    DrvNfo.ReadDVDR = IsBitSet(.ReadSupported, 4)   ' Can read DVD-R?
                    DrvNfo.ReadDVDRAM = IsBitSet(.ReadSupported, 5) ' Can read DVD-RAM?
                    DrvNfo.ReadDASectors = IsBitSet(.misc(1), 0)    ' Can read CDDA?
                    DrvNfo.Mode2Form1 = IsBitSet(.misc(0), 4)       ' Can read M2F1?
                    DrvNfo.Mode2Form2 = IsBitSet(.misc(0), 5)       ' Can read M2F2?
                    DrvNfo.MultiSession = IsBitSet(.misc(0), 6)     ' Can handle MM?
                    DrvNfo.ReadISRC = IsBitSet(.misc(1), 5)         ' Can read ISRC?
                    DrvNfo.ReadUPC = IsBitSet(.misc(1), 6)          ' Can read UPC?
                    DrvNfo.ReadBarCode = IsBitSet(.misc(1), 7)      ' Can read BC?

                    DrvNfo.ReadSubChannels = IsBitSet(.misc(1), 2)  ' Can read Subs?
                    DrvNfo.ReadSubChannelsCorrected = IsBitSet(.misc(1), 3) ' Can read Subs interleaved?
                    DrvNfo.ReadSubChannelsLeadIn = IsBitSet(.misc(3), 5)    ' Can read Subs from the Lead-In?

                    DrvNfo.WriteCDR = IsBitSet(.WriteSupported, 0)  ' Can write CD-R?
                    DrvNfo.WriteCDRW = IsBitSet(.WriteSupported, 1) ' Can write CD-RW?
                    DrvNfo.TestMode = IsBitSet(.WriteSupported, 2)  ' Can write in Test Mode?
                    DrvNfo.WriteDVDR = IsBitSet(.WriteSupported, 4) ' Can write DVD-R?
                    DrvNfo.WriteDVDRAM = IsBitSet(.WriteSupported, 5) ' Can write DVD-RAM?
                    DrvNfo.BURNProof = IsBitSet(.misc(0), 7)        ' Hast BURN-Proof technologie?
                    DrvNfo.BufferSize = LShift(.BufferSize(0), 8) Or .BufferSize(1) ' Read/Write Buffer Size

                    DrvNfo.MaxReadSpeed = LShift(.MaxReadSpeed(0), 8) Or .MaxReadSpeed(1)
                    DrvNfo.CurrReadSpeed = LShift(.CurrReadSpeed(0), 8) Or .CurrReadSpeed(1)
                    DrvNfo.MaxWriteSpeed = LShift(.MaxWriteSpeed(0), 8) Or .MaxWriteSpeed(1)
                    DrvNfo.CurrWriteSpeed = LShift(.CurrWriteSpeed(0), 8) Or .CurrWriteSpeed(1)

                    DrvNfo.AnalogAudio = IsBitSet(.misc(0), 0)      ' Can play CDDA analog?
                    DrvNfo.JitterCorrection = IsBitSet(.misc(1), 1) ' Has built in Jitter-Correction?
                    DrvNfo.ReadC2ErrorPointers = IsBitSet(.misc(1), 4)  ' Can read and handle C2 Error Pointers?
                    DrvNfo.LockMedia = IsBitSet(.misc(2), 0)        ' Can lock the media?

                    If WriteParams(ChrW$(i + 64), False, False, 150, 1, 0, 0, False) Then _
                        DrvNfo.WriteModeTao = True

                    If WriteParams(ChrW$(i + 64), True, False, 150, 1, 0, 0, False) Then _
                        DrvNfo.WriteModeTaoTest = True

                    If WriteParams(ChrW$(i + 64), False, False, 150, 2, 0, 8, False) Then _
                        DrvNfo.WriteModeSao = True

                    If WriteParams(ChrW$(i + 64), True, False, 150, 2, 0, 8, False) Then _
                        DrvNfo.WriteModeSaoTest = True

                    If WriteParams(ChrW$(i + 64), False, False, 150, 3, 0, 1, False) Then _
                        DrvNfo.WriteModeRaw16 = True

                    If WriteParams(ChrW$(i + 64), True, False, 150, 3, 0, 1, False) Then _
                        DrvNfo.WriteModeRaw16Test = True

                    If WriteParams(ChrW$(i + 64), False, False, 150, 3, 0, 3, False) Then _
                        DrvNfo.WriteModeRaw96 = True

                    If WriteParams(ChrW$(i + 64), True, False, 150, 3, 0, 3, False) Then _
                        DrvNfo.WriteModeRaw96Test = True

                    'determine the loading mechanism
                    If IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_CADDY
                    ElseIf IsBitSet(.misc(2), 5) And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_TRAY
                    ElseIf IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) And IsBitSet(.misc(2), 7) = False Then
                        DrvNfo.LoadingMechanism = LOAD_POPUP
                    ElseIf IsBitSet(.misc(2), 5) = False And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) Then
                        DrvNfo.LoadingMechanism = LOAD_CHANGER
                    ElseIf IsBitSet(.misc(2), 5) And IsBitSet(.misc(2), 6) = False And IsBitSet(.misc(2), 7) Then
                        DrvNfo.LoadingMechanism = LOAD_CHANGER
                    Else
                        DrvNfo.LoadingMechanism = LOAD_UNKNWN
                    End If

                    DrvNfo.Interface = GetInterface(ChrW$(i + 64))  ' Has information about its phyiscal interface?
                End With

                'get the current drive's handle
                If GetDriveHandle(ChrW$(i + 64), fh) Then

                    'ask it for its SCSI address
                    status = DeviceIoControl(fh, IOCTL_SCSI_GET_ADDRESS, _
                                             0, 0, pscsiAddr, Len(pscsiAddr), _
                                             returned, OL)

                    'close that handle!!
                    CloseHandle fh

                    'success?
                    If status = 1 Then

                        'add the drive to the drive collection
                        drives.Add ChrW$(i + 64), InquiryDrive(ChrW$(i + 64), INQ_VENDOR), _
                                   InquiryDrive(ChrW$(i + 64), INQ_NAME), InquiryDrive(ChrW$(i + 64), INQ_FIRMWARE), _
                                   pscsiAddr.PortNumber, pscsiAddr.TargetId, ChrW$(i + 64), DrvNfo

                    End If

                End If

        End If

'Next one!!
Nextdrv:
    Next i

    If drives.Count = 0 Then
        SPTIEnumerateDrives = 0                     ' No drives found
    Else
        SPTIEnumerateDrives = 1                     ' everything fine
    End If
End Function

Private Sub Class_Initialize()
    Dim drives As New cDrives

    'If NT kernel, then ASPI = False
    If IsNT Then

        bUseASPI = False

        'didn't find any drives?
        If EnumerateDrives(drives) < 1 Then

            'switch ASPI on
            bUseASPI = True

            'if this doesn't find any drives, switch back and exit the sub :)
            If EnumerateDrives(drives) < 1 Then
                bUseASPI = False
                Exit Sub
            End If

        End If

    'If 32 kernel, then ASPI = True, but some XP SP2 machines
    'say they are Win9x/Me...
    Else
        bUseASPI = True

        If EnumerateDrives(drives) < 1 Then
            bUseASPI = False

            If EnumerateDrives(drives) < 1 Then
                'Debug.Print "No drives through SPTI..."
            Else
                bUseASPI = True
                Exit Sub
            End If

        End If
    End If
End Sub

'**************************************
'** Function: GetDrvChrASPI32
'** Returns: Byte
'**             0: error
'**            >0: Ascii Char
'** Usage: detects the drive char of a drive found with ASPI
'**        through the Int 13h function associated with the
'**        ASPI GetDiskInfo command.
'**        Works only for Win9x/Me.
'**************************************
Private Function GetDrvChrASPI32(ByVal DrvID As String) As Byte
    Dim diskinfo As SRB_GetDiskInfo

    diskinfo.SRB_Cmd = SC_GET_DISK_INFO         ' ASPI GetDiskInfo command
    diskinfo.SRB_HaID = ExtractHAID(DrvID, 0)   ' Host Adapter identifier
    diskinfo.SRB_Target = ExtractHAID(DrvID, 1) ' Target identifier
    diskinfo.SRB_Flags = SRB_DIR_IN             ' get data

    SendASPI32CommandEx diskinfo                ' execute

    GetDrvChrASPI32 = diskinfo.SRB_Int13HDriveInfo + 65 ' return the drive char
End Function

'**************************************
'** Function: GetDrvChrASPINT
'** Returns: Byte
'**             0: error
'**            >0: Ascii Char
'** Usage: detects the drive char of a drive found with ASPI
'**        through the DeviceIoControl function "GET SCSI ADDRESS".
'**        Works only for NT/2000/XP.
'**************************************
Private Function GetDrvChrASPINT(ByVal DrvID As String) As Byte
    Dim pscsiAddr As t_SCSI_ADDRESS
    Dim fh As Long: fh = FreeFile
    Dim i As Integer
    Dim returned As Long, status As Long
    Dim OL As OVERLAPPED
    Dim sd As SECURITY_ATTRIBUTES

    'go through all the 26 possible drives
    For i = 1 To 26

        'CD-ROM?
        If GetDriveType(Chr(i + 64) & ":") = DTYPE_CDROM Then
            'yep, get the drive handle
            If GetDriveHandle(Chr(i + 64), fh) Then

                'ask for the SCSI Address
                pscsiAddr.Length = Len(pscsiAddr)
                status = DeviceIoControl(fh, IOCTL_SCSI_GET_ADDRESS, pscsiAddr, Len(pscsiAddr), _
                                         pscsiAddr, Len(pscsiAddr), returned, OL)
                'we have full rights, don't lose the handle!! so better close it...
                CloseHandle fh

                'worked?
                If status = 1 Then
                    With pscsiAddr
                        'compare it with the ASPI Drive ID
                        If .PortNumber = ExtractHAID(DrvID, 0) And .TargetId = ExtractHAID(DrvID, 1) Then
                            'yeah! found it!
                            GetDrvChrASPINT = i + 64
                            Exit Function
                        End If
                    End With
                End If

            End If
        End If

    Next
End Function
