VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cISO9660Read"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Type ISO_VOL_DESCRIPTOR                 'ISO9660 Volume Descriptor
    Vol_Desc_Type As Byte                       'Volume Descriptor Type (1)
    Std_ID        As String * 6                 'Standard Identifier (CD001 )
    Vol_Desc_Ver  As Byte                       'Volume Descriptor Version (0)
End Type

Private Type ISO_HEADER                         'ISO9660 Header
     zeros       As String                      'some zeros at the beginning
     Vol_Desc    As ISO_VOL_DESCRIPTOR          'Primary volume descriptor
End Type

Private Enum eFileFlags                         'Directory Record Flags
    FLAG_HIDDEN = &H1                           'Record Flag Hidden
    FLAG_DIRECTORY = &H2                        'Record Flag Directory
    FLAG_ASSOCIATED = &H4                       'Record Flag associated file
    FLAG_RECFMTSPEC = &H8                       'Record Flag Fecord Format Specific
    FLAG_PERM = &H16                            'Record Flag Permission
    FLAG_NOTFINAL = &H80                        'Record Flag Not Final Version
End Enum

Private Enum eVolDesc                           'Enumeration of Volume Descriptor Types
    VOL_DESC_BOOT_RECORD                        'Boot Record
    VOL_DESC_PRIM_VOL_DESCRIPTOR                'Primary Volume Descriptor
    VOL_DESC_SUPPL_VOL_DESCRIPTOR               'Supplementary Volume Descriptor
    VOL_DESC_VOL_PART_DESCRIPTOR                'Volume Partition Descriptor
    VOL_DESC_SET_TERMINATOR                     'Volume Descriptor Set Terminator
End Enum

Private Type PTableRecord                       'path table record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    FirstSector As String * 4                   'First sector in the directory
    RecNumParentDir As String * 2               'Record number of the parent directory
End Type

Private Type DirRecord                          'Directory record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    LocExt As String * 8                        'Location of Extent
    DataLen As String * 8                       'Data Length
    RecDateTime As String * 7                   'Recording Date & Time
    FileFlags As Byte                           'File Flags
    FileUnitSize As Byte                        'File Unit Size
    InterleaveGapSize As Byte                   'InterleaveGapSize
    VolSeqNum As String * 4                     'Volume Sequence Number
    FileIDLen As Byte                           'Length of filename
End Type

Private Type RootDirRecord                      'Root Directory record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    LocExt As String * 8                        'Location of Extent
    DataLen As String * 8                       'Data Length
    RecDateTime As String * 7                   'Recording Date & Time
    FileFlags As Byte                           'File Flags
    FileUnitSize As Byte                        'File Unit Size
    InterleaveGapSize As Byte                   'InterleaveGapSize
    VolSeqNum As String * 4                     'Volume Sequence Number
    FileIDLen As Byte                           'Length of filename
    FileID(0) As Byte                           'Filename
End Type

Private Type ISO_PRIMARY_DESCRIPTOR             'Primary Volume Descriptor
    System_ID As String * 32                    'System ID
    Volume_ID As String * 32                    'Volume ID

    unused2(1 To 8) As Byte                     '8 unused bytes

    Vol_Space_Size As String * 8                'Volume Space Size
    unused3(1 To 32) As Byte                    '32 unused bytes
    Vol_Set_Size As String * 4                  'Volume Set Size (1)
    Vol_Seq_Number As String * 4                'Volume Sequence Number (1)
    Logical_Block_Size As String * 4            'Logical Block Size (2048)

    Path_Table_Size As String * 8               'Path Table Size
    Location_TypeL_Path_Table As String * 4     'Location of Type L Path Table
    Location_TypeL_Opt_Path_Table As String * 4 'Location of optional Type L Path Table
    Location_TypeM_Path_Table As String * 4     'Location of Type M Path Table
    Location_TypeM_Opt_Oath_Table As String * 4 'Location of optional Type M Path Table
    Dir_Root_Record As RootDirRecord            'Directory Record for Root Dir

    Vol_Set_ID As String * 128                  'Volume Set Identifier
    Publisher_ID As String * 128                'Publisher Identifier
    Data_Preparer_ID As String * 128            'Data Preparer Identifier
    App_ID As String * 128                      'Application Identifie

    Copyright_File_ID As String * 37            'Copyright File Identifier
    Abstract_File_ID As String * 37             'Abstract File Identifier
    Bibliographic_File_ID As String * 37        'Bibliographic File Idenfitier

    Vol_Creation_Date As String * 17            'Volume Creation Date & Time
    Vol_Modification_Date As String * 17        'Volume Modification Date & Time
    Vol_Expiration_Date As String * 17          'Volume Expiration Date & Time
    Vol_Effective_Date As String * 17           'Volume Effective Date & Time

    File_Structure_Ver As Byte                  'File Structure Version (1)
    rsvd As Byte                                'reserved
    app_use(1 To 512) As Byte                   'reserved for application use
    rsvd2(1 To 653) As Byte                     'reserved
End Type

Private sVolID As String
Private sSysID As String
Private sAppID As String
Private sPubID As String
Private sPrepID As String
Private sVolCreateDate As String
Private sVolModDate As String
Private sVolExpDate As String
Private sVolEffDate As String
Private sAbstrFile As String
Private sBiblioFile As String
Private sCopyFile As String
Private lVolSize As Long
Private sCurrFile As String

Private dirs() As String
Private files() As String
Private filecnt As Long

'some properties of the ISO image
Public Property Get CurrentISO() As String
    CurrentISO = sCurrFile
End Property

Public Property Get VolumeID() As String
    VolumeID = sVolID
End Property

Public Property Get SystemID() As String
    SystemID = sSysID
End Property

Public Property Get ApplicationID() As String
    ApplicationID = sAppID
End Property

Public Property Get PublisherID() As String
    PublisherID = sPubID
End Property

Public Property Get DataPreparerID() As String
    DataPreparerID = sPrepID
End Property

Public Property Get VolumeCreationDate() As String
    VolumeCreationDate = sVolCreateDate
End Property

Public Property Get VolumeModificationDate() As String
    VolumeModificationDate = sVolModDate
End Property

Public Property Get VolumeExpirationDate() As String
    VolumeExpirationDate = sVolExpDate
End Property

Public Property Get VolumeEffectiveDate() As String
    VolumeEffectiveDate = sVolEffDate
End Property

Public Property Get AbstractFile() As String
    AbstractFile = sAbstrFile
End Property

Public Property Get BibliographicFile() As String
    BibliographicFile = sBiblioFile
End Property

Public Property Get CopyrightFile() As String
    CopyrightFile = sCopyFile
End Property

Public Property Get VolumeSize() As Long
    VolumeSize = lVolSize
End Property



'add a slash to the end of a string
Private Function AddSlash(ByVal sVal As String) As String
    If Not Right$(sVal, 1) = "\" Then sVal = sVal & "\"
    AddSlash = sVal
End Function

'get the size (in bytes) of a string in Mode-1 (2048 bytes) sectors
Private Function GetCDFSSize(ByVal length As Long) As Long
    Dim i As Integer, j As Long
    While Not i = -1
        j = j + 1
        If j * 2048 >= length Then i = -1
    Wend
    GetCDFSSize = j * 2048
End Function

'shift bits to the left (like the << operator in C++)
Private Function LShift(ByVal pnValue As Long, ByVal pnShift As Long) As Long
    LShift = pnValue * (2 ^ pnShift)
End Function

'shift bits to the right (like the >> operator in C++)
Private Function RShift(ByVal pnValue As Long, ByVal pnShift As Long) As Long
    RShift = CLng(pnValue \ (2 ^ pnShift))
End Function

'convert a 16bit least significant byte ordered value to a VB6 integer
Private Function From16BitLS(l As String) As Integer
    Dim b0 As Byte, b1 As Byte
    b0 = Asc(Right(l, 1))
    b1 = Asc(Left(l, 1))
    From16BitLS = b0 Or LShift(b1, 8)
End Function

'convert a 16bit most significant byte ordered value to a VB6 integer
Private Function From16BitMS(l As String) As Integer
    Dim b0 As Byte, b1 As Byte
    b0 = Asc(Right(l, 1))
    b1 = Asc(Left(l, 1))
    From16BitMS = b1 Or LShift(b0, 8)
End Function

'convert a 32bit least significant byte ordered value to a VB6 long
Private Function From32BitLS(l As String) As Long
    Dim b0 As Byte, b1 As Byte, b2 As Byte, b3 As Byte
    b0 = Asc(Left$(l, 1))
    b1 = Asc(Mid$(l, 2, 1))
    b2 = Asc(Mid$(l, 3, 1))
    b3 = Asc(Right$(l, 1))
    From32BitLS = LShift(b0, 24) Or LShift(b1, 16) Or LShift(b2, 8) Or b3
End Function

'convert a 32bit most significant byte ordered value to a VB6 long
Private Function From32BitMS(l As String) As Long
    Dim b0 As Byte, b1 As Byte, b2 As Byte, b3 As Byte
    b0 = Asc(Left$(l, 1))
    b1 = Asc(Mid$(l, 2, 1))
    b2 = Asc(Mid$(l, 3, 1))
    b3 = Asc(Right$(l, 1))
    From32BitMS = LShift(b3, 24) Or LShift(b2, 16) Or LShift(b1, 8) Or b0
End Function

'format a volume descriptor date
Private Function VolDateToStr(ByVal sDate As String) As String
    Dim sYear As String, sMonth As String, sDay As String
    Dim sHour As String, sMinute As String, sSecond As String
    sYear = Left$(sDate, 4)
    sMonth = Mid$(sDate, 5, 2)
    sDay = Mid$(sDate, 7, 2)
    sHour = Mid$(sDate, 10, 2)
    sMinute = Mid$(sDate, 13, 2)
    sSecond = Mid$(sDate, 16, 2)
    VolDateToStr = sDay & "," & sMonth & "," & sYear & " " & sHour & ":" & sMinute & ":" & sSecond
End Function

'get the directory name of an absolute path
Public Function GetDirname(ByVal sDir As String) As String
    Dim parts() As String
    parts = Split(AddSlash(sDir), "\")
    GetDirname = parts(UBound(parts) - 1)
End Function

'get the parent dir of a directory
Public Function GetUpperDir(ByVal sDir As String) As String
    Dim parts() As String, i As Integer
    sDir = AddSlash(sDir)
    parts = Split(sDir, "\")
    For i = 0 To UBound(parts) - 2
        GetUpperDir = GetUpperDir & parts(i) & "\"
    Next
    GetUpperDir = AddSlash(GetUpperDir)
End Function

'checks if a directory is the parent directory of an other directory
Private Function IsSubDir(ByVal sDir1 As String, ByVal sDir2 As String) As Boolean
    If GetUpperDir(sDir2) = sDir1 Then IsSubDir = True
End Function

Private Function ReadSector(ByVal sPath As String, ByVal lLBA As Long, Optional ByVal lSectors As Long = 1) As String
    Dim FF As Integer: FF = FreeFile
    'read sectors from the ISO image
    Open sPath For Binary As #FF
        ReadSector = Space(lSectors * 2048)
        Seek FF, lLBA * 2048 + 1
        Get #FF, , ReadSector
    Close #FF
End Function




Public Function HasSubDirs(ByVal sPath As String) As Boolean
    Dim i As Integer

    'check if a directory has directories inside
    sPath = LCase(sPath)
    For i = 1 To UBound(dirs)
        If LCase(GetUpperDir(dirs(i))) = sPath Then
            HasSubDirs = True
            Exit Function
        End If
    Next
End Function

Public Function HasSubFiles(ByVal sPath As String) As Boolean
    Dim i As Integer

    'check if a directory has files inside
    sPath = LCase(AddSlash(sPath))
    For i = 0 To UBound(files)
        If Not files(i) = vbNullString Then
            If LCase(GetUpperDir(AddSlash(files(i)))) = sPath Then
                HasSubFiles = True
                Exit Function
            End If
        End If
    Next
End Function

Public Function GetSubDirs(ByVal sPath As String) As Variant
    Dim i As Integer
    Dim dirz() As String
    Dim dircnt As Integer

    'get all directories in a directory
    sPath = LCase(sPath)
    For i = 1 To UBound(dirs)
        If Not (dirs(i) = "\" And sPath = "\") Then
            If LCase(GetUpperDir(dirs(i))) = sPath Then
                ReDim Preserve dirz(dircnt)
                dirz(dircnt) = dirs(i)
                dircnt = dircnt + 1
            End If
        End If
    Next
    GetSubDirs = dirz
End Function

Public Function GetSubFiles(ByVal sPath As String) As Variant
    On Error Resume Next
    Dim i As Integer
    Dim filez() As String
    Dim fcnt As Long

    'get all files in a directory
    sPath = LCase(AddSlash(sPath))
    For i = 0 To UBound(files)
        If LCase(GetUpperDir(AddSlash(files(i)))) = sPath Then
            ReDim Preserve filez(fcnt)
            filez(fcnt) = Mid$(Left$(files(i), InStr(1, files(i), ChrW$(0)) - 1), Len(sPath) + 1)
            fcnt = fcnt + 1
        End If
    Next
    GetSubFiles = filez
End Function

Public Function GetFilesize(ByVal sFile As String) As Long
    On Error Resume Next
    Dim i As Integer
    Dim sCurr As String
    Dim parts() As String

    'get the size of a file inside the ISO image
    sFile = LCase(sFile)
    For i = 0 To UBound(files)
        If LCase(Left$(files(i), InStr(1, files(i), ChrW$(0)) - 1)) = sFile Then
            parts = Split(files(i), Chr(0))
            GetFilesize = parts(2)
            Exit Function
        End If
    Next
End Function

Public Function GetFileDate(ByVal sFile As String) As String
    On Error Resume Next
    Dim i As Integer
    Dim sCurr As String
    Dim parts() As String

    'get the size of a file inside the ISO image
    sFile = LCase(sFile)
    For i = 0 To UBound(files)
        If LCase(Left$(files(i), InStr(1, files(i), ChrW$(0)) - 1)) = sFile Then
            parts = Split(files(i), Chr(0))
            GetFileDate = parts(3)
            Exit Function
        End If
    Next
End Function

Public Function ReadFileToStr(ByVal sFile As String) As String
    Dim i As Integer
    Dim sCurr As String
    Dim parts() As String

    'read a file from the ISO image and return it as a string
    sFile = LCase(sFile)
    For i = 0 To UBound(files)
        If LCase(Left$(files(i), InStr(1, files(i), ChrW$(0)) - 1)) = sFile Then
            parts = Split(files(i), Chr(0))
            ReadFileToStr = ReadFileStr(sCurrFile, parts(1) * 2048, parts(2))
            Exit Function
        End If
    Next
End Function

Public Function ReadFileToFile(ByVal sFile As String, ByVal sDest As String) As String
    Dim i As Integer
    Dim sCurr As String
    Dim parts() As String
    Dim FF As Integer: FF = FreeFile

    'read a file from the ISO image and save it to hard disk
    sFile = LCase(sFile)
    For i = 0 To UBound(files)
        If LCase(Left$(files(i), InStr(1, files(i), ChrW$(0)) - 1)) = sFile Then
            parts = Split(files(i), Chr(0))
            Open sDest For Binary As #FF
                Put #FF, , ReadFileStr(sCurrFile, parts(1) * 2048, parts(2))
            Close #FF
            Exit Function
        End If
    Next
End Function

Private Function ReadFileStr(ByVal sFile As String, ByVal start As Long, ByVal length As Long) As String
    Dim FF As Integer: FF = FreeFile

    'read a part from a file to a string
    Open sFile For Binary As #FF
        ReadFileStr = Space(length)
        Seek #FF, start + 1
        Get #FF, , ReadFileStr
    Close #FF
End Function

Private Sub ParseFileTable(ByVal sPath As String, ByVal lStart As Long, _
                           ByVal lLen As Long, ByVal sParent As String)

    On Error Resume Next
    Dim sBuf As String, sName As String, sNameLen As Integer
    Dim i As Long, dircnt As Long: dircnt = -1
    Dim lRecLen As Byte, LocExt As Long, DataLen As Long, isDir As Boolean, sDate As String
    Dim dirz() As String, splt() As String

    'get the content of the directory record's sectors
    sBuf = ReadSector(sPath, lStart, lLen \ 2048)

    'parse it
    While i > -1
        lRecLen = Asc(Mid$(sBuf, i + 1, 1))

        If lRecLen = 0 Then
            'no record found at this position, move to the next byte
            lRecLen = 1
        Else
            'get record info
            LocExt = From32BitMS(Mid$(sBuf, i + 3, 4))              ' start sector of record's content
            DataLen = From32BitMS(Mid$(sBuf, i + 11, 4))            ' data length of current record
            sDate = format(Asc(Mid$(sBuf, i + 21, 1)), "00") & "," & _
                    format(Asc(Mid$(sBuf, i + 20, 1)), "00") & "," & _
                    (Asc(Mid$(sBuf, i + 19, 1)) + 1900) & " " & _
                    format(Asc(Mid$(sBuf, i + 22)), "00") & ":" & _
                    format(Asc(Mid$(sBuf, i + 23)), "00") & ":" & _
                    format(Asc(Mid$(sBuf, i + 24)), "00")           ' get the date and the time of the record
            isDir = Asc(Mid$(sBuf, i + 26, 1)) And FLAG_DIRECTORY   ' is directory or file?
            sNameLen = Asc(Mid$(sBuf, i + 33, 1))                   ' name length
            sName = Mid$(sBuf, i + 34, sNameLen)                    ' name
            If Not sName = vbNullString Then
                If Asc(sName) > 1 Then
                    If isDir Then                                   ' add a directory for parsing
                        dircnt = dircnt + 1
                        ReDim Preserve dirz(dircnt)
                        dirz(dircnt) = sName & Chr(0) & LocExt & Chr(0) & DataLen
                    Else
                        ReDim Preserve files(filecnt)               ' add a file
                        If Right$(sName, 2) = ";1" Then sName = Left$(sName, Len(sName) - 2)
                        files(filecnt) = AddSlash(sParent) & sName & Chr(0) & LocExt & Chr(0) & DataLen & Chr(0) & sDate
                        filecnt = filecnt + 1
                    End If
                End If
            End If
        End If

        i = i + lRecLen
        If i >= lLen Then i = -1
    Wend

    'parse all found directories
    If dircnt > -1 Then
        For i = 0 To UBound(dirz)
            splt = Split(dirz(i), Chr(0))
            ParseFileTable sPath, splt(1), splt(2), AddSlash(sParent) & AddSlash(splt(0))
        Next
    End If
End Sub

Private Sub ParseDirs(ByVal sData As String, ByVal lDataLen As Long)
    Dim i As Integer: i = 1
    Dim iLen As Byte, ParentDir As Long, sName As String
    Dim dircnt As Integer: dircnt = 1

    'parse the path table
    While i > -1
        iLen = Asc(Mid$(sData, i, 1))                               ' length of directory name
        ParentDir = From16BitMS(Mid$(sData, i + 6, 2))              ' parent directorie's record ID
        sName = Mid$(sData, i + 8, iLen)                            ' directory name
        If sName = Chr(0) Then sName = "\"                          ' Chr(0) = "\" = root

        If Not sName = vbNullString Then
            ReDim Preserve dirs(dircnt)

            If Not sName = "\" Then
                dirs(dircnt) = dirs(ParentDir) & AddSlash(sName)    ' add directory
            Else
                dirs(dircnt) = sName                                ' add root
            End If
        End If

        dircnt = dircnt + 1

        If i >= lDataLen Then Exit Sub                              ' end of path table?

        i = i + iLen + 8
        If iLen Mod 2 Then i = i + 1
        If iLen = 0 Then i = i + 1
        If ParentDir = 0 Then i = i + 1
    Wend
End Sub

Public Function ReadISO(ByVal sPath As String) As Boolean
    Dim hdr As ISO_HEADER
    Dim PrimDesc As ISO_PRIMARY_DESCRIPTOR
    Dim FF As Integer: FF = FreeFile
    Dim lPathTable As Long, lFileTable As Long
    Dim lPathTableStart As Long, lFileTableStart As Long

    Open sPath For Binary As #FF
        Seek #FF, 32767
        Get #FF, , hdr
        Get #FF, , PrimDesc
    Close #FF

    If Not Left$(hdr.Vol_Desc.Std_ID, 5) = "CD001" Then Exit Function

    ReDim files(0): filecnt = 0
    ReDim dirs(0)

    sCurrFile = sPath

    'some information about the file system
    sVolID = Trim$(PrimDesc.Volume_ID)
    sSysID = Trim$(PrimDesc.System_ID)
    sAppID = Trim$(PrimDesc.App_ID)
    sPubID = Trim$(PrimDesc.Publisher_ID)
    sPrepID = Trim$(PrimDesc.Data_Preparer_ID)
    sAbstrFile = Trim$(PrimDesc.Abstract_File_ID)
    sBiblioFile = Trim$(PrimDesc.Bibliographic_File_ID)
    sCopyFile = Trim$(PrimDesc.Copyright_File_ID)
    sVolCreateDate = VolDateToStr(PrimDesc.Vol_Creation_Date)
    sVolModDate = VolDateToStr(PrimDesc.Vol_Modification_Date)
    sVolExpDate = VolDateToStr(PrimDesc.Vol_Effective_Date)
    sVolEffDate = VolDateToStr(PrimDesc.Vol_Effective_Date)
    lVolSize = From32BitMS(Left$(PrimDesc.Vol_Space_Size, 4))

    'read the little endian path table
    lPathTable = From32BitMS(Left$(PrimDesc.Path_Table_Size, 4))
    lPathTableStart = From32BitMS(PrimDesc.Location_TypeL_Path_Table)
    ParseDirs ReadSector(sPath, lPathTableStart, GetCDFSSize(lPathTable) \ 2048), lPathTable

    lFileTableStart = From32BitMS(Left$(PrimDesc.Dir_Root_Record.LocExt, 4))
    lFileTable = From32BitMS(Left$(PrimDesc.Dir_Root_Record.DataLen, 4))

    'read the file table
    ParseFileTable sPath, lFileTableStart, lFileTable, "\"

    ReadISO = True
End Function
