VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cISO9660Write"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Declare Function FindFirstFile Lib "kernel32" Alias "FindFirstFileA" ( _
    ByVal lpFileName As String, lpFindFileData As WIN32_FIND_DATA _
) As Long

Private Declare Function FindNextFile Lib "kernel32" Alias "FindNextFileA" ( _
    ByVal hFindFile As Long, lpFindFileData As WIN32_FIND_DATA _
) As Long

Private Declare Function GetFileAttributes Lib "kernel32" Alias "GetFileAttributesA" ( _
    ByVal lpFileName As String _
) As Long

Private Declare Function FindClose Lib "kernel32" ( _
    ByVal hFindFile As Long _
) As Long

Const MAX_PATH = 260
Const MAXDWORD = &HFFFF
Const INVALID_HANDLE_VALUE = -1
Const FILE_ATTRIBUTE_ARCHIVE = &H20
Const FILE_ATTRIBUTE_DIRECTORY = &H10
Const FILE_ATTRIBUTE_HIDDEN = &H2
Const FILE_ATTRIBUTE_NORMAL = &H80
Const FILE_ATTRIBUTE_READONLY = &H1
Const FILE_ATTRIBUTE_SYSTEM = &H4
Const FILE_ATTRIBUTE_TEMPORARY = &H100

Private Type FILETIME
    dwLowDateTime As Long
    dwHighDateTime As Long
End Type

Private Type WIN32_FIND_DATA
    dwFileAttributes As Long
    ftCreationTime As FILETIME
    ftLastAccessTime As FILETIME
    ftLastWriteTime As FILETIME
    nFileSizeHigh As Long
    nFileSizeLow As Long
    dwReserved0 As Long
    dwReserved1 As Long
    cFileName As String * MAX_PATH
    cAlternate As String * 14
End Type

Private Declare Function WritePrivateProfileString Lib "kernel32" Alias "WritePrivateProfileStringA" ( _
    ByVal lpApplicationName As String, _
    ByVal lpKeyName As Any, ByVal lpString As Any, _
    ByVal lpFileName As String _
) As Long

Private Declare Function GetPrivateProfileString Lib "kernel32" Alias "GetPrivateProfileStringA" ( _
    ByVal lpApplicationName As String, _
    ByVal lpKeyName As Any, ByVal lpDefault As String, _
    ByVal lpReturnedString As String, ByVal nSize As Long, _
    ByVal lpFileName As String _
) As Long

Private sCurrPrj As String                      'current project
Private sVolID As String                        'volume identifier

Dim files() As String                           'Recursively found files

Private Enum RecTypes
    TYPE_FILE                                   'Project File Entry
    TYPE_DIR                                    'Project Directory Entry
    TYPE_DELETED                                'Project Deleted Entry
End Enum

Private Type ISO_VOL_DESCRIPTOR                 'ISO9660 Volume Descriptor
    Vol_Desc_Type As Byte                       'Volume Descriptor Type (1)
    Std_ID        As String * 6                 'Standard Identifier (CD001 )
    Vol_Desc_Ver  As Byte                       'Volume Descriptor Version (0)
End Type

Private Type ISO_HEADER                         'ISO9660 Header
     zeros       As String                      'some zeros at the beginning
     Vol_Desc    As ISO_VOL_DESCRIPTOR          'Primary volume descriptor
End Type

Public Enum eFileFlags                          'Directory Record Flags
    FLAG_HIDDEN = &H1                           'Record Flag Hidden
    FLAG_DIRECTORY = &H2                        'Record Flag Directory
    FLAG_ASSOCIATED = &H4                       'Record Flag associated file
    FLAG_RECFMTSPEC = &H8                       'Record Flag Fecord Format Specific
    FLAG_PERM = &H16                            'Record Flag Permission
    FLAG_NOTFINAL = &H80                        'Record Flag Not Final Version
End Enum

Public Enum eVolDesc                            'Enumeration of Volume Descriptor Types
    VOL_DESC_BOOT_RECORD                        'Boot Record
    VOL_DESC_PRIM_VOL_DESCRIPTOR                'Primary Volume Descriptor
    VOL_DESC_SUPPL_VOL_DESCRIPTOR               'Supplementary Volume Descriptor
    VOL_DESC_VOL_PART_DESCRIPTOR                'Volume Partition Descriptor
    VOL_DESC_SET_TERMINATOR                     'Volume Descriptor Set Terminator
End Enum

Private Type PTableRecord                       'path table record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    FirstSector As String * 4                   'First sector in the directory
    RecNumParentDir As String * 2               'Record number of the parent directory
End Type

Private Type DirRecord                          'Directory record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    LocExt As String * 8                        'Location of Extent
    DataLen As String * 8                       'Data Length
    RecDateTime As String * 7                   'Recording Date & Time
    FileFlags As Byte                           'File Flags
    FileUnitSize As Byte                        'File Unit Size
    InterleaveGapSize As Byte                   'InterleaveGapSize
    VolSeqNum As String * 4                     'Volume Sequence Number
    FileIDLen As Byte                           'Length of filename
End Type

Private Type RootDirRecord                      'Root Directory record
    length As Byte                              'length of the record
    ExtAttrRecLen As Byte                       'Extended Attribute Record Length
    LocExt As String * 8                        'Location of Extent
    DataLen As String * 8                       'Data Length
    RecDateTime As String * 7                   'Recording Date & Time
    FileFlags As Byte                           'File Flags
    FileUnitSize As Byte                        'File Unit Size
    InterleaveGapSize As Byte                   'InterleaveGapSize
    VolSeqNum As String * 4                     'Volume Sequence Number
    FileIDLen As Byte                           'Length of filename
    FileID(0) As Byte                           'Filename
End Type

Private Type ISO_PRIMARY_DESCRIPTOR             'Primary Volume Descriptor

    System_ID As String * 32                    'System ID
    Volume_ID As String * 32                    'Volume ID

    unused2(1 To 8) As Byte                     '8 unused bytes

    Vol_Space_Size As String * 8                'Volume Space Size
    unused3(1 To 32) As Byte                    '32 unused bytes
    Vol_Set_Size As String * 4                  'Volume Set Size (1)
    Vol_Seq_Number As String * 4                'Volume Sequence Number (1)
    Logical_Block_Size As String * 4            'Logical Block Size (2048)

    Path_Table_Size As String * 8               'Path Table Size
    Location_TypeL_Path_Table As String * 4     'Location of Type L Path Table
    Location_TypeL_Opt_Path_Table As String * 4 'Location of optional Type L Path Table
    Location_TypeM_Path_Table As String * 4     'Location of Type M Path Table
    Location_TypeM_Opt_Oath_Table As String * 4 'Location of optional Type M Path Table
    Dir_Root_Record As RootDirRecord            'Directory Record for Root Dir

    Vol_Set_ID As String * 128                  'Volume Set Identifier
    Publisher_ID As String * 128                'Publisher Identifier
    Data_Preparer_ID As String * 128            'Data Preparer Identifier
    App_ID As String * 128                      'Application Identifie

    Copyright_File_ID As String * 37            'Copyright File Identifier
    Abstract_File_ID As String * 37             'Abstract File Identifier
    Bibliographic_File_ID As String * 37        'Bibliographic File Idenfitier

    Vol_Creation_Date As String * 17            'Volume Creation Date & Time
    Vol_Modification_Date As String * 17        'Volume Modification Date & Time
    Vol_Expiration_Date As String * 17          'Volume Expiration Date & Time
    Vol_Effective_Date As String * 17           'Volume Effective Date & Time

    File_Structure_Ver As Byte                  'File Structure Version (1)
    rsvd As Byte                                'reserved
    app_use(1 To 512) As Byte                   'reserved for application use
    rsvd2(1 To 653) As Byte                     'reserved
End Type

Private Const a_chars = "!""%&'()*+,-./0123456789:;<>=?ABCDEFGHIJKLMNOPQRSTUVWXYZ_ "
Private Const d_chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_.-"

Public Event FileProgress(ByVal max As Long, ByVal value As Long)
Public Event FSProgress(ByVal max As Long, ByVal value As Long)

Dim dirs As Long
Const MFPB As Long = 30                         'max files per block
Dim IniMap() As String
Dim bDOSNames As Boolean
Dim bCachedSections As Boolean

'get the size of a file in 2048-bytes-blocks
Private Function GetCDFSSize(length As Long) As Long
    Dim i As Integer, j As Long
    While Not i = -1
        j = j + 1
        If j * 2048 >= length Then i = -1
    Wend
    GetCDFSSize = j * 2048
End Function

Private Function StripNulls(ByVal OriginalStr As String) As String
    If (InStr(OriginalStr, Chr(0)) > 0) Then
        OriginalStr = Left(OriginalStr, InStr(OriginalStr, Chr(0)) - 1)
    End If
    StripNulls = OriginalStr
End Function

Private Function FindFilesAPI(Path As String, SearchStr As String, Optional rec As Boolean)
    'KPD-Team 1999
    'E-Mail: KPDTeam@Allapi.net
    'URL: http://www.allapi.net/

    Dim FileName As String ' Walking filename variable...
    Dim DirName As String ' SubDirectory Name
    Dim dirNames() As String ' Buffer for directory name entries
    Dim nDir As Integer ' Number of directories in this path
    Dim i As Integer ' For-loop counter...
    Dim hSearch As Long ' Search Handle
    Dim WFD As WIN32_FIND_DATA
    Dim Cont As Integer
    If Right(Path, 1) <> "\" Then Path = Path & "\"
    ' Search for subdirectories.
    nDir = 0
    ReDim dirNames(nDir)
    Cont = True
    hSearch = FindFirstFile(Path & "*", WFD)
    If hSearch <> INVALID_HANDLE_VALUE Then
        Do While Cont
        DirName = StripNulls(WFD.cFileName)
        ' Ignore the current and encompassing directories.
        If (DirName <> ".") And (DirName <> "..") Then
            ' Check for directory with bitwise comparison.
            If GetFileAttributes(Path & DirName) And FILE_ATTRIBUTE_DIRECTORY Then
                dirNames(nDir) = DirName
                nDir = nDir + 1
                ReDim Preserve files(UBound(files) + 1)
                files(UBound(files)) = AddSlash(Path) & DirName
                ReDim Preserve dirNames(nDir)
            End If
        End If
        Cont = FindNextFile(hSearch, WFD) 'Get next subdirectory.
        Loop
        Cont = FindClose(hSearch)
    End If
    ' Walk through this directory and sum file sizes.
    hSearch = FindFirstFile(Path & SearchStr, WFD)
    Cont = True
    If hSearch <> INVALID_HANDLE_VALUE Then
        While Cont
            FileName = StripNulls(WFD.cFileName)
            If (FileName <> ".") And (FileName <> "..") Then
                FindFilesAPI = FindFilesAPI + (WFD.nFileSizeHigh * MAXDWORD) + WFD.nFileSizeLow
                ReDim Preserve files(UBound(files) + 1)
                files(UBound(files)) = Path & FileName
            End If
            Cont = FindNextFile(hSearch, WFD) ' Get next file
        Wend
        Cont = FindClose(hSearch)
    End If
    ' If there are sub-directories...
    If rec Then
        If nDir > 0 Then
            ' Recursively walk into them...
            For i = 0 To nDir - 1
                FindFilesAPI = FindFilesAPI + FindFilesAPI(Path & dirNames(i) & "\", SearchStr, rec)
            Next i
        End If
    End If
End Function

'return the current project's filename
Public Property Get CurrProject() As String
    CurrProject = sCurrPrj
End Property

'return the current volume ID
Public Property Get VolumeID() As String
    VolumeID = sVolID
End Property

'set the current volume ID
Public Property Let VolumeID(ByVal sVal As String)
    If sCurrPrj = vbNullString Then Exit Property
    sVolID = sVal
    IniWriteString "info", "VolID", sVal
End Property

Private Function RemBreaks(ByVal sVal As String) As String
    Dim i As Long
    Dim sBuf As String

    For i = 1 To Len(sVal)
        If Not Mid$(sVal, i, 1) = vbCr Then sBuf = sBuf & Mid$(sVal, i, 1)
    Next

    For i = 1 To Len(sBuf)
        If Not Mid$(sBuf, i, 1) = vbLf Then RemBreaks = RemBreaks & Mid$(sBuf, i, 1)
    Next
End Function

'Reads a INI value
'rewritten because of the 64k API limit
Private Function IniRaw(ByVal section As String, ByVal title As String, _
                        ByVal default As String, ByVal file As String) As String

    On Error Resume Next

    Dim sFile As String, sLines() As String
    Dim i As Long, j As Long, FF As Integer: FF = FreeFile
    Dim FilePos As Long

    For i = 0 To UBound(IniMap)
        If IniMap(i, 0) = section Then
            FilePos = IniMap(i, 1)
            Exit For
        End If
    Next

    Open file For Binary As #FF
        sFile = Space(FileLen(file) - FilePos)
        Seek #FF, FilePos
        Get #FF, , sFile
    Close #FF

    sLines = Split(sFile, vbCrLf)
    For i = 0 To UBound(sLines)

        If Len(sLines(i)) > 0 Then
            If Left$(sLines(i), 1) = "[" Then
                If Right$(sLines(i), 1) = "]" Then
                    If Mid$(sLines(i), 2, Len(sLines(i)) - 2) = section Then

                        For j = i + 1 To UBound(sLines)

                            If Left$(sLines(j), 1) = "[" Then
                                If Right$(sLines(j), 1) = "]" Then
                                    IniRaw = default
                                    Exit Function
                                End If
                            End If

                            If Left$(sLines(j), InStr(1, sLines(j), "=") - 1) = title Then
                                IniRaw = RemBreaks(Trim$(Mid$(sLines(j), InStr(1, sLines(j), "=") + 1)))
                                Exit Function
                            End If

                        Next

                    End If
                End If
            End If
        End If

    Next
End Function

'Delete a section from an INI file
Private Function IniDeleteSection(ByVal sSection As String)
    Dim sCnt As String, sLines() As String
    Dim FF As Integer: FF = FreeFile
    Dim i As Long, j As Long, k As Long

    Open sCurrPrj For Binary As #FF
        sCnt = Space(FileLen(sCurrPrj))
        Get #FF, , sCnt
    Close #FF

    sLines = Split(sCnt, vbCrLf)

    Open sCurrPrj For Output As #FF
        For i = 0 To UBound(sLines)
            If Not Len(sLines(i)) = 0 Then
                If Left$(sLines(i), 1) = "[" Then
                    If Right$(sLines(i), 1) = "]" Then
                        If Mid$(sLines(i), 2, Len(sLines(i)) - 2) = sSection Then
                            For j = i + 1 To UBound(sLines)
                                If Left$(sLines(j), 1) = "[" Then
                                    If Right(sLines(j), 1) = "]" Then
                                        For k = j To UBound(sLines)
                                            If Not Len(sLines(i)) = 0 Then
                                                Print #FF, sLines(k)
                                            End If
                                        Next
                                        Close #FF
                                        Exit Function
                                    End If
                                End If
                            Next
                        Else
                            Print #FF, sLines(i)
                        End If
                    Else
                        Print #FF, sLines(i)
                    End If
                Else
                    Print #FF, sLines(i)
                End If
            End If
        Next
    Close #FF
End Function

'delete a whole directory from the INI file
Public Sub DeleteDir(ByVal sISODir As String)
    Dim sections() As String
    Dim i As Long
    sections = IniGetSections

    For i = 0 To UBound(sections)
        If Left$(sections(i), Len(sISODir)) = sISODir Then
            If val(IniGetString(sections(i), "type", TYPE_DELETED)) = TYPE_DIR Then
                dirs = dirs - 1
            End If
            IniDeleteSection sections(i)
        End If
    Next

    IniWriteString GetUpperDir(sISODir), "files", _
                   IniGetString(GetUpperDir(sISODir), "files", 1) - 1

    bCachedSections = False
End Sub

'delete a file from the INI file
Public Sub DeleteFile(ByVal sISOFile As String)
    IniDeleteSection sISOFile
    bCachedSections = False
    IniWriteString GetUpperDir(sISOFile), "files", _
                   IniGetString(GetUpperDir(sISOFile), "files", 1) - 1
End Sub

'read a value from the project file
'the API can only read 64kb per file -> rewritten
Private Function IniGetString(ByVal section As String, ByVal title As String, _
                              ByVal default As String) As String

    IniGetString = IniRaw(section, title, default, sCurrPrj)
End Function

'write a value to the project file
Private Sub IniWriteString(ByVal section As String, ByVal title As String, _
                           ByVal sValue As String)

    WritePrivateProfileString section, title, sValue, sCurrPrj
End Sub

'get the names of all sections in the project file
'the API can only read up to 64kb per file -> rewritten
Private Function IniGetSections() As Variant
    Dim sFile As String, sLines() As String
    Static sApps() As String, OldLen As Long
    Dim i As Long, FF As Integer: FF = FreeFile
    Dim fcnt As Long

'    If OldLen = FileLen(sCurrPrj) Then
'        'nothing changed in the project, return a cached result
'        IniGetSections = sApps
'        Exit Function
'    End If
'
'    OldLen = FileLen(sCurrPrj)

    If bCachedSections Then
        IniGetSections = sApps
        Exit Function
    End If

    Open sCurrPrj For Binary As #FF
        sFile = Space(FileLen(sCurrPrj))
        Get #FF, , sFile
    Close #FF

    sLines = Split(sFile, vbCrLf)
    ReDim IniMap(UBound(sLines), 1)
    For i = 0 To UBound(sLines)
        If Len(sLines(i)) > 0 Then

            If Left(sLines(i), 1) = "[" Then
                If Right(sLines(i), 1) = "]" Then

                    'cache the results
                    IniMap(i, 0) = Mid(sLines(i), 2, Len(sLines(i)) - 2)
                    IniMap(i, 1) = InStr(1, sFile, sLines(i))

                    ReDim Preserve sApps(fcnt)
                    sApps(UBound(sApps)) = Mid(sLines(i), 2, Len(sLines(i)) - 2)
                    fcnt = fcnt + 1

                End If
            End If

        End If
    Next

    'return the sections
    IniGetSections = sApps
    bCachedSections = True
End Function

'create a new project (if file exists, will be overwritten)
Public Sub NewProject(ByVal sFile As String)
    'set some vars
    ReDim IniMap(0)
    bCachedSections = False
    dirs = 0
    sVolID = "NEW_IMAGE"
    sCurrPrj = sFile
    CreateFile sFile

    'write the version info the project
    IniWriteString "info", "ver", "cISO9660"
    IniWriteString "info", "type", TYPE_DELETED
    IniWriteString "info", "VolID", "NEW_IMAGE"

    'create the root directory
    NewDir "\"
End Sub

'open a project
Public Function OpenProject(ByVal sFile As String) As Boolean
    Dim sVal As String
    Dim lResult As Long

    If Not FileExists(sFile) Then Exit Function

    'lResult = GetPrivateProfileString("info", "ver", "no", sVal, 256, sFile)
    If IniRaw("info", "ver", "no", sFile) = "cISO9660" Then
        dirs = 0
        sCurrPrj = sFile
        sVolID = IniGetString("info", "VolID", "NEW_IMAGE")
        OpenProject = True
    End If
End Function

'moves the project to another location
Public Function MoveProject(sMoveTo As String) As Boolean
    On Error Resume Next
    FileCopy sCurrPrj, sMoveTo
    If FileExists(sMoveTo) Then
        Kill sCurrPrj
        sCurrPrj = sMoveTo
        MoveProject = True
    End If
End Function

'delete the current project
Public Sub DeleteProject()
    On Error Resume Next
    If sCurrPrj <> vbNullString Then
        Kill sCurrPrj
    End If
    sCurrPrj = vbNullString
End Sub

'create a file
Private Sub CreateFile(ByVal sFile As String)
    Dim FF As Integer: FF = FreeFile
    Open sFile For Output As #FF: Close #FF
End Sub

'create a new directory
Public Sub NewDir(ByVal sISODir As String)
    If sCurrPrj = "" Then Exit Sub

    sISODir = AddSlash(sISODir)
    If val(IniGetString(sISODir, "nr", "0")) > 0 Then Exit Sub

    IniWriteString sISODir, "type", TYPE_DIR
    IniWriteString sISODir, "files", "1"
    dirs = dirs + 1
    IniWriteString sISODir, "nr", dirs
    If Not sISODir = "\" Then
        IniWriteString GetUpperDir(sISODir), "files", _
                       IniGetString(GetUpperDir(sISODir), "files", 0) + 1
    End If
End Sub

'recursively search a folder and add its contents to a directory in the project
Public Sub AddDir(ByVal sISODir As String, ByVal sLocalDir As String, Optional ByVal subdirs As Boolean = True)
    Dim i As Long

    ReDim files(0)
    FindFilesAPI sLocalDir, "*.*", subdirs

    If subdirs Then
        For i = 0 To UBound(files)
            If Not files(i) = "" Then
                If DirExists(files(i)) Then
                    If isDir(files(i)) Then
                        DoEvents
                        NewDir AddSlash(sISODir) & Mid(files(i), Len(sLocalDir) + 1)
                    End If
                End If
            End If
        Next
    End If

    For i = 0 To UBound(files)
        If Not files(i) = "" Then
            If FileExists(files(i)) Then
                DoEvents
                AddFile AddSlash(sISODir) & Mid(GetPath(files(i)), Len(sLocalDir) + 1), files(i)
            End If
        End If
    Next

    bCachedSections = False
End Sub

'checks if a dir as subdirs
Public Function HasSubDirs(ByVal sISODir As String) As Boolean
    Dim i As Long
    Dim sections() As String

    sections = IniGetSections

    For i = 0 To UBound(sections)
        If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_DIR Then
            If Not sISODir = sections(i) Then
                If IsSubDir(sISODir, sections(i)) Then
                    HasSubDirs = True
                    Exit Function
                End If
            End If
        End If
    Next
End Function

'checks if a dir has subfiles
Public Function HasSubFiles(ByVal sISODir As String) As Boolean
    Dim i As Long
    Dim sections() As String
    Dim filecnt As Long

    sections = IniGetSections

    For i = 0 To UBound(sections)
        If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then
            If IsSubFile(sISODir, sections(i)) Then
                HasSubFiles = True
                Exit Function
            End If
        End If
    Next
End Function

'return the files from a directory
Public Function GetSubFiles(ByVal sISODir As String) As Variant
    Dim i As Long
    Dim sections() As String
    Dim ret() As String
    Dim retcnt As Long
    Dim filecnt As Long
    Dim IniRet As Long

    sections = IniGetSections

    filecnt = IniGetString(sISODir, "files", 1)

    For i = 0 To UBound(sections)
        If Left$(sections(i), Len(sISODir)) = sISODir Then
            IniRet = IniGetString(sections(i), "type", TYPE_DELETED)
            If IniRet = TYPE_FILE Then
                If IsSubFile(sISODir, sections(i)) Then
                    DoEvents
                    ReDim Preserve ret(retcnt)
                    ret(retcnt) = sections(i)
                    retcnt = retcnt + 1
                    If retcnt + 1 = filecnt Then
                        GetSubFiles = ret
                        Exit Function
                    End If
                End If
            ElseIf IniRet = TYPE_DIR Then
                If Not sISODir = sections(i) Then
                    If IsSubDir(sISODir, sections(i)) Then
                        filecnt = filecnt - 1
                    End If
                End If
            End If
        End If
    Next

    GetSubFiles = ret
End Function

'return the sub-directories from a directory
Public Function GetSubDirs(ByVal sISODir As String) As Variant
    On Error Resume Next

    Dim i As Long
    Dim sections() As String
    Dim ret() As String
    Dim retcnt As Long
    Dim filecnt As Long
    Dim IniRet As Long

    sections = IniGetSections

    filecnt = IniGetString(sISODir, "files", 1)

    For i = 0 To UBound(sections)
        If Left$(sections(i), Len(sISODir)) = sISODir Then
            IniRet = IniGetString(sections(i), "type", TYPE_DELETED)
            
            If IniRet = TYPE_DIR Then
                If IsSubDir(sISODir, sections(i)) Then
                    DoEvents
                    ReDim Preserve ret(retcnt)
                    ret(retcnt) = sections(i)
                    retcnt = retcnt + 1
                    If retcnt = filecnt Then
                        GetSubDirs = ret
                        Exit Function
                    End If
                End If
            ElseIf IniRet = TYPE_FILE Then
                filecnt = filecnt - 1
            End If

        End If
    Next

    GetSubDirs = ret
End Function

'add a file to a directory
Public Sub AddFile(ByVal sISODir As String, ByVal sLocalFile As String)
    Dim fcnt As Long
    If sCurrPrj = "" Then Exit Sub
    sISODir = AddSlash(sISODir)
    IniWriteString sISODir & CutFile(GetFilename(sLocalFile)), "type", TYPE_FILE
    IniWriteString sISODir & CutFile(GetFilename(sLocalFile)), "local", sLocalFile

    fcnt = IniGetString(sISODir, "files", 0): fcnt = fcnt + 1
    IniWriteString sISODir, "files", fcnt

    bCachedSections = False
End Sub

'pull out D-Chars
Private Function To_D_Chars(ByVal str As String) As String
    Dim sBuf As String, i As Integer

    str = UCase(str)

    For i = 1 To Len(str)
        If InStr(1, d_chars, Mid(str, i, 1)) > 0 Then sBuf = sBuf & Mid(str, i, 1)
    Next

    To_D_Chars = sBuf
End Function

'pull out A-Chars
Private Function To_A_Chars(ByVal str As String) As String
    Dim sBuf As String, i As Integer

    str = UCase(str)

    For i = 1 To Len(str)
        If InStr(1, a_chars, Mid(str, i, 1)) > 0 Then sBuf = sBuf & Mid(str, i, 1)
    Next

    To_A_Chars = sBuf
End Function

'convert a long to a 16 bit least significant byte value
Private Function To16BitLS(l As Integer) As String
    Dim b0 As Byte, b1 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    To16BitLS = Chr(b1) & Chr(b0)
End Function

'convert a long to a 16 bit most significant byte value
Private Function To16BitMS(l As Integer) As String
    Dim b0 As Byte, b1 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    To16BitMS = Chr(b0) & Chr(b1)
End Function

'convert a long to a 16 bit both byte ordered value
Private Function To16BitBothByte(l As Long) As String
    Dim b0 As Byte, b1 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    To16BitBothByte = Chr(b0) & Chr(b1) & Chr(b1) & Chr(b0)
End Function

'convert a long to a 32 bit most significant byte value
Private Function To32BitMS(l As Long) As String
    Dim b0 As Byte, b1 As Byte, b2 As Byte, b3 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    b2 = (l And &HFF0000) \ &H10000
    b3 = (l And &H7F000000) \ &H1000000 - 128 * (l < 0)
    To32BitMS = Chr(b0) & Chr(b1) & Chr(b2) & Chr(b3)
End Function

'convert a long to a 32 bit least significant byte value
Private Function To32BitLS(l As Long) As String
    Dim b0 As Byte, b1 As Byte, b2 As Byte, b3 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    b2 = (l And &HFF0000) \ &H10000
    b3 = (l And &H7F000000) \ &H1000000 - 128 * (l < 0)
    To32BitLS = Chr(b3) & Chr(b2) & Chr(b1) & Chr(b0)
End Function

'convert a long to a 32 bit both byte ordered value
Private Function To32BitBothByte(l As Long) As String
    Dim b0 As Byte, b1 As Byte, b2 As Byte, b3 As Byte
    b0 = (l And &HFF&)
    b1 = (l And &HFF00&) \ &H100&
    b2 = (l And &HFF0000) \ &H10000
    b3 = (l And &H7F000000) \ &H1000000 - 128 * (l < 0)
    To32BitBothByte = Chr(b0) & Chr(b1) & Chr(b2) & Chr(b3) & _
                      Chr(b3) & Chr(b2) & Chr(b1) & Chr(b0)
End Function

'round a value up, if needed
Private Function RoundUp(ByVal nr As Currency) As Long
    If nr \ 1 <> nr Then
        If CLng(nr) < nr Then nr = CLng(nr) + 1
    End If
    RoundUp = nr
End Function

'create the start- and end-offsets for the directories
Private Function CreateDirStarts() As Long
    Dim entries() As String
    Dim i As Long, start As Long, fcnt As Long

    start = 19

    entries = IniGetSections

    For i = 0 To UBound(entries)
        DoEvents
        If Not Trim$(entries(i)) = vbNullString Then
            If IniGetString(entries(i), "type", TYPE_DELETED) = TYPE_DIR Then
                fcnt = IniGetString(entries(i), "files", 0)
                IniWriteString entries(i), "blocks", RoundUp(fcnt / MFPB)
                IniWriteString entries(i), "start", start
                start = start + RoundUp(fcnt / MFPB)
            End If
        End If
    Next

    CreateDirStarts = start
End Function

'create the start- and end-offsets for the files
Private Sub CreateFileStarts(ByVal startblock As Long)
    Dim entries() As String
    Dim i As Long, p As Long

    entries = IniGetSections

    For i = 0 To UBound(entries)
        DoEvents
        If Not Len(Trim$(entries(i))) = 0 Then
            If IniGetString(entries(i), "type", TYPE_DELETED) = TYPE_FILE Then
                IniWriteString entries(i), "start", startblock
                IniWriteString entries(i), "length", FileLen(IniGetString(entries(i), "local", ""))
                startblock = startblock + (GetCDFSSize(FileLen(IniGetString(entries(i), "local", ""))) / 2048)
            End If
        End If
    Next
End Sub

'get the size of a file from the project file
Private Function GetFileBlocks() As Long
    Dim i As Long, cnt As Long
    Dim entries() As String

    entries = IniGetSections

    For i = 0 To UBound(entries)
        If Not Trim$(entries(i)) = vbNullString Then
            If IniGetString(entries(i), "type", TYPE_DELETED) = TYPE_FILE Then
                cnt = cnt + (FileLen(IniGetString(entries(i), "local", "")))
            End If
        End If
    Next

    GetFileBlocks = RoundUp(cnt / 2048)
End Function

'pad a string to the left or right
Private Function Pad(ByVal str As String, ByVal val As Long, ByVal Dir As Integer, ByVal char As String)
    Dim i As Long
    If Len(str) < val Then
        For i = 0 To val - Len(str)
            If Dir = 0 Then
                str = char & str
            Else
                str = str & char
            End If
        Next
    End If
    Pad = str
End Function

'create date and time for the primary volume descriptor according to ECMA-119
Private Function CreateDateTime() As String
    CreateDateTime = format(Year(Date), "0000") & format(Month(Date), "00") & format(Day(Date), "00") & _
                     format(Hour(Time), "00") & format(Minute(Time), "00") & format(Second(Time), "00") & _
                     "00" & Chr(0)
End Function

'create date and time for a directory record according to ECMA-119
Private Function CreateDateTimeRec() As String
    CreateDateTimeRec = Chr(Year(Date) - 1900) & Chr(Month(Date)) & Chr(Day(Date)) & Chr(Hour(Time)) & _
                        Chr(Minute(Time)) & Chr(Second(Time)) & Chr(0)
End Function

'write a directory record
Private Function WriteDirRecord(ByVal sDir As String, ByVal fh As Long) As Long
    Dim tdir As DirRecord, rec As DirRecord
    Dim UpperDir As String, sFile As String
    Dim StartLBA As Long, upperstartlba As Long
    Dim Size As Long, uppersize As Long, dirsize As Long, dirstart As Long
    Dim sections() As String, Rest() As Byte
    Dim i As Long, lWritten As Long, fcnt As Long
    Dim totalcnt As Long: totalcnt = 1

    'get the INI sections (files & dirs)
    sections = IniGetSections

    'get the parent directory from the current directory (sDir)
    UpperDir = GetUpperDir(sDir)
    'get the calculated start LBA of the dir
    StartLBA = IniGetString(sDir, "start", 0)
    dirstart = IniGetString(sDir, "start", 0)
    'get the calculated start LBA of the parent dir
    upperstartlba = IniGetString(UpperDir, "start", 0)
    uppersize = IniGetString(UpperDir, "blocks", 1)
    'get the size in sectors of the directory
    Size = IniGetString(sDir, "blocks", 1)
    dirsize = IniGetString(sDir, "blocks", 1)

    'this directory directory is like "."
    With tdir
        .length = 34
        .LocExt = To32BitBothByte(StartLBA)
        .DataLen = To32BitBothByte(Size * 2048)
        .RecDateTime = CreateDateTimeRec
        .VolSeqNum = To16BitBothByte(1)
        .FileFlags = FLAG_DIRECTORY Or FLAG_HIDDEN
        .FileIDLen = 1
    End With

    Put #fh, , tdir: Put #fh, , CByte(0)
    lWritten = lWritten + tdir.length

    'change 2 values, so the directory record becomes ".."
    tdir.LocExt = To32BitBothByte(upperstartlba)
    tdir.DataLen = To32BitBothByte(uppersize)
    Put #fh, , tdir: Put #fh, , CByte(1)
    lWritten = lWritten + tdir.length

    For i = 0 To UBound(sections)
        DoEvents
        If Not Trim$(sections(i)) = vbNullString Then
            If Not sections(i) = sDir Then

                'add a directory to the record
                If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_DIR Then

                    'sub-directory, so we can add it?
                    If IsSubDir(sDir, sections(i)) Then

                        'max. records per block = 30,
                        'if this record is 31, fill up the sector with zeros
                        If totalcnt = MFPB Then
                            FillUpSector fh
                            totalcnt = 0
                        Else
                            totalcnt = totalcnt + 1
                        End If

                        StartLBA = IniGetString(sections(i), "start", 0)
                        Size = IniGetString(sections(i), "blocks", 1)

                        'create the directory record
                        With rec
                            .length = 34 + Len(Left$(GetDirname(sections(i)), IIf(bDOSNames, 8, 31)))
                            .LocExt = To32BitBothByte(StartLBA)
                            .DataLen = To32BitBothByte(Size * 2048)
                            .RecDateTime = CreateDateTimeRec
                            .VolSeqNum = To16BitBothByte(1)
                            .FileFlags = FLAG_DIRECTORY
                            .FileIDLen = Len(Left$(GetDirname(sections(i)), IIf(bDOSNames, 8, 31)))
                        End With

                        'and write it
                        Put #fh, , rec
                        Put #fh, , Left$(GetDirname(sections(i)), IIf(bDOSNames, 8, 31))
                        Put #fh, , CByte(0)

                        lWritten = lWritten + rec.length

                        WriteDirRecord = WriteDirRecord + 1

                    End If

                'add a file to the record
                ElseIf IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then

                    'sub-file, so we can add it?
                    If IsSubFile(sDir, sections(i)) Then

                        'max. records per block = 30,
                        'if this record is 31, fill up the sector with zeros
                        If totalcnt = MFPB Then
                            FillUpSector fh
                            totalcnt = 0
                        Else
                            totalcnt = totalcnt + 1
                        End If

                        StartLBA = IniGetString(sections(i), "start", 0)
                        Size = IniGetString(sections(i), "length", 0)

                        'create the file record
                        With rec
                            '.Length = 34 + Len(CutFile(GetFilename(sections(i)))) + 2
                            .length = 34 + Len(GetFilename(sections(i))) + 2
                            .LocExt = To32BitBothByte(StartLBA)
                            .DataLen = To32BitBothByte(Size)
                            .RecDateTime = CreateDateTimeRec
                            .VolSeqNum = To16BitBothByte(1)
                            .FileFlags = 0
                            '.FileIDLen = Len(CutFile(GetFilename(sections(i))) & ";1")
                            .FileIDLen = Len(GetFilename(sections(i)) & ";1")
                        End With

                        'write the record
                        Put #fh, , rec
                        'Put #fh, , CutFile(GetFilename(sections(i))) & ";1"
                        Put #fh, , GetFilename(sections(i)) & ";1"
                        Put #fh, , CByte(0)

                        lWritten = lWritten + rec.length

                        WriteDirRecord = WriteDirRecord + 1

                    End If

                End If
            End If
        End If
    Next

    'fill the rest of the last written sector with zeros
    FillUpSector fh

    'check for non-written sectors
    If Seek(fh) - ((dirstart * 2048) + (dirsize * 2048)) - 1 <> 0 Then
        Put #fh, , CStr(String(Abs(Seek(fh) - ((dirstart * 2048) + (dirsize * 2048)) - 1), Chr(0)))
    End If
End Function

Private Function CutDir(ByVal sName As String) As String
    On Error GoTo ErrHndl
    Dim MaxLen As Byte

    If bDOSNames = True Then
        MaxLen = 8    ' ISO Level-1
    Else
        MaxLen = 31   ' ISO Level-2
    End If

    'sName = To_D_Chars(sName)
    If Len(sName) > MaxLen Then sName = Left$(sName, MaxLen - 2) & "~1"
    CutDir = sName

ErrHndl:
    CutDir = UCase(sName)
End Function

'cut a filename, if length > 31,
'and remove non-D-characters
Private Function CutFile(ByVal sName As String) As String
    On Error GoTo ErrHndl
    Dim MaxLen As Byte

    If bDOSNames = True Then
        MaxLen = 8    ' ISO Level-1
    Else
        MaxLen = 31   ' ISO Level-2
    End If

    'If Left$(Right$(sName, 3), 1) = "." Then sName = sName & " "
    If InStrRev(sName, ".") > Len(sName) - 3 Then
        sName = sName & Space(Len(sName) - InStrRev(sName, ".") + 1)
    End If
    sName = To_D_Chars(UCase(Left(sName, InStrRev(sName, ".") - 1))) & "." & Right(sName, 3)
    If InStr(1, sName, ".") > MaxLen - 4 Then
        CutFile = UCase(Left(Left(sName, InStrRev(sName, ".") - 1), MaxLen - 2) & "~1." & Right(sName, 3))
    Else
        CutFile = UCase(sName)
    End If
    
    Exit Function

ErrHndl:
    CutFile = UCase(sName)
End Function

'directory parent for the file?
Private Function IsSubFile(ByVal sDir1 As String, ByVal sDir2 As String) As Boolean
    If sDir1 = GetPath(sDir2) Then IsSubFile = True
End Function

'directory parent for the directory?
Private Function IsSubDir(ByVal sDir1 As String, ByVal sDir2 As String) As Boolean
    If GetUpperDir(sDir2) = sDir1 Then IsSubDir = True
End Function

'write a file to the image
Private Sub WriteFile(ByVal section As String, ByVal fh As Long)
    Dim FF As Integer: FF = FreeFile
    Dim sFile As String, buffer() As Byte
    Dim LastPos As Long
    Dim Rest() As Byte

    LastPos = Seek(fh) - 1

    sFile = IniGetString(section, "local", "")
    Open sFile For Binary As #FF
        ReDim buffer(FileLen(sFile))
        Get #FF, , buffer
    Close #FF
    DoEvents
    Put #fh, , buffer
    buffer = vbNullString

    If GetCDFSSize(FileLen(sFile)) - (Seek(fh) - LastPos) > 0 Then
        ReDim Rest(GetCDFSSize(FileLen(sFile)) - (Seek(fh) - LastPos))
        Put #fh, , Rest
    Else
        Seek fh, Seek(fh) - 1
    End If
End Sub

'get the directory's parent number
Private Function GetLevel(sDir As String) As Byte
    GetLevel = IniGetString(GetUpperDir(sDir), "nr", 1)
End Function

'write the little-endian path table
Private Function WriteLEndianPTable(ByVal fh As Long) As Long
    Dim ptable As PTableRecord
    Dim sections() As String
    Dim buffer As String
    Dim i As Long, FirstPos As Long
    Dim Rest() As Byte

    ptable.length = 1
    ptable.FirstSector = To32BitMS(19)
    ptable.RecNumParentDir = To16BitMS(1)

    FirstPos = Seek(fh) - 1

    Put #fh, , ptable
    Put #fh, , CByte(0): Put #fh, , CByte(0)

    sections = IniGetSections

    For i = 0 To UBound(sections)
        DoEvents
        If Not Trim$(sections(i)) = vbNullString Then
            If Not sections(i) = "\" Then
                If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_DIR Then
                    ptable.length = Len(GetDirname(sections(i)))
                    ptable.FirstSector = To32BitMS(IniGetString(sections(i), "start", 0))
                    ptable.RecNumParentDir = To16BitMS(GetLevel(sections(i)))
                    Put #fh, , ptable
                    Put #fh, , CStr(CutDir(UCase(GetDirname(sections(i)))))
                    If Len(CutDir(UCase(GetDirname(sections(i))))) Mod 2 Then
                        Put #fh, , CByte(0)
                    End If
                End If
            End If
        End If
    Next

    WriteLEndianPTable = (Seek(fh) - FirstPos) - 1

    If 2048 - (Seek(fh) - FirstPos) > 0 Then
        ReDim Rest(2048 - (Seek(fh) - FirstPos))
        Put #fh, , Rest
    End If
End Function

Public Function ISOSize() As Long
    On Error Resume Next
    Dim sections() As String
    Dim i As Long
    sections = IniGetSections

    For i = 0 To UBound(sections)
        If Not sections(i) = vbNullString Then
            If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then
                ISOSize = ISOSize + FileLen(IniGetString(sections(i), "local", ""))
            End If
        End If
    Next
End Function

'write the big endian path table
Private Sub WriteBEndianPTable(ByVal fh As Long)
    Dim ptable As PTableRecord
    Dim sections() As String
    Dim buffer As String
    Dim i As Long, FirstPos As Long
    Dim Rest() As Byte

    ptable.length = 1
    ptable.FirstSector = To32BitLS(19)
    ptable.RecNumParentDir = To16BitLS(1)

    FirstPos = Seek(fh) - 1

    Put #fh, , ptable
    Put #fh, , CByte(0): Put #fh, , CByte(0)

    sections = IniGetSections

    For i = 0 To UBound(sections)
        DoEvents
        If Not Trim$(sections(i)) = vbNullString Then
            If Not sections(i) = "\" Then
                If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_DIR Then
                    ptable.length = Len(GetDirname(sections(i)))
                    ptable.FirstSector = To32BitLS(IniGetString(sections(i), "start", 0))
                    ptable.RecNumParentDir = To16BitLS(GetLevel(sections(i)))
                    Put #fh, , ptable
                    Put #fh, , CStr(CutDir(UCase(GetDirname(sections(i)))))
                    If Len(CutDir(UCase(GetDirname(sections(i))))) Mod 2 Then
                        Put #fh, , CByte(0)
                    End If
                End If
            End If
        End If
    Next

    If 2048 - (Seek(fh) - FirstPos) > 0 Then
        ReDim Rest(2048 - (Seek(fh) - FirstPos))
        Put #fh, , Rest
    End If
End Sub

Private Sub FillUpSector(fh As Long)
    If GetCDFSSize(Seek(fh)) > Seek(fh) Then
        Put #fh, , CStr(String(GetCDFSSize(Seek(fh)) - Seek(fh) + 1, Chr(0)))
    End If
End Sub

Public Function GetLocalFileList() As Variant
    Dim i As Long
    Dim sections() As String, files() As String
    Dim filecnt As Long

    sections = IniGetSections

    For i = 0 To UBound(sections)
        If Not Trim$(sections(i)) = vbNullString Then
            If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then
                ReDim Preserve files(filecnt)
                files(filecnt) = IniGetString(sections(i), "local", "")
                filecnt = filecnt + 1
            End If
        End If
    Next

    GetLocalFileList = files
End Function

'create the ISO9660 Image
Public Function CreateISO(sPath As String, Optional WriteFiles As Boolean = True, Optional Level1 As Boolean = False) As Boolean
    On Error GoTo ErrH

    Dim IsoHdr As ISO_HEADER
    Dim PrimDesc As ISO_PRIMARY_DESCRIPTOR
    Dim FF As Integer: FF = FreeFile
    Dim lPTable As Long, lBytesWritten As Long
    Dim i As Long, PVDPos As Long, BPTPos As Long
    Dim sections() As String
    Dim filecnt As Long, currfile As Long
    Dim WrittenRecs As Long

    If sCurrPrj = vbNullString Then Exit Function

    bDOSNames = Level1

    sections = IniGetSections

    RaiseEvent FSProgress(UBound(sections) + 80, 0)

    'generate the directory + file start offsets
    lPTable = CreateDirStarts
    RaiseEvent FSProgress(UBound(sections) + 80, 20)

    CreateFileStarts lPTable
    RaiseEvent FSProgress(UBound(sections) + 80, 40)

    If FileExists(sPath) Then Kill sPath

    Open sPath For Binary As #FF

        'prepare and write the ISO9660 header

        With IsoHdr
            .zeros = String(32766, Chr(0))
            .Vol_Desc.Std_ID = "CD001" & Chr(1)
            .Vol_Desc.Vol_Desc_Type = VOL_DESC_PRIM_VOL_DESCRIPTOR
        End With

        With PrimDesc
            .System_ID = To_A_Chars("WIN32")
            .Volume_ID = To_D_Chars(sVolID)

            .Vol_Space_Size = To32BitBothByte(lPTable + GetFileBlocks)
            .Vol_Set_Size = To16BitBothByte(1)
            .Vol_Seq_Number = To16BitBothByte(1)
            .Logical_Block_Size = To16BitBothByte(2048)
            .Location_TypeL_Path_Table = To32BitMS(17)
            .Location_TypeM_Path_Table = To32BitLS(18)

            .Abstract_File_ID = Space(37)
            .App_ID = Pad("cISO96660", 128, 1, " ")
            .Bibliographic_File_ID = Space(37)
            .Copyright_File_ID = Space(37)
            .Data_Preparer_ID = To_A_Chars(Pad("Data Preparer", 128, 1, " "))
            .Publisher_ID = To_A_Chars(Pad("Publisher", 128, 1, " "))
            .Vol_Set_ID = Space(128)
            .Path_Table_Size = To32BitBothByte(1)  'just a place holder, will be
                                                   'edited below

            .Vol_Creation_Date = CreateDateTime
            .Vol_Effective_Date = CreateDateTime
            .Vol_Expiration_Date = CreateDateTime
            .Vol_Modification_Date = CreateDateTime

            .Dir_Root_Record.length = 34
            .Dir_Root_Record.LocExt = To32BitBothByte(19)
            .Dir_Root_Record.DataLen = To32BitBothByte(IniGetString("\", "blocks", 0) * 2048)
            .Dir_Root_Record.RecDateTime = CreateDateTimeRec
            .Dir_Root_Record.VolSeqNum = To16BitBothByte(1)
            .Dir_Root_Record.FileFlags = FLAG_DIRECTORY
            .Dir_Root_Record.FileIDLen = 1

        End With

        Put #FF, , IsoHdr
        PVDPos = Seek(FF)
        Put #FF, , PrimDesc

        'create the little endian path table
        PrimDesc.Path_Table_Size = To32BitBothByte(WriteLEndianPTable(FF))
        RaiseEvent FSProgress(UBound(sections) + 80, 60)
        BPTPos = Seek(FF)

        Seek FF, PVDPos
        Put #FF, , PrimDesc

        'create the big endian path table
        Seek FF, BPTPos
        WriteBEndianPTable FF
        RaiseEvent FSProgress(UBound(sections) + 80, 80)

        'create the file table
        For i = 0 To UBound(sections)
            DoEvents
            If Not Trim$(sections(i)) = vbNullString Then
                If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_DIR Then
                    WrittenRecs = WrittenRecs + WriteDirRecord(sections(i), FF)
                    RaiseEvent FSProgress(UBound(sections) + 80, 80 + WrittenRecs)
                ElseIf IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then
                    filecnt = filecnt + 1
                End If
            End If
        Next

        If WriteFiles Then
            'write the files to the image
            For i = 0 To UBound(sections)
                If Not Trim$(sections(i)) = vbNullString Then
                    If IniGetString(sections(i), "type", TYPE_DELETED) = TYPE_FILE Then
                        WriteFile sections(i), FF
                        currfile = currfile + 1
                        RaiseEvent FileProgress(filecnt, currfile)
                    End If
                End If
            Next

            'generate some empty sectors
            'to make sure cBurnISO doesn't forget any important ones :D
            Put #FF, , CStr(String(2048 * 8, Chr(0)))
        End If

    Close #FF
    bDOSNames = False
    CreateISO = True
    Exit Function

ErrH:
    bDOSNames = False
    Close #FF
End Function

Private Sub Class_Initialize()
    ReDim IniMap(0)
End Sub
