Attribute VB_Name = "modFunctions"
Option Explicit

'Windows function to get the type of a specific drive
Public Declare Function GetDriveType Lib "kernel32" Alias "GetDriveTypeA" ( _
    ByVal nDrive As String _
) As Long

'Sleep function
'Public Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

Public Const VER_PLATFORM_WIN32s = 0            'Windows 3.x
Public Const VER_PLATFORM_WIN32_WINDOWS = 1     'Windows 9x/Me
Public Const VER_PLATFORM_WIN32_NT = 2          'Windows NT 3/4/2000/XP/Longhorn

Public Declare Sub SleepA Lib "kernel32" Alias "Sleep" (ByVal dwMilliseconds As Long)

'windows function to get the windows version
Public Declare Function GetVersionEx Lib "kernel32" Alias "GetVersionExA" ( _
        lpVersionInformation As OSVERSIONINFOEX _
) As Long

'windows function to get the system directory
Private Declare Function GetSystemDirectory Lib "kernel32" Alias "GetSystemDirectoryA" ( _
    ByVal lpBuffer As String, ByVal nSize As Long _
) As Long

Public CurrPrj As Form

Public WrtDrvID As String
Public bEjectAfterWrite As Boolean
Public bUseTestMode As Boolean
Public bUseBURNProof As Boolean
Public bFinalizeDisc As Boolean

Public Function GetSysDir() As String
    Dim buf As String * 256
    Dim lRet As Long

    lRet = GetSystemDirectory(buf, Len(buf))
    GetSysDir = Left$(buf, lRet)
End Function

'Found @ http://www.pscode.com/vb/scripts/ShowCode.asp?txtCodeId=7331&lngWId=1
Public Function FormatFileSize(ByVal dblFileSize As Double, _
    Optional ByVal strFormatMask As String) As String

    Select Case dblFileSize
        Case 0 To 1023 ' Bytes
            FormatFileSize = format(dblFileSize) & " bytes"
        Case 1024 To 1048575 ' KB
            If strFormatMask = Empty Then strFormatMask = "###0"
            FormatFileSize = format(dblFileSize \ 1024, strFormatMask) & " KB"
        Case 1024# ^ 2 To 1073741823 ' MB
            If strFormatMask = Empty Then strFormatMask = "####0.00"
            FormatFileSize = format(dblFileSize \ (1024 ^ 2), strFormatMask) & " MB"
        Case Is > 1073741823# ' GB
            If strFormatMask = Empty Then strFormatMask = "####0.00"
            FormatFileSize = format(dblFileSize \ (1024 ^ 3), strFormatMask) & " GB"
    End Select

End Function

'secs to mins:secs
Public Function FormatTime(ByVal secs As Long) As String
    Dim mins As Long
    If secs > 60 Then
        mins = Fix(secs / 60)
        secs = secs - (mins * 60)
        FormatTime = format(mins, "00") & ":" & format(secs, "00") & " min"
    Else
        FormatTime = "00:" & format(secs, "00") & " min"
    End If
End Function

'seconds to minutes:seconds
Public Function FormatSecs(ByVal iSecs As Integer) As String
    Dim iMin As Integer, iSec As Integer
    iMin = 0: iSec = 0

    If iSecs > 60 Then
        iMin = Fix(iSecs / 60)
        iSec = iSecs - iMin * 60
    Else
        iSec = iSecs
    End If

    FormatSecs = format(iMin, "00") & ":" & format(iSec, "00")
End Function

'Checks if a wav has the typical CDDA format 44.1kHz, 16-bit, stereo
Public Function CheckWav(ByVal sFile As String) As Boolean
    On Error GoTo ErrH
    Dim wHdr As t_WavHdr
    Dim FF As Integer: FF = FreeFile

    Open sFile For Binary As #FF
        'read the wave header
        Get #FF, , wHdr
        
        'Check the format
        CheckWav = (wHdr.BitsPerSample = 16) _
                 = (wHdr.NumChannels = 2) _
                 = (wHdr.SampleRate = 44100)

    Close #FF

    Exit Function

ErrH:
End Function

Public Function GetCDDASecs(ByVal sFile As String) As Integer
    '176400 b/s for CDDA (16 bit, 44.1kHz, stereo)
    GetCDDASecs = Fix((FileLen(sFile) - 44) \ 176400)
End Function

'remove 0-chars from a string
Public Function RemNulls(ByVal sVal As String) As String
    Dim i As Integer
    Dim sBuf As String
    For i = 1 To Len(sVal)
        If Asc(Mid(sVal, i, 1)) > 0 Then
            sBuf = sBuf & Mid(sVal, i, 1)
        End If
    Next
    RemNulls = sBuf
End Function

'get the parent for a directory
Public Function GetUpperDir(ByVal sDir As String)
    Dim parts() As String, i As Integer
    sDir = AddSlash(sDir)
    parts = Split(sDir, "\")
    For i = 0 To UBound(parts) - 2
        GetUpperDir = GetUpperDir & parts(i) & "\"
    Next
    GetUpperDir = AddSlash(GetUpperDir)
End Function

'get the filename from path + filename
Public Function GetFilename(ByVal sVal As String) As String
    Dim parts() As String
    parts = Split(sVal, "\")
    GetFilename = parts(UBound(parts))
End Function

'get the path name from path + filename
Public Function GetDirname(ByVal sDir As String) As String
    Dim parts() As String
    parts = Split(AddSlash(sDir), "\")
    GetDirname = parts(UBound(parts) - 1)
End Function

'add a slash to the end of a string
Public Function AddSlash(ByVal sVal As String) As String
    If Not Right$(sVal, 1) = "\" Then sVal = sVal & "\"
    AddSlash = sVal
End Function

'get the path for a file
Public Function GetPath(ByVal sFile As String) As String
    Dim parts() As String, i As Byte
    parts = Split(sFile, "\")
    For i = 0 To UBound(parts) - 1
        GetPath = GetPath & parts(i) & "\"
    Next
End Function

'is a directory?
Public Function isDir(ByVal curdir As String) As Boolean
    If GetAttr(curdir) And vbDirectory Then
        isDir = True
    Else
        isDir = False
    End If
End Function

'file exists?
Public Function FileExists(ByVal Path As String) As Boolean
    On Error Resume Next
    FileExists = (GetAttr(Path) And (vbDirectory Or vbVolume)) = 0
End Function

'directory exists?
Public Function DirExists(ByVal Path As String) As Boolean
    On Error Resume Next
    DirExists = CBool(GetAttr(Path) And vbDirectory)
End Function

Public Sub KillEx(file As String)
    On Error Resume Next
    Kill file
End Sub

Public Sub Sleep(ByVal lMillisecs As Long)
    Dim i As Integer

    For i = 1 To lMillisecs
        SleepA 10
        DoEvents
    Next
End Sub


'**************************************
'** Function: IsNT
'** Returns: Boolean
'** Usage: Returns TRUE, if the current platform is NT based
'**************************************
Public Function IsNT() As Boolean
    Dim osver As OSVERSIONINFOEX

    osver.dwOSVersionInfoSize = Len(osver)  ' set the type length
    GetVersionEx osver                      ' get the windows version

    ' check the version and return TRUE for NT systems
    If osver.dwPlatformId = VER_PLATFORM_WIN32_NT Then
        If osver.dwMajorVersion >= 4 Then IsNT = True
        If osver.dwMajorVersion >= 5 Then bW2K = True
    End If
End Function

'**************************************
'** Function: WriteWavHdr
'** Returns: void
'** Usage: writes a standard wave header (44.1kHz 16-bit stereo) to a file handle
'**************************************
Public Sub WriteWavHdr(ByVal FileHandle As Integer, ByVal F_Len As Long)
    Dim WavHdr As t_WavHdr

    With WavHdr
        .riff = "RIFF"                  ' RIFF chunk
        .len = F_Len + 44 - 8           ' length of the file
        .WavFmt = "WAVEfmt "            ' WAVEfmt
        .HdrLen = 16                    ' header length
        .format = 1                     ' format
        .NumChannels = 2                ' number of channels (stereo)
        .SampleRate = 44100             ' sample rate (44.1kHz)
        .BytesPerSec = 44100 * 2 * 2    ' Bytes Per Second
        .BlockAlign = 4                 ' Block Align
        .BitsPerSample = 16             ' Bits per sample
        .Data = "data"                  ' data chunk
        .DataLen = F_Len - 44           ' data length without header length
    End With

    Put #FileHandle, , WavHdr
End Sub

'**************************************
'** Function: LShift
'** Returns: Long
'** Usage: shifts a value by pnShift bits to the left
'** Source: Planet-Source-Code
'**************************************
Public Function LShift(ByVal pnValue As Long, ByVal pnShift As Long) As Long
    LShift = pnValue * (2 ^ pnShift)
End Function

'**************************************
'** Function: RShift
'** Returns: Long
'** Usage: shifts a value by pnShift bits to the right
'** Source: Planet-Source-Code
'**************************************
Public Function RShift(ByVal pnValue As Long, ByVal pnShift As Long) As Long
    RShift = CLng(pnValue \ (2 ^ pnShift))
End Function

'**************************************
'** Function: IsBitSet
'** Returns: Boolean
'** Usage: detects if a specific bit in a byte is set
'** Source: Planet-Source-Code
'**************************************
Public Function IsBitSet(ByVal InByte As Byte, ByVal Bit As Byte) As Boolean
    IsBitSet = ((InByte And (2 ^ Bit)) > 0)
End Function

'if LBA is positive but negative on media, convert to negative value
Public Function CHECKLBA(ByVal LBA As Long) As Long
    Dim sMSF As String, mins As Long, secs As Long, frames As Long
    sMSF = LBA2MSF(LBA)

    mins = Left(sMSF, 2)
    secs = Mid(sMSF, 4, 2)
    frames = Right(sMSF, 2)

    CHECKLBA = MSF2LBA(mins, secs, frames, False)
End Function

'**************************************
'** Function: MSF2LBA
'** Returns: Long
'** Usage: converts  MSF (min:sec:frm) to LBA (logical block address)
'**************************************
Public Function MSF2LBA(ByVal mins As Long, ByVal secs As Long, ByVal frames As Long, Optional positive As Boolean) As Long
    Dim calcbuf As Long
    calcbuf = mins * 60
    calcbuf = calcbuf * 75 + (secs * 75)
    MSF2LBA = calcbuf + frames

    If mins < 90 Or positive Then
        MSF2LBA = MSF2LBA - 150
    Else
        MSF2LBA = MSF2LBA - 450150
    End If
End Function

'Decode SCSI.crp to SCSI.dll
Public Sub DecodeSCSILib()
    Dim i As Long
    Dim b As Byte
    Dim FF1 As Integer: FF1 = FreeFile
    Dim FF2 As Integer: FF2 = FreeFile + 1

    If FileExists(App.Path & "\SCSI.dll") Then Exit Sub
    'If Not FileExists(App.Path & "\SCSI.crp") Then Exit Sub

    Open App.Path & "\SCSI.crp" For Binary As #FF1
    Open App.Path & "\SCSI.dll" For Binary As #FF2

        For i = 1 To FileLen(App.Path & "\SCSI.crp")
            Get #FF1, , b
            Put #FF2, , CByte(255 - b)
            DoEvents
        Next

    Close #FF1: Close #FF2
    FileCopy App.Path & "\SCSI.dll", AddSlash(GetSysDir) & "SCSI.dll"
    'Kill App.Path & "\SCSI.crp"
End Sub

'**************************************
'** Function: LBA2MSF
'** Returns: String [8 chars], format: "00:00:00"
'** Usage: converts LBA (logical block address) to MSF (min:sec:frm)
'**************************************
Public Function LBA2MSF(ByVal LBA As Long) As String
    Dim M As Long, S As Long, f As Long

    If LBA >= -150 Then
        M = Fix((LBA + 150) / (60 * 75))
        S = Fix((LBA + 150 - M * 60 * 75) / 75)
        f = Fix(LBA + 150 - M * 60 * 75 - S * 75)
    ElseIf LBA >= -45150 And LBA <= -151 Then
        M = Fix((LBA + 450150) / (60 * 75))
        S = Fix((LBA + 450150 - M * 60 * 75) / 75)
        f = Fix(LBA + 450150 - M * 60 * 75 - S * 75)
    End If
    LBA2MSF = format(M, "00") & ":" & format(S, "00") & ":" & format(f, "00")
End Function
