Attribute VB_Name = "modMMC"
Option Explicit

Private m_lPower2(0 To 31) As Long          ' from vb-accelerator: unsigned shfiting values

Public Const DTYPE_CDROM As Byte = 5           ' Device Type [CD-Rom]

'Synchronization pattern for data sectors (Mode-1 and Mode-2 [Form1/2])
Public Const SyncPattern = "00" & "FF" & "FF" & "FF" & "FF" & "FF" & _
                           "FF" & "FF" & "FF" & "FF" & "FF" & "00"

Public Const SCSI_TST_U_RDY As Byte = &H0      ' SCSI command [Test Unit Ready]
Public Const SCSI_REQ_SENSE As Byte = &H3      ' SCSI command [Request Sense]
Public Const SCSI_READ As Byte = &H8           ' SCSI command [Read]
Public Const SCSI_WRITE As Byte = &HA          ' SCSI command [Write]
Public Const SCSI_INQUIRY As Byte = &H12       ' SCSI command [Inquiry]
Public Const SCSI_MODE_SEL6 As Byte = &H15     ' SCSI command [ModeSelect6]
Public Const SCSI_MODE_SEN6 As Byte = &H1A     ' SCSI command [ModeSense6]
Public Const SCSI_MODE_SEL10 As Byte = &H55    ' SCSI command [ModeSelect10]
Public Const SCSI_MODE_SEN10 As Byte = &H5A    ' SCSI command [ModeSense10]
Public Const SCSI_SEEK As Byte = &H2B          ' SCSI command [Seek10]
Public Const SCSI_READ_CAP As Byte = &H25      ' SCSI command [Read Capacity]
Public Const SCSI_READ_DISC_NFO As Byte = &H51 ' SCSI command [Read Disc Information]
Public Const SCSI_READ_TOC As Byte = &H43      ' SCSI command [Read TOC/PMA/ATIP]
Public Const SCSI_READ_TRK_NFO As Byte = &H52  ' SCSI command [Read Track Information]
Public Const SCSI_READ_CD As Byte = &HBE       ' SCSI command [Read CD]
Public Const SCSI_READ10 As Byte = &H28        ' SCSI command [Read10]
Public Const SCSI_READD4 As Byte = &HD4        ' SCSI command [Read D4]
Public Const SCSI_READD8 As Byte = &HD8        ' SCSI command [Read D8]
Public Const SCSI_WRITE10 As Byte = &H2A       ' SCSI command [Write10]
Public Const SCSI_SETSPEED As Byte = &HBB      ' SCSI command [Set CD Speed]
Public Const SCSI_BLANK As Byte = &HA1         ' SCSI command [Blank disc]
Public Const SCSI_SYNC_CACHE As Byte = &H35    ' SCSI command [Synchronize Cache]
Public Const SCSI_CLOSE_TRK_SESS As Byte = &H5B ' SCSI command [Close Track/Session]
Public Const SCSI_LOUNLOAD As Byte = &H1B      ' SCSI command [Load/Unload]
Public Const SCSI_LOUNLOCK As Byte = &H1E      ' SCSI command [Lock/Unlock]
Public Const SCSI_GETEVENTSTATUS As Byte = &H4A ' SCSI command [Get Event Status]
Public Const SCSI_OPCINFO As Byte = &H54       ' SCSI command [OPC Information]
Public Const SCSI_READ_SUBCH As Byte = &H42    ' SCSI command [Read Sub-Channel]
Public Const SCSI_SENDCUE As Byte = &H5D       ' SCSI command [Send Cue-Sheet]
Public Const SCSI_READ_BUF_CAP As Byte = &H5C  ' SCSI command [Read Buffer Capacity]
Public Const SCSI_GET_CONFIG As Byte = &H46    ' SCSI command [Get Configuration]

'**************************************
'** Function: TestUnitReady
'** Returns: Boolean
'** Usage: returns TRUE if the drive is ready
'**************************************
Public Function TestUnitReady(ByVal DrvID As String) As Boolean
    Dim cmd(5) As Byte
    TestUnitReady = ExecCMD(DrvID, cmd, 6, 0, False, SCSI_DIR_IN, 0, 0, 5)
End Function

'**************************************
'** Function: EraseCDRW
'** Returns: TRUE if it could blank the disc
'** Usage: blanks a disc, either in quick or full blank mode.
'**        Note: Program will hang until it's finished.
'**************************************
Public Function EraseCDRW(ByVal DrvID As String, ByVal BlankType As e_BLANKING_TYPE) As Boolean
    Dim cmd(11) As Byte

    cmd(0) = SCSI_BLANK ' BLANK command [A1h]
    cmd(1) = BlankType  ' Blanking type

    'execute
    EraseCDRW = ExecCMD(DrvID, cmd, 12, 0, True, SCSI_DIR_IN, 0, 0)
End Function

'**************************************
'** Function: LockMedia
'** Returns: TRUE if it could lock the media/tray
'** Usage: locks the media/tray,
'**        useful for CD writing or screen locking :)
'**************************************
Public Function LockMedia(ByVal DrvID As String) As Boolean
    Dim cmd(5) As Byte

    cmd(0) = SCSI_LOUNLOCK  ' LOCK/UNLOCK command [1Eh]
    cmd(4) = 1              ' lock flag

    'execute
    LockMedia = ExecCMD(DrvID, cmd, 6, 0, False, SCSI_DIR_IN, 0, 0, 5)
End Function

'**************************************
'** Function: UnlockMedia
'** Returns: TRUE if it could unlock the media/tray
'** Usage: unlocks the media/tray
'**************************************
Public Function UnlockMedia(ByVal DrvID As String) As Boolean
    Dim cmd(5) As Byte

    cmd(0) = SCSI_LOUNLOCK  ' LOCK/UNLOCK command [1Eh]
    cmd(4) = 0              ' remove all flags to unlock the media/tray

    'execute
    UnlockMedia = ExecCMD(DrvID, cmd, 6, 0, False, SCSI_DIR_IN, 0, 0, 5)
End Function

'**************************************
'** Function: GetEventStatusNotification
'** Returns: TRUE if it could return information
'** Usage: event/status info
'**************************************
Public Function GetEventStatusNotification(ByVal DrvID As String, ByVal Request As Byte, _
                         ByVal PtrBuffer As Long, ByVal BufferLen As Long) As Boolean
    Dim cmd(9) As Byte

    cmd(0) = SCSI_GETEVENTSTATUS    ' GET EVENT/STATUS NOTIFICATION command [4Ah]
    cmd(1) = 1                      '
    cmd(4) = Request                ' Request type
    cmd(8) = BufferLen              ' allocation length

    'execute
    GetEventStatusNotification = ExecCMD( _
                                    DrvID, cmd, 10, 0, False, SCSI_DIR_IN, _
                                    PtrBuffer, BufferLen, 5 _
                                 )
End Function

'**************************************
'** Function: SendOPCInformation
'** Returns: Boolean
'** Usage: Calibrates the laser to its optimal power (Optimal Power Calibration)
'**************************************
Public Function SendOPCInformation(ByVal DrvID As String, ByVal OPC As Boolean, _
                           PtrBuffer As Long, ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = SCSI_OPCINFO   ' SEND OPC INFORMATION command [54h]
    cmd(1) = IIf(OPC, 1, 0) ' Do OPC?
    cmd(8) = BufferLen      ' allocation length

    'execute
    SendOPCInformation = ExecCMD( _
                              DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, _
                              PtrBuffer, BufferLen _
                         )
End Function

'**************************************
'** Function: ReadTrackInformation
'** Returns: Boolean
'** Usage: reads various information about a specific track
'**************************************
Public Function ReadTrackInformation(ByVal DrvID As String, ByVal AdrType As Byte, _
                                      ByVal LBA_TRACK_SESSION As Byte, _
                                      ByVal PtrBuffer As Long, ByVal BufferLen As Long) _
                                      As Boolean

    Dim cmd(9) As Byte

    cmd(0) = SCSI_READ_TRK_NFO              ' READ TRACK INFORMATION command [52h]
    cmd(1) = AdrType                        ' Address Type
    cmd(2) = RShift(LBA_TRACK_SESSION, 24) And &HFF  ' Track/Session MSB
    cmd(3) = RShift(LBA_TRACK_SESSION, 16) And &HFF
    cmd(4) = RShift(LBA_TRACK_SESSION, 8) And &HFF
    cmd(5) = LBA_TRACK_SESSION And &HFF              ' Track/Session LSB
    cmd(7) = BufferLen \ &HFF               ' allocation length
    cmd(8) = BufferLen And &HFF             ' allocation length

    'execute
    ReadTrackInformation = ExecCMD( _
                                DrvID, cmd, 10, 0, False, SCSI_DIR_IN, _
                                PtrBuffer, BufferLen, 5 _
                           )
End Function

'**************************************
'** Function: GetSectorMode
'** Returns: e_SectorModes (Byte)
'** Usage: determines the mode of a specific sector
'** Code: translated from CDR-DAO
'**************************************
Public Function GetSectorMode(ByVal DrvID As String, ByVal SectorLBA As Long) As e_SectorModes
    Dim buffer(2351) As Byte
    Dim Sync As String, cnt As Integer

    'read the sector raw
    If Not ReadCD(DrvID, SectorLBA, 1, VarPtr(buffer(0)), UBound(buffer), &HF8) Then _
        GetSectorMode = -1: Exit Function

    'get the first 12 bytes (= SYNC Pattern)
    For cnt = 0 To 11
        Sync = Sync & format(Hex(buffer(cnt)), "00")
    Next

    'SYNC Pattern exists = Data sector
    If Sync = SyncPattern Then

        Select Case buffer(15)

            Case 1
                'Mode-1 Sector (2048 bytes userdata)
                GetSectorMode = MODE_MODE1

            Case 2
                If (buffer(16) = buffer(20)) And (buffer(17) = buffer(21)) _
                    And (buffer(18) = buffer(22)) And (buffer(19) = buffer(23)) Then

                    If Not buffer(18) And &H20 = 0 Then
                        'Mode 2 form 2 sector
                        GetSectorMode = MODE_MODE2_FORM2
                    Else
                        'Mode 2 form 1 sector
                        GetSectorMode = MODE_MODE2_FORM1
                    End If

                Else
                    'plain Mode-2 sector (2336 bytes userdata)
                    GetSectorMode = MODE_MODE2
                End If

        End Select

    Else
        'Audio or Mode-0 (2352 bytes userdata)
        GetSectorMode = MODE_AUDIO
    End If
End Function

'**************************************
'** Function: SectorMode2Str
'** Returns: String [undefined length]
'** Usage: converts value returned by GetSectorMode to a string
'**************************************
Public Function SectorMode2Str(ByVal sMode As e_SectorModes) As String
    Select Case sMode
        Case MODE_AUDIO
            SectorMode2Str = "AUDIO"
        Case MODE_MODE1
            SectorMode2Str = "MODE 1"
        Case MODE_MODE2
            SectorMode2Str = "MODE 2"
        Case MODE_MODE2_FORM1
            SectorMode2Str = "MODE 2 FORM 1"
        Case MODE_MODE2_FORM2
            SectorMode2Str = "MODE 2 FORM 2"
    End Select
End Function

'**************************************
'** Function: ReadSubChannel
'** Returns: Boolean
'** Usage: reads Sub-Channel data
'**************************************
Public Function ReadSubChannel(ByVal DrvID As String, ByVal MSF As Boolean, _
                                ByVal SUBQ As Boolean, ParamList As Byte, _
                                TrackNum As Byte, PtrBuffer As Long, BufferLen As Long) _
                                As Boolean

    Dim cmd(9) As Byte

    cmd(0) = SCSI_READ_SUBCH            ' READ SUB-CHANNEL [42h]
    cmd(1) = IIf(MSF, 3, 0)             ' MSF or LBA?
    cmd(2) = IIf(SUBQ, LShift(1, 6), 0) ' Q Sub-Channel?
    cmd(3) = ParamList                  ' Parameter list
    cmd(6) = TrackNum                   ' specific track number
    cmd(7) = BufferLen \ &HFF           ' allocation length
    cmd(8) = BufferLen Mod &HFF         ' allocation length

    'execute
    ReadSubChannel = ExecCMD( _
                        DrvID, cmd, 10, 0, False, SCSI_DIR_IN, PtrBuffer, BufferLen, 5 _
                     )
End Function

'**************************************
'** Function: Write10
'** Returns: Boolean
'** Usage: writes data to a disc, if supported
'**************************************
Public Function Write10(ByVal DrvID As String, ByVal LBA As Long, _
                         ByVal WrtSektors As Long, ByVal PtrBuffer As Long, _
                         ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_WRITE10               ' WRITE10 command [2Ah]
    cmd(2) = RShift(LBA, 24) And &HFF   ' MSB of LBA
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF               ' LSB of LBA
    cmd(7) = WrtSektors \ &HFF          ' allocation length
    cmd(8) = WrtSektors And &HFF        ' allocation length

    'execute
    Write10 = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, PtrBuffer, BufferLen)
End Function

'**************************************
'** Function: SyncCache
'** Returns: Boolean
'** Usage: writes rest data from the buffer to the disc, if supported
'**************************************
Public Function SyncCache(ByVal DrvID As String, ByVal LBA As Long) As Boolean
    Dim cmd(9) As Byte

    cmd(0) = SCSI_SYNC_CACHE            ' SYNCHRONIZE CACHE command [35h]
    cmd(2) = RShift(LBA, 24) And &HFF   ' MSB of LBA
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF               ' LSB of LBA
    cmd(8) = 1                          '

    'execute
    SyncCache = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, 0, 0)
End Function

'**************************************
'** Function: CloseCD
'** Returns: Boolean
'** Usage: writes a lead-out for a track/session/disc, if supported
'**************************************
Public Function CloseCD(ByVal DrvID As String, ByVal CloseFunction As Byte, _
                                       ByVal TrackNum As Integer) As Boolean

    Dim cmd(9) As Byte

    cmd(0) = SCSI_CLOSE_TRK_SESS        ' CLOSE TRACK/SESSION command [5Bh]
    cmd(2) = CloseFunction              ' Close function
    cmd(5) = TrackNum                   ' Track/Session number

    'execute
    CloseCD = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, 0, 0)
End Function

'**************************************
'** Function: BuildCDRVendorList
'** Returns: string array
'** Usage: generates a list of cd-r vendors with their typical lead-in start times
'**************************************
Private Function BuildCDRVendorList() As Variant
    Dim cdrv() As String
    ReDim cdrv(69)

     cdrv(0) = "97,28,30||97,46,50||Auvistar Industry Co.,Ltd."
     cdrv(1) = "97,26,60||97,46,60||CMC Magnetics Corporation"
     cdrv(2) = "97,23,10||00,00,00||Doremi Media Co., Ltd."
     cdrv(3) = "97,26,00||97,45,00||FORNET International PTE Ltd."
     cdrv(4) = "97,46,40||97,46,40||FUJI Photo Film Co., Ltd."
     cdrv(5) = "97,26,40||00,00,00||FUJI Photo Film Co., Ltd."
     cdrv(6) = "97,28,10||97,49,10||Gigastore Corporation"
     cdrv(7) = "97,25,20||97,47,10||Hitachi Maxwell, Ltd."
     cdrv(8) = "97,27,40||97,48,10||Kodak Japan Limited"
     cdrv(9) = "97,26,50||97,48,60||Lead Data Inc."
    cdrv(10) = "97,27,50||97,48,50||Mitsui Chemicals, Inc."
    cdrv(11) = "97,34,20||97,50,20||Mitsubishi Chemical Corporation"
    cdrv(12) = "97,28,20||97,46,20||Multi Media Masters & Machinary SA"
    cdrv(13) = "97,21,40||00,00,00||Optical Disc Manufacturing Equipment"
    cdrv(14) = "97,27,30||97,48,30||Pioneer Video Corporation"
    cdrv(15) = "97,27,10||97,48,20||Plasmon Data systems Ltd."
    cdrv(16) = "97,26,10||97,47,40||POSTECH Corporation"
    cdrv(17) = "97,27,20||97,47,20||Princo Corporation"
    cdrv(18) = "97,32,10||00,00,00||Prodisc Technology Inc."
    cdrv(19) = "97,27,60||97,48,00||Ricoh Company Limited"
    cdrv(20) = "97,31,00||97,47,50||Ritek Co."
    cdrv(21) = "97,26,20||00,00,00||SKC Co., Ltd."
    cdrv(22) = "97,24,10||00,00,00||SONY Corporation"
    cdrv(23) = "97,24,00||97,46,00||Taiyo Yuden Company Limited"
    cdrv(24) = "97,32,00||97,49,00||TDK Corporation"
    cdrv(25) = "97,25,60||97,45,60||Xcitek Inc."
    cdrv(26) = "97,22,60||97,45,20||Acer Media Technology, Inc"
    cdrv(27) = "97,25,50||00,00,00||AMS Technology Inc."
    cdrv(28) = "97,23,30||00,00,00||Audio Distributors Co., Ltd."
    cdrv(29) = "97,21,30||00,00,00||Bestdisc Technology Corporation"
    cdrv(30) = "97,30,10||97,50,30||CDA Datentraeger Albrechts GmbH"
    cdrv(31) = "97,22,40||97,45,40||CIS Technology Inc."
    cdrv(32) = "97,24,20||97,46,30||Computer Support Italy s.r.l."
    cdrv(33) = "97,23,60||00,00,00||Customer Pressing Oosterhout"
    cdrv(34) = "97,28,50||00,00,00||Delphi Technology Inc."
    cdrv(35) = "97,27,00||97,48,40||Digital Storage Technology Co., Ltd."
    cdrv(36) = "97,22,30||00,00,00||EXIMPO"
    cdrv(37) = "97,28,60||00,00,00||Friendly CD-Tek Co."
    cdrv(38) = "97,31,30||97,51,10||Grand Advance Technology Ltd."
    cdrv(39) = "97,29,50||00,00,00||General Magnetics Ld"
    cdrv(40) = "97,24,50||97,45,50||Guann Yinn Co.,Ltd."
    cdrv(41) = "97,29,00||00,00,00||Harmonic Hall Optical Disc Ltd."
    cdrv(42) = "97,29,30||97,51,50||Hile Optical Disc Technology Corp."
    cdrv(43) = "97,46,10||97,22,50||Hong Kong Digital Technology Co., Ltd."
    cdrv(44) = "97,25,30||97,51,20||INFODISC Technology Co., Ltd."
    cdrv(45) = "97,24,40||00,00,00||kdg mediatech AG"
    cdrv(46) = "97,28,40||97,49,20||King Pro Mediatek Inc."
    cdrv(47) = "97,23,00||97,49,60||Matsushita Electric Industrial Co., Ltd."
    cdrv(48) = "97,15,20||00,00,00||Mitsubishi Chemical Corporation"
    cdrv(49) = "97,25,00||00,00,00||Hi-SPACE, MPO"
    cdrv(50) = "97,23,20||00,00,00||Nacar Media sr"
    cdrv(51) = "97,26,30||00,00,00||Optical Disc Corporation"
    cdrv(52) = "97,28,00||97,49,30||Opti.Me.S. S.p.A."
    cdrv(53) = "97,23,50||00,00,00||OPTROM.INC."
    cdrv(54) = "97,47,60||00,00,00||Prodisc Technology Inc."
    cdrv(55) = "97,15,10||00,00,00||Ritek Co."
    cdrv(56) = "97,22,10||00,00,00||Seantram Technology Inc."
    cdrv(57) = "97,21,50||00,00,00||Sound Sound Multi-Media Development Limited"
    cdrv(58) = "97,29,00||00,00,00||Taeil Media Co.,Ltd."
    cdrv(59) = "97,18,60||00,00,00||Taroko International Co.,Ltd."
    cdrv(60) = "97,15,00||00,00,00||TDK Corporation."
    cdrv(61) = "97,29,20||00,00,00||Unidisc Technology Co.,Ltd."
    cdrv(62) = "97,24,30||97,45,10||Unitech Japan Inc."
    cdrv(63) = "97,29,10||97,50,10||Vanguard Disc Inc."
    cdrv(64) = "97,49,40||97,23,40||Victor Company of Japan, Ltd."
    cdrv(65) = "97,29,40||00,00,00||Viva Magnetics, Ltd."
    cdrv(66) = "97,25,40||00,00,00||Vivastar AG"
    cdrv(67) = "97,18,10||00,00,00||Wealth Fair Investment, Ltd."
    cdrv(68) = "97,22,00||00,00,00||Woongjin Media corp."
    cdrv(69) = "97,17,00||00,00,00||Moser Baer India Limited"

    BuildCDRVendorList = cdrv
End Function

'**************************************
'** Function: GetCDRVendor
'** Returns: Boolean
'** Usage: wrapper for BuildCDRVendorList()
'**        If it finds a valid vendor,
'**        sVendor gets the value.
'**        If not, GetCDRVendor is FALSE.
'**************************************
Public Function GetCDRVendor(ByVal DrvID As String, ByRef sVendor As String) As Boolean
    Dim cdrv() As String
    Dim atip As t_ATIP
    Dim i As Integer

    Dim MSF_CD As String
    Dim MSF_V1 As String
    Dim MSF_V2 As String

    'read the disc's ATIP
    If Not ReadTOC(DrvID, 4, True, 0, VarPtr(atip), Len(atip)) Then _
        Exit Function

    'get some vendors
    cdrv = BuildCDRVendorList()

    'get the Lead-In of the disc and round it up
    atip.LeadIn_Frm = val(Left(atip.LeadIn_Frm, 1) & "0")
    'convert it to a better format for better comparing
    MSF_CD = atip.LeadIn_Min & "," & atip.LeadIn_Sec & "," & atip.LeadIn_Frm

    For i = 0 To UBound(cdrv)
        'get the first possible Lead-In time for the current vendor
        MSF_V1 = Left(cdrv(i), InStr(1, cdrv(i), "||") - 1)
        'get the second possible Lead-In time for the current vendor
        MSF_V2 = Mid(cdrv(i), InStr(1, cdrv(i), "||") + 2, InStrRev(cdrv(i), "||") - 3 - Len(MSF_V1))

        If (MSF_CD = MSF_V1) Or (MSF_CD = MSF_V2) Then
            'Bingo!
            sVendor = Mid(cdrv(i), Len(MSF_V1) + Len(MSF_V2) + 5)
            GetCDRVendor = True
            Exit Function
        End If
    Next
End Function

'**************************************
'** Function: IsTrayOpen
'** Returns: Long
'**          -1: command failed
'**           0: closed
'**           1: open
'** Usage: checks, if the drive's tray is open
'**************************************
Public Function IsTrayOpen(ByVal DrvID As String) As Long
    Dim media_event_req(8) As Byte

    'get information about the tray
    If Not GetEventStatusNotification(DrvID, &H10, VarPtr(media_event_req(0)), UBound(media_event_req)) Then _
        IsTrayOpen = -1: Exit Function

    'check for valid data
    If media_event_req(0) = 0 And media_event_req(1) = 0 Then
        IsTrayOpen = -1: Exit Function
    Else
        If IsBitSet(media_event_req(5), 0) Then
            'tray is open
            IsTrayOpen = 1
        Else
            'tray is closed
            IsTrayOpen = 0
        End If
    End If
End Function

'**************************************
'** Function: IsDiscPresent
'** Returns: Long
'**          -1: command failed
'**           0: not present
'**           1: present
'** Usage: checks, if a disc is present or not
'**************************************
Public Function IsDiscPresent(ByVal DrvID As String) As Long
    Dim media_event_req(8) As Byte

    'get the tray status
    If Not GetEventStatusNotification(DrvID, &H10, VarPtr(media_event_req(0)), UBound(media_event_req)) Then _
        IsDiscPresent = -1: Exit Function

    'check for valid data
    If media_event_req(0) = 0 And media_event_req(1) = 0 Then
        IsDiscPresent = -1: Exit Function
    Else
        If IsBitSet(media_event_req(5), 1) Then
            'disc is present
            IsDiscPresent = 1
        Else
            'no disc in there
            IsDiscPresent = 0
        End If
    End If
End Function

'**************************************
'** Function: IsTrayLocked
'** Returns: Long
'**          -1: function failed
'**           0: tray isn't locked
'**           1: trayk is locked
'** Usage: checks if the media/tray is locked
'**************************************
Public Function IsTrayLocked(ByVal DrvID As String) As Long
    Dim mmc As t_MMC

    'get the multimedia capabilities page
    If Not ModeSense10(DrvID, &H2A, VarPtr(mmc), Len(mmc)) Then _
        IsTrayLocked = -1: Exit Function

    'check the "Lock State" bit
    IsTrayLocked = Abs(IsBitSet(mmc.misc(2), 1))
End Function

'**************************************
'** Function: Seek10
'** Returns: Boolean
'** Usage: Sets the laser-position
'**************************************
Public Function Seek10(ByVal DrvID As String, ByVal LBA As Long) As Boolean
    Dim cmd(9) As Byte

    cmd(0) = SCSI_SEEK                      ' SEEK10 command [2Bh]
    cmd(2) = RShift(LBA, 24) And &HFF       ' MSB of LBA
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF                   ' LSB of LBA

    'execute
    Seek10 = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_IN, 0, 0)
End Function

'**************************************
'** Function: Read10
'** Returns: Boolean
'** Usage: reads Mode-1 sectors
'**************************************
Public Function Read10(ByVal DrvID As String, ByVal LBA As Long, _
                        ByVal PtrBuf As Long, ByVal BufLen As Long, _
                        ByVal Blocks As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_READ10                ' READ10 command (28h)
    cmd(2) = RShift(LBA, 24) And &HFF   ' MSB of LBA
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF               ' LSB of LBA
    cmd(7) = Blocks \ &HFF              ' number of blocks to read
    cmd(8) = Blocks And &HFF            ' number of blocks to read

    'execute
    Read10 = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_IN, PtrBuf, BufLen)
End Function

'**************************************
'** Function: SessionInfo
'** Returns: cSessInfo
'** Usage: collects various information about a session and returns it in a class
'**************************************
Public Function SessionInfo(ByVal DrvID As String, ByVal SessNum As Integer) As cSessInfo
    Dim nfo As New cSessInfo
    Dim tnfo As New cTrackInfo
    Dim TOC As t_RTOC_STRUCT
    Dim i As Long

    'read the RAW TOC
    If Not ReadTOC(DrvID, 2, True, 1, VarPtr(TOC), Len(TOC)) Then Exit Function

        'parse the packets of the TOC
        For i = 0 To ((LShift(TOC.dummy(0), 8) Or TOC.dummy(1)) \ 11) - 1

            Select Case TOC.packet(i).point

                'get the LBA of the first track in the session
                Case nfo.FirstTrack
                    nfo.FirstTrackLBA = MSF2LBA(TOC.packet(i).pmin, TOC.packet(i).psec, TOC.packet(i).pframe)

                'found the first track of the session
                Case &HA0
                    If TOC.packet(i).sessionNr = SessNum Then _
                        nfo.FirstTrack = TOC.packet(i).pmin

                'found the last track of the session
                Case &HA1
                    If TOC.packet(i).sessionNr = SessNum Then _
                        nfo.LastTrack = TOC.packet(i).pmin

                'found the Lead-Out of the session
'                Case &HA2
'                    nfo.LeadOutLBA = MSF2LBA(toc.packet(i).pmin, toc.packet(i).psec, toc.packet(i).pframe)

            End Select

        Next i

    Set tnfo = TrackInfo(DrvID, nfo.LastTrack)

    nfo.SessionNumber = SessNum
    nfo.LeadOutLBA = tnfo.EndLBA
    nfo.Size = nfo.LeadOutLBA - nfo.FirstTrackLBA

    'return the class
    Set SessionInfo = nfo
End Function

'**************************************
'** Function: CDInfo
'** Returns: cCDInfo
'** Usage: collects various information about a disc and returns it in a class
'**************************************
Public Function CDInfo(ByVal DrvID As String) As cCDInfo
    Dim nfo As New cCDInfo
    Dim BufRDI As t_RDI
    Dim BufCap As t_ReadCap
    Dim BufAtip As t_ATIP
    Dim conf_hdr(512) As Byte
    Dim sBuf As String

    'collect information about the disc
    ReadDiscInformation DrvID, VarPtr(BufRDI), Len(BufRDI)
    ReadCapacity DrvID, VarPtr(BufCap), Len(BufCap)
    ReadTOC DrvID, 4, True, 0, VarPtr(BufAtip), Len(BufAtip)

    'fill the class with main information
    
    
    With BufAtip

        nfo.LeadInLBA = MSF2LBA(.LeadIn_Min, .LeadIn_Sec, .LeadIn_Frm)
        nfo.LeadOutLBA = MSF2LBA(.LeadOut_Min, .LeadOut_Sec, .LeadOut_Frm)
        nfo.Capacity = MSF2LBA(.LeadOut_Min, .LeadOut_Sec, .LeadOut_Frm)

    End With

    With BufRDI

        If IsBitSet(.states, 1) = False And IsBitSet(.states, 0) = False Then
            nfo.DiscStatus = STAT_EMPTY
        ElseIf IsBitSet(.states, 1) = False And IsBitSet(.states, 0) Then
            nfo.DiscStatus = STAT_INCOMPLETE
        ElseIf IsBitSet(.states, 1) And IsBitSet(.states, 0) = False Then
            nfo.DiscStatus = STAT_COMPLETE
        Else
            nfo.DiscStatus = STAT_UNKNWN
        End If

        If Not IsBitSet(.states, 3) And Not IsBitSet(.states, 2) Then
            nfo.LastSessionStatus = STAT_EMPTY
        ElseIf Not IsBitSet(.states, 3) And IsBitSet(.states, 2) Then
            nfo.LastSessionStatus = STAT_INCOMPLETE
        ElseIf IsBitSet(.states, 3) And IsBitSet(.states, 2) Then
            nfo.LastSessionStatus = STAT_COMPLETE
        Else
            nfo.DiscStatus = STAT_UNKNWN
        End If

        If .DiscType = &H0 Then
            nfo.CDSubType = STYPE_CDROMDA
        ElseIf .DiscType = &H10 Then
            nfo.CDSubType = STYPE_CDI
        ElseIf .DiscType = &H20 Then
            nfo.CDSubType = STYPE_XA
        Else
            nfo.CDSubType = STYPE_UNKNWN
        End If

        nfo.Erasable = IsBitSet(.states, 4)

        If Not GetCDRVendor(DrvID, sBuf) Then
            nfo.Vendor = "unknown"
        Else
            nfo.Vendor = sBuf
        End If

        nfo.FirstTrackInLastSession = LShift(.FirstTrackLastSessionMSB, 8) Or .FirstTrackLastSessionLSB
        nfo.LastTrackInLastSession = LShift(.LastTrackLastSessionMSB, 8) Or .LastTrackLastSessionLSB
        nfo.Sessions = LShift(.NumSessionsMSB, 8) Or .NumSessionsLSB

    End With

    With BufCap
        nfo.Size = LShift(.Blocks(0), 24) Or _
                   LShift(.Blocks(1), 16) Or _
                   LShift(.Blocks(2), 8) Or _
                   .Blocks(3)

    End With

    If GetConfiguration(DrvID, 0, 2, VarPtr(conf_hdr(0)), UBound(conf_hdr)) Then

        'detect the type of the disc by the currently active profile
        Select Case (LShift(conf_hdr(6), 8) Or conf_hdr(7))
            Case &H8: nfo.CDType = TYPE_CDROM
            Case &H9: nfo.CDType = TYPE_CDR
            Case &HA: nfo.CDType = TYPE_CDRW
            Case &H10: nfo.CDType = TYPE_DVD_ROM
            Case &H11: nfo.CDType = TYPE_DVD_R
            Case &H12: nfo.CDType = TYPE_DVD_RAM
            Case &H13: nfo.CDType = TYPE_DVD_RW
            Case &H14: nfo.CDType = TYPE_DVD_RW
            Case &H1A: nfo.CDType = TYPE_DVD_P_RW
            Case &H1B: nfo.CDType = TYPE_DVD_P_R
            Case Else: GoTo DTOW
        End Select

    Else

DTOW:
        'detect the type of the disc
        'DVD?
        If IsDVD(DrvID) Then
            Select Case GetDVDBookType(DrvID)
                Case VDB_DVD_P_R: nfo.CDType = TYPE_DVD_P_R
                Case VDB_DVD_P_RW: nfo.CDType = TYPE_DVD_P_RW
                Case VDB_DVD_R: nfo.CDType = TYPE_DVD_R
                Case VDB_DVD_RW: nfo.CDType = TYPE_DVD_RW
                Case VDB_DVD_RAM: nfo.CDType = TYPE_DVD_RAM
                Case VDB_DVD_ROM: nfo.CDType = TYPE_DVD_ROM
                Case VDB_UNKNWN: GoTo MBCD
            End Select
        Else
MBCD:
            If Not ReadTOC(DrvID, 4, True, 0, VarPtr(BufAtip), Len(BufAtip)) Then
                'CD-ROM?
                If MSF2LBA(BufRDI.LastPossibleLeadOutStart(1), BufRDI.LastPossibleLeadOutStart(2), BufRDI.LastPossibleLeadOutStart(3)) = 1166730 Then
                    nfo.CDType = TYPE_CDROM
                Else
                    'couldn't detect wether it is a CD-ROM, CD-R or CD-RW
                    nfo.CDType = TYPE_CDROM_R_RW
                End If
            Else
                If IsBitSet(BufAtip.DiscType, 6) Then
                    'disc is a CD-R
                    nfo.CDType = TYPE_CDRW
                Else
                    'disc is a CD-RW
                    nfo.CDType = TYPE_CDR
                End If
            End If
        End If

    End If

    'read the MCN (Media Catalog Number)
    nfo.MCN = ReadMCN(DrvID)

    Set CDInfo = nfo

End Function

'**************************************
'** Function: IsDVD
'** Returns: Boolean
'** Usage: checks if the inserted disc is a DVD
'**************************************
Private Function IsDVD(ByVal DrvID As String) As Boolean
    Dim dummy(512) As Byte

    If READ_DVD_STRUCTURE(DrvID, 0, 0, 0, VarPtr(dummy(0)), UBound(dummy)) Then
        If dummy(0) > 0 Or dummy(1) > 0 Then IsDVD = True
    End If
End Function

'**************************************
'** Function: DeactivateECC
'** Returns: Boolean
'** Usage: trys to deactive error correction for a drive
'**************************************
Public Function DeactivateECC(ByVal DrvID As String) As Boolean
    Dim page(16)
    page(8) = 1
    page(9) = 6
    page(10) = &H11

    'set the new mode page
    DeactivateECC = ModeSelect10(DrvID, VarPtr(page(0)), UBound(page))
End Function

'**************************************
'** Function: ActivateECC
'** Returns: Boolean
'** Usage: trys to active error correction for a drive
'**************************************
Public Function ActivateECC(ByVal DrvID As String) As Boolean
    Dim page(16)
    page(8) = 1
    page(9) = 6
    page(10) = 1

    'set the new mode page
    ActivateECC = ModeSelect10(DrvID, VarPtr(page(0)), UBound(page))
End Function

'**************************************
'** Function: GetDVDBookType
'** Returns: Boolean
'** Usage: gets the book type of the inserted DVD
'**************************************
Private Function GetDVDBookType(ByVal DrvID As String) As e_DVDBooks
    Dim physdata As t_DVD_Phys
    Dim book As Byte

    If READ_DVD_STRUCTURE(DrvID, 0, 0, 0, VarPtr(physdata), Len(physdata)) Then
        With physdata
            If IsBitSet(.booktype, 4) Then book = 1
            If IsBitSet(.booktype, 5) Then book = book Or 2
            If IsBitSet(.booktype, 6) Then book = book Or 4
            If IsBitSet(.booktype, 7) Then book = book Or 8

            Select Case book
                Case 0: GetDVDBookType = VDB_DVD_ROM
                Case 1: GetDVDBookType = VDB_DVD_RAM
                Case 2: GetDVDBookType = VDB_DVD_R
                Case 3: GetDVDBookType = VDB_DVD_RW
                Case 9: GetDVDBookType = VDB_DVD_P_RW
                Case 10: GetDVDBookType = VDB_DVD_P_R
                Case Else: GetDVDBookType = VDB_UNKNWN
            End Select

        End With
    End If
End Function

'**************************************
'** Function: READ_DVD_STRUCTURE
'** Returns: Boolean
'** Usage: reads information depending on format value from a DVD
'**************************************
Private Function READ_DVD_STRUCTURE(ByVal DrvID As String, ByVal LBA As Long, _
                                    ByVal LayerNr As Byte, ByVal format As Byte, _
                                    ByVal PtrBuffer As Long, ByVal BufferLen As Long _
                                   ) As Boolean

    Dim cmd(11) As Byte

    cmd(0) = &HAD                           ' READ DVD STRUCTURE command
    cmd(2) = RShift(LBA, 24) And &HFF       ' LBA LSB
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF                   ' LBA MSB
    cmd(6) = LayerNr                        ' Layer Number
    cmd(7) = format                         ' information format
    cmd(8) = BufferLen \ &HFF               ' allocation length
    cmd(9) = BufferLen And &HFF             ' allocation length

    'execute
    READ_DVD_STRUCTURE = ExecCMD(DrvID, cmd, 12, 0, False, SCSI_DIR_IN, PtrBuffer, BufferLen, 5)
End Function

'**************************************
'** Function: ReadISRC
'** Returns: String (15 chars)
'** Usage: trys to read the ISRC of a track stored in the Q Sub-Channel
'**************************************
Private Function ReadISRC(ByVal DrvID As String, ByVal TrkNum As Byte) As String
    Dim ISRC As t_Q_ISRC
    Dim i As Integer

    'read the Q Sub-Channel
    ReadSubChannel DrvID, True, True, &H3, TrkNum, VarPtr(ISRC), Len(ISRC)

    'return the ISRC (International Standard Recording Code)
    With ISRC
        ReadISRC = ChrW$(.ISRC(1)) & ChrW$(.ISRC(2)) & " " & _
                   ChrW$(.ISRC(3)) & ChrW$(.ISRC(4)) & ChrW$(.ISRC(5)) & " " & _
                   ChrW$(.ISRC(6)) & ChrW$(.ISRC(7)) & " " & _
                   ChrW$(.ISRC(8)) & ChrW$(.ISRC(9)) & ChrW$(.ISRC(10)) & ChrW$(.ISRC(11)) & ChrW$(.ISRC(12))
    End With
End Function

'**************************************
'** Function: ReadMCN
'** Returns: String (13 chars)
'** Usage: trys to read the MCN of a disc stored in the Q Sub-Channel
'**************************************
Private Function ReadMCN(ByVal DrvID As String) As String
    Dim MCN As t_Q_MCN
    Dim i As Integer

    'read the Q Sub-Channel
    ReadSubChannel DrvID, True, True, &H2, 0, VarPtr(MCN), Len(MCN)

    'return the MCN (media catalog number)
    For i = 1 To 13
        ReadMCN = ReadMCN & ChrW$(MCN.MCN(i))
    Next
End Function

'**************************************
'** Function: Interface2Str
'** Returns: String (undefined length)
'** Usage: convert the value returned by GetInterface to a string
'**************************************
Public Function Interface2Str(ByVal iVal As e_Interfaces) As String
    Select Case iVal
        Case IF_ATAPI: Interface2Str = "ATAPI"
        Case IF_IEEE: Interface2Str = "IEEE"
        Case IF_SCSI: Interface2Str = "SCSI"
        Case IF_UNKNWN: Interface2Str = "unknown"
    End Select
End Function

'**************************************
'** Function: SendCueSheet
'** Returns: Boolean
'** Usage: sends a cue sheet to the drive,
'**        returns FALSE if format isn't supported or
'**        the drive doesn't support write mode (WMP) + Q sheet
'**************************************
Public Function SendCueSheet(ByVal DrvID As String, ByVal PtrBuf As Long, BufLen As Long, CueSheetLen As Long) As Boolean
    Dim cmd(9) As Byte
    cmd(0) = SCSI_SENDCUE
    cmd(6) = LShift(CueSheetLen, 16) And &HFF
    cmd(7) = LShift(CueSheetLen, 8) And &HFF
    cmd(8) = CueSheetLen And &HFF

    SendCueSheet = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, PtrBuf, BufLen)
End Function

'**************************************
'** Function: TrackInfo
'** Returns: cTrackInfo
'** Usage: collects various information about a track
'**************************************
Public Function TrackInfo(ByVal DriveID As String, ByVal Track As Integer) As cTrackInfo
    Dim nfo As New cTrackInfo

    Dim TOC As t_RTOC_STRUCT: Dim rti As t_RTI
    Dim sectors As Long, LeadOut As Long, LastTrk As Long
    Dim q_pos As t_CDCurrPos
    Dim i As Integer

    'read the RAW TOC
    If Not ReadTOC(DriveID, 2, True, 1, VarPtr(TOC), Len(TOC)) Then Exit Function

        'parse the packets of the TOC
        For i = 0 To ((LShift(TOC.dummy(0), 8) Or TOC.dummy(1)) \ 11) - 1

            Select Case TOC.packet(i).point

                'search for the track number
                Case Is = Track

                    nfo.Track = Track
                    nfo.session = TOC.packet(i).sessionNr

                    nfo.StartLBA = MSF2LBA(TOC.packet(i).pmin, TOC.packet(i).psec, TOC.packet(i).pframe)
                    If TOC.packet(i).point = LastTrk Then
                        nfo.EndLBA = LeadOut
                    Else
                        nfo.EndLBA = MSF2LBA(TOC.packet(i + 1).pmin, TOC.packet(i + 1).psec, TOC.packet(i + 1).pframe)
                    End If
                    nfo.length = nfo.EndLBA - nfo.StartLBA

                    If ReadSubChannel(DriveID, True, True, 1, 0, VarPtr(q_pos), Len(q_pos)) Then

                        If Not (IsBitSet(q_pos.ADRCTRL, 0) And IsBitSet(q_pos.ADRCTRL, 2) And IsBitSet(q_pos.ADRCTRL, 3)) Then
                            nfo.TrackMode = TRK_STEREOAUDIO
                        ElseIf IsBitSet(q_pos.ADRCTRL, 0) And Not (IsBitSet(q_pos.ADRCTRL, 2) And IsBitSet(q_pos.ADRCTRL, 3)) Then
                            nfo.TrackMode = TRK_STEREOAUDIO_EMPHASIS
                        ElseIf Not (IsBitSet(q_pos.ADRCTRL, 0) And IsBitSet(q_pos.ADRCTRL, 2)) And IsBitSet(q_pos.ADRCTRL, 3) Then
                            nfo.TrackMode = TRK_AUDIO
                        ElseIf IsBitSet(q_pos.ADRCTRL, 0) And Not IsBitSet(q_pos.ADRCTRL, 2) And IsBitSet(q_pos.ADRCTRL, 3) Then
                            nfo.TrackMode = TRK_AUDIO_EMPHASIS
                        ElseIf Not IsBitSet(q_pos.ADRCTRL, 0) And IsBitSet(q_pos.ADRCTRL, 2) And Not IsBitSet(q_pos.ADRCTRL, 3) Then
                            nfo.TrackMode = TRK_DATA_UNINTERRUPTED
                        ElseIf IsBitSet(q_pos.ADRCTRL, 0) And IsBitSet(q_pos.ADRCTRL, 2) And Not IsBitSet(q_pos.ADRCTRL, 3) Then
                            nfo.TrackMode = TRK_DATA_INCREMENTAL
                        Else
                            nfo.TrackMode = TRK_UNKNWN
                        End If

                    Else

                        ReadTrackInformation DriveID, 0, TOC.packet(i).point, VarPtr(rti), Len(rti)

                        If Not (IsBitSet(rti.TrackMode, 0) And IsBitSet(rti.TrackMode, 2) And IsBitSet(rti.TrackMode, 3)) Then
                            nfo.TrackMode = TRK_STEREOAUDIO
                        ElseIf IsBitSet(rti.TrackMode, 0) And Not (IsBitSet(rti.TrackMode, 2) And IsBitSet(rti.TrackMode, 3)) Then
                            nfo.TrackMode = TRK_STEREOAUDIO_EMPHASIS
                        ElseIf Not (IsBitSet(rti.TrackMode, 0) And IsBitSet(rti.TrackMode, 2)) And IsBitSet(rti.TrackMode, 3) Then
                            nfo.TrackMode = TRK_AUDIO
                        ElseIf IsBitSet(rti.TrackMode, 0) And Not IsBitSet(rti.TrackMode, 2) And IsBitSet(rti.TrackMode, 3) Then
                            nfo.TrackMode = TRK_AUDIO_EMPHASIS
                        ElseIf Not IsBitSet(rti.TrackMode, 0) And IsBitSet(rti.TrackMode, 2) And Not IsBitSet(rti.TrackMode, 3) Then
                            nfo.TrackMode = TRK_DATA_UNINTERRUPTED
                        ElseIf IsBitSet(rti.TrackMode, 0) And IsBitSet(rti.TrackMode, 2) And Not IsBitSet(rti.TrackMode, 3) Then
                            nfo.TrackMode = TRK_DATA_INCREMENTAL
                        Else
                            nfo.TrackMode = TRK_UNKNWN
                        End If

                    End If

                    Select Case GetSectorMode(DriveID, nfo.StartLBA)
                        Case MODE_AUDIO
                            nfo.DataMode = MODE_AUDIO
                        Case MODE_MODE1
                            nfo.DataMode = MODE_MODE1
                        Case MODE_MODE2
                            nfo.DataMode = MODE_MODE2
                        Case MODE_MODE2_FORM1
                            nfo.DataMode = MODE_MODE2_FORM1
                        Case MODE_MODE2_FORM2
                            nfo.DataMode = MODE_MODE2_FORM2
                    End Select
                    nfo.ISRC = ReadISRC(DriveID, Track)
                    GoTo ret

                'found the last track of the session
                Case &HA1
                    LastTrk = TOC.packet(i).pmin

                'found the Lead-Out of the session
                Case &HA2
                    LeadOut = MSF2LBA(TOC.packet(i).pmin, TOC.packet(i).psec, TOC.packet(i).pframe)

            End Select

        Next i

ret:
    'return the class
    Set TrackInfo = nfo
End Function

'**************************************
'** Function: IsAudioTrack
'** Returns: Boolean
'** Usage: checks if a specific track on a disc is an audio track
'**************************************
Public Function IsAudioTrack(ByVal DriveID As String, ByVal TrkNum As Integer) As Boolean
    Dim TOC As t_STOC_STRUCT
    ReadTOC DriveID, 0, True, 0, VarPtr(TOC), Len(TOC)              ' Read the simple TOC
    IsAudioTrack = IsBitSet(TOC.TocTrack(TrkNum).ADR, 2) = False    'check the ADR field of the track
End Function

'**************************************
'** Function: EjectTray
'** Returns: Boolean
'** Usage: returns TRUE if the drive's tray could be ejected
'**************************************
Public Function EjectTray(ByVal DrvID As String) As Boolean
    Dim cmd(6) As Byte

    cmd(0) = SCSI_LOUNLOAD  ' LOUNLOAD command [1Bh]
    cmd(4) = &H2            ' eject flag

    EjectTray = ExecCMD(DrvID, cmd, 6, 0, False, SCSI_DIR_IN, 0, 0, 5)
End Function

'**************************************
'** Function: LoadTray
'** Returns: Boolean
'** Usage: returns TRUE if the drive's tray/media could be loaded
'**************************************
Public Function LoadTray(ByVal DrvID As String) As Boolean
    Dim cmd(6) As Byte

    cmd(0) = SCSI_LOUNLOAD  ' LOUNLOAD command [1Bh]
    cmd(4) = &H3            ' load flag

    LoadTray = ExecCMD(DrvID, cmd, 6, 0, False, SCSI_DIR_IN, 0, 0, 5)
End Function

'**************************************
'** Function: ReadCD
'** Returns: Boolean
'** Usage: reads a specific sector on a disc,
'**        returns up to 2448 bytes of sector data
'**        depending on the read and sub-channel flags.
'**        returns TRUE, if the sector could be read.
'**************************************
Public Function ReadCD(ByVal DrvID As String, ByVal LBA As Long, _
                       ByVal NumSectors As Long, ByVal PtrBuf As Long, _
                       ByVal BufLen As Long, ByVal ReadFlags As e_READCD_FLAGS, _
                       Optional ByVal SubchBits As e_READCD_SUBCH_FLAGS) As Boolean

    Dim cmd(11) As Byte

    cmd(0) = SCSI_READ_CD                       ' READ CD command [BEh]
    cmd(2) = RShift(LBA, 24) And &HFF           ' MSB of LBA
    cmd(3) = RShift(LBA, 16) And &HFF
    cmd(4) = RShift(LBA, 8) And &HFF
    cmd(5) = LBA And &HFF                       ' LSB of LBA
    cmd(6) = RShift(NumSectors, 16) And &HFF    ' MSB of number of sectors
    cmd(7) = RShift(NumSectors, 8) And &HFF
    cmd(8) = NumSectors And &HFF                ' LSB of number of sectors
    cmd(9) = ReadFlags                          ' read flags
    cmd(10) = SubchBits                         ' Sub-Channel flags

    'execute
    ReadCD = ExecCMD(DrvID, cmd, 12, 0, False, SCSI_DIR_IN, PtrBuf, BufLen + 1, 10)
End Function

'**************************************
'** Function: ReadTOC
'** Returns: Boolean
'** Usage: reads simple TOC, raw TOC, PMA, ATIP and CD-Text
'**************************************
Public Function ReadTOC(ByVal DrvID As String, ByVal TOC_Format As Integer, _
                        ByVal MSF As Boolean, ByVal Track_Session As Integer, _
                        ByVal PtrBuffer As Long, ByVal BufferLen As Long) As Boolean
    Dim cmd(9) As Byte

    cmd(0) = SCSI_READ_TOC                  ' READ TOC command [43h]
    cmd(1) = IIf(MSF, &H2, 0)               ' MSF or LBA?
    cmd(2) = TOC_Format                     ' requested format (TOC, PMA, ATIP, CD-Text)
    cmd(6) = Track_Session                  ' specific track/session
    cmd(7) = BufferLen \ &HFF               ' allocation length
    cmd(8) = BufferLen And &HFF             ' allocation length

    'execute
    ReadTOC = ExecCMD(DrvID, cmd, 10, 0, True, SCSI_DIR_IN, PtrBuffer, BufferLen)
End Function

'**************************************
'** Function: SetCDSpeed
'** Returns: TRUE if it could set the new speeds
'** Usage: sets the reading and writing speed.
'**        0 sets the values to its maximum.
'**        If a value isn't supported (e.g. writing),
'**        it will be ignored.
'**************************************
Public Function SetCDSpeed(ByVal DrvID As String, ByVal NewReadSpeed As Long, _
                  ByVal NewWriteSpeed As Long, ByVal CLV As Boolean) As Boolean

    Dim cmd(11) As Byte
    cmd(0) = SCSI_SETSPEED                  ' SET CD SPEED command [BBh]
    cmd(1) = IIf(CLV, 0, 1)                 ' CLV?

    If NewReadSpeed > 0 Then                ' New Read Speed is greater than zero?
        cmd(2) = NewReadSpeed \ &HFF        ' NewReadSpeed MSB
        cmd(3) = NewReadSpeed Mod &HFF      ' NewReadSpeed LSB
    Else                                    ' New read speed is zero
        cmd(2) = &HFF                       ' set the MSB to its max.
        cmd(3) = &HFF                       ' set the LSB to its max.
    End If

    If NewWriteSpeed > 0 Then               ' New Write Speed is greater than zero?
        cmd(4) = NewWriteSpeed \ &HFF       ' NewWriteSpeed MSB
        cmd(5) = NewWriteSpeed Mod &HFF     ' NewWriteSpeed LSB
    Else                                    ' New write speed is zero
        cmd(4) = &HFF                       ' set the MSB to its max.
        cmd(5) = &HFF                       ' set the LSB to its max.
    End If

    'execute
    SetCDSpeed = ExecCMD(DrvID, cmd, 12, 0, False, SCSI_DIR_OUT, 0, 0)
End Function

'**************************************
'** Function: ReadDiscInformation
'** Returns: TRUE if it could read disc information
'** Usage: reads various information about an inserted disc
'**************************************
Private Function ReadDiscInformation(ByVal DrvID As String, ByVal PtrBuffer As Long, _
                                                ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_READ_DISC_NFO ' READ DISC INFORMATION command [51h]
    cmd(7) = BufferLen \ &HFF   ' Allocation length
    cmd(8) = BufferLen Mod &HFF ' Allocation length

    'execute
    ReadDiscInformation = ExecCMD( _
                            DrvID, cmd, 10, 0, False, SCSI_DIR_IN, PtrBuffer, BufferLen _
                          )
End Function

'**************************************
'** Function: ReadCapacity
'** Returns: TRUE if it could read the disc's capacity
'** Usage: reads number of written sectors
'**************************************
Public Function ReadCapacity(ByVal DrvID As String, ByVal PtrBuffer As Long, _
                                            ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_READ_CAP  ' READ CAPACITY command [25h]

    'execute
    ReadCapacity = ExecCMD( _
                        DrvID, cmd, 10, 0, False, SCSI_DIR_IN, PtrBuffer, BufferLen _
                   )
End Function

'**************************************
'** Function: CDDB_Disc_ID
'** Returns: String [8 chars]
'** Usage: generates a disc identifier (almost unique for every disc),
'**        compatible to the CDDB and FreeDB
'** Code: Parts from VB-Accelerator, got the main function from FreeDB
'**************************************
Public Function CDDB_Disc_ID(ByVal DrvID As String) As String
    Dim tTOC As t_STOC_STRUCT
    Dim i As Long, t As Long, n As Long

    InitShifting    ' Init VB-Accelerator bit shifting routines

    'Read the simple TOC
    ReadTOC DrvID, 0, True, 0, VarPtr(tTOC), Len(tTOC)

    'generate a unsigned sum of all tracks
    For i = 0 To tTOC.LastTrack - 1
        n = UnsignedAdd( _
                n, CDDB_Sum( _
                    (tTOC.TocTrack(i).addr(1) * 60) + tTOC.TocTrack(i).addr(2) _
                   ) _
            )
    Next i

    'get the size of the disc
    t = ((tTOC.TocTrack(tTOC.LastTrack).addr(1) * 60) + tTOC.TocTrack(tTOC.LastTrack).addr(2)) _
         - ((tTOC.TocTrack(0).addr(1) * 60) + tTOC.TocTrack(0).addr(2))

    'generate the CDDB ID
    CDDB_Disc_ID = lPAD(Hex(RShiftAcc((n Mod &HFF), 24) Or RShiftAcc(t, 8) Or tTOC.LastTrack), 8, 0)
End Function

'**************************************
'** Function: CDDB_Sum
'** Returns: Integer
'** Usage: sub-function of CDDB_Disc_ID, calculates a sum from a value...?
'** Code: Parts from VB-Accelerator, got the main function from FreeDB
'**************************************
Private Function CDDB_Sum(ByVal n As Integer) As Integer
    Dim ret As Long
    ret = 0

    While n > 0
        ret = UnsignedAdd(ret, n Mod 10)
        n = n \ 10
    Wend

    CDDB_Sum = ret
End Function

'**************************************
'** Function: UnsignedAdd
'** Returns: Long
'** Usage: adds 2 unsigned values
'** Code: from VB-Accelerator
'**************************************
Private Function UnsignedAdd(start As Long, Incr As Long) As Long
   If start And &H80000000 Then
      UnsignedAdd = start + Incr
   ElseIf (start Or &H80000000) < -Incr Then
      UnsignedAdd = start + Incr
   Else
      UnsignedAdd = (start + &H80000000) + (Incr + &H80000000)
   End If
End Function

'**************************************
'** Function: lPAD
'** Returns: String [undefined length]
'** Usage: padds a string with x characters to the left
'**************************************
Private Function lPAD(ByVal str As String, ByVal str_len As Integer, ByVal char As String) As String
    If Not (Len(str) = str_len Or Len(str) > str_len) Then
        lPAD = String(str_len - Len(str), char) & str
    Else
        lPAD = str
    End If
End Function

'**************************************
'** Function: RShiftAcc
'** Returns: Long
'** Usage: shifts bits of a value to the right (unsigned)
'** Code: from VB-Accelerator
'**************************************
Private Function RShiftAcc(ByVal lThis As Long, ByVal lBits As Long) As Long
   If (lBits <= 0) Then
      RShiftAcc = lThis
   ElseIf (lBits > 63) Then
      Exit Function
   ElseIf (lBits > 31) Then
      RShiftAcc = 0
   Else
      If (lThis And m_lPower2(31 - lBits)) = m_lPower2(31 - lBits) Then
         RShiftAcc = (lThis And (m_lPower2(31 - lBits) - 1)) * m_lPower2(lBits) Or m_lPower2(31)
      Else
         RShiftAcc = (lThis And (m_lPower2(31 - lBits) - 1)) * m_lPower2(lBits)
      End If
   End If
End Function

'**************************************
'** Function: InitShifting
'** Returns: void
'** Usage: needed values for unsigned bit shifting
'** Code: from VB-Accelerator
'**************************************
Private Sub InitShifting()
   m_lPower2(0) = &H1&: m_lPower2(1) = &H2&: m_lPower2(2) = &H4&
   m_lPower2(3) = &H8&: m_lPower2(4) = &H10&: m_lPower2(5) = &H20&
   m_lPower2(6) = &H40&: m_lPower2(7) = &H80&: m_lPower2(8) = &H100&
   m_lPower2(9) = &H200&: m_lPower2(10) = &H400&: m_lPower2(11) = &H800&
   m_lPower2(12) = &H1000&: m_lPower2(13) = &H2000&: m_lPower2(14) = &H4000&
   m_lPower2(15) = &H8000&: m_lPower2(16) = &H10000: m_lPower2(17) = &H20000
   m_lPower2(18) = &H40000: m_lPower2(19) = &H80000: m_lPower2(20) = &H100000
   m_lPower2(21) = &H200000: m_lPower2(22) = &H400000: m_lPower2(23) = &H800000
   m_lPower2(24) = &H1000000: m_lPower2(25) = &H2000000: m_lPower2(26) = &H4000000
   m_lPower2(27) = &H8000000: m_lPower2(28) = &H10000000: m_lPower2(29) = &H20000000
   m_lPower2(30) = &H40000000: m_lPower2(31) = &H80000000
End Sub

'**************************************
'** Function: ReadCDText
'** Returns: String Array if CD-Text could be read
'**          FALSE if CD-Text couldn't be read
'** Usage: reads CD-Text and returns it
'**        reads artist, album and track names
'**        there is quiet more information that can be stored
'**        in CD-Text, but we will only parse the basic stuff...
'**************************************
Public Function ReadCDText(ByVal DrvID As String) As Variant
    On Error Resume Next
    Dim artistname As String, albname As String         ' artist & album
    Dim albyear As String, albgenre As String           ' year and genre
    Dim albnotes As String                              ' album notes
    Dim trkindex As Integer, trkname As String          ' current track index & name
    Dim trknames(99) As String, RetArr() As String      ' name buffer
    Dim j As Integer, i As Integer                      ' counters
    Dim albnameready As Boolean, IsEmpty As Boolean     '
    Dim tCDT As t_CD_TEXT                               ' CD-Text container

    'try to read CD-Text
    If Not ReadTOC(DrvID, &H5, True, 0, VarPtr(tCDT), Len(tCDT)) Then _
        Exit Function

    'go through very packet
    For j = 0 To 255
        If tCDT.CDTEXT(j).idSeq <> j Then Exit For

            'language is right?
            If ((tCDT.CDTEXT(j).idFlg And &H30) = 0) Then

                'detect the type of the packet
                Select Case tCDT.CDTEXT(j).idType

                    'album or track name
                    Case &H80
                        If (tCDT.CDTEXT(j).idTrk = 0) Then
                            For i = 0 To 11
                                If (tCDT.CDTEXT(j).txt(i) <> 0) Then
                                  If albnameready = False Then
                                    'album name
                                    albname = albname & ChrW$(tCDT.CDTEXT(j).txt(i))
                                  Else
                                    'track name
                                    trkname = trkname & ChrW$(tCDT.CDTEXT(j).txt(i))
                                  End If
                                Else
                                  Trim$ (albname)
                                  albnameready = True
                                End If
                            Next i

                        Else
                            'track name
                            For i = 0 To 11
                                If (tCDT.CDTEXT(j).txt(i) <> 0) Then
                                  trkname = trkname & ChrW$(tCDT.CDTEXT(j).txt(i))
                                Else
                                  Trim$ (trkname)
                                  trknames(trkindex) = trkname
                                  trkname = vbNullString
                                  trkindex = trkindex + 1
                                  If (tCDT.CDTEXT(j).txt(i + 1) = 0) Then Exit For
                                End If
                            Next
                        End If

                    'artist name
                    Case &H81
                        If (tCDT.CDTEXT(j).idTrk = 0) Then
                            For i = 0 To 11
                                If (tCDT.CDTEXT(j).txt(i) <> 0) Then
                                  artistname = artistname & ChrW$(tCDT.CDTEXT(j).txt(i))
                                Else
                                  Trim$ (artistname)
                                  Exit For
                                End If
                            Next
                        End If

                    'album notes
                    Case &H85
                        If (tCDT.CDTEXT(j).idTrk = 0) Then
                            For i = 0 To 11
                                   If (tCDT.CDTEXT(j).txt(i) <> 0) Then
                                    albnotes = albnotes & ChrW$(tCDT.CDTEXT(j).txt(i))
                                Else
                                    Trim$ (albnotes)
                                    Exit For
                                End If
                            Next
                        End If

                    'album genre
                    Case &H87
                        If (tCDT.CDTEXT(j).idTrk = 0) Then
                            For i = 0 To 11
                                albgenre = albgenre & ChrW$(tCDT.CDTEXT(j).txt(i))
                            Next
                        End If

                End Select
            End If
    Next j

    RemNulls albgenre
    If Len(albgenre) > 0 Then
        albgenre = Asc(Left(albgenre, 1))
    End If

    'check if CD-Text data is empty
    IsEmpty = True
    If artistname = vbNullString And albname = vbNullString Then

        'check every track
        For i = LBound(trknames) To UBound(trknames)

            'if current track name isn't empty, then there exists some data, w00t!
            If Not trknames(i) = vbNullString Then
                IsEmpty = False
                GoTo NextStep
            End If

        Next i

    Else
        'if artist and album are existing, there is data!
        IsEmpty = False
    End If

NextStep:
    If IsEmpty Then ReadCDText = False: Exit Function

    'prepare an array to return it
    ReDim RetArr(0 To UBound(trknames) + 2)
    RetArr(0) = albname
    RetArr(1) = artistname
    RetArr(2) = albgenre
    For i = LBound(trknames) To UBound(trknames)
        RetArr(i + 3) = Trim$(trknames(i))
    Next

    'return the array
    ReadCDText = RetArr
End Function

'**************************************
'** Function: GetHostAdapters
'** Returns: Long
'**          > -1: number of Host Adapters
'**          = -1: ASPI error
'** Usage: gets the number of host adapters and checks the ASPI status
'**************************************
Public Function GetHostAdapters() As Long
    Dim nRet As Long
    Dim sts As Long, cnt As Long, i As Long, fh As Long
    Dim buffer(255) As Byte
    Dim pscsiAddr As t_SCSI_ADDRESS
    Dim OL As OVERLAPPED

    If bUseASPI Then

        nRet = GetASPI32SupportInfo()   ' get ASPI status and number of HAs
        sts = nRet \ &HFF               ' extract the status
        cnt = nRet And &HFF             ' extract the number of HAs

        If sts = SS_COMP Then           ' if the ASPI status is SS_COMP,
            GetHostAdapters = cnt       ' return the number of HAs,
        Else
            GetHostAdapters = -1        ' if not, return -1.
        End If

    Else

        For i = 1 To 24

            If GetDriveType(ChrW$(i + 64) & ":") <> 5 Then GoTo Nextdrv

            'get the current drive's handle
            If GetDriveHandle(ChrW$(i + 64), fh) Then

                'ask it for its SCSI address
                sts = DeviceIoControl(fh, IOCTL_SCSI_GET_ADDRESS, _
                                      0, 0, pscsiAddr, Len(pscsiAddr), _
                                      nRet, OL)

                'close that handle!!
                CloseHandle fh

                'success?
                If sts = 1 Then
                    'add the HA
                    buffer(pscsiAddr.PortNumber) = 1
                End If

            End If

Nextdrv:
        Next

        'return the number of HAs found
        For i = 0 To UBound(buffer)
            If buffer(i) Then GetHostAdapters = GetHostAdapters + 1
        Next

    End If
End Function

'**************************************
'** Function: GetConfiguration
'** Returns: Boolean
'** Usage: reads features of a drive
'**************************************
Public Function GetConfiguration(ByVal DrvID As String, ByVal StartFeature As Byte, ByVal RT As Byte, _
                                 ByVal PtrBuffer As Long, ByVal BufLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_GET_CONFIG        ' GET CONFIGURATION command [46h]
    cmd(1) = RT                     ' RT byte
    cmd(2) = StartFeature \ &HFF    ' Feature to start with
    cmd(3) = StartFeature And &HFF  '    "     "   "    "
    cmd(7) = BufLen \ &HFF          ' allocation length
    cmd(8) = BufLen And &HFF        '    "         "

    'execute
    GetConfiguration = ExecCMD(DrvID, cmd, 10, 0, False, SCSI_DIR_IN, PtrBuffer, BufLen)
End Function

'**************************************
'** Function: GetInterface
'** Returns: Boolean
'** Usage: trys to get the phyiscal interface of a drive
'**************************************
Public Function GetInterface(ByVal DrvID As String) As e_Interfaces
    Dim buffer(15) As Byte
    Dim cmd(9) As Byte
    Dim inquiry As t_InqDat

    'try to read the core feature of the drive
    GetConfiguration DrvID, 1, 2, VarPtr(buffer(0)), UBound(buffer)

    'determine the interface
    Select Case LShift(buffer(12), 24) Or LShift(buffer(13), 16) Or _
                LShift(buffer(14), 8) Or buffer(15)

        Case 1: GetInterface = IF_SCSI
        Case 2: GetInterface = IF_ATAPI
        Case 3 Or 4: GetInterface = IF_IEEE
        Case Else: GetInterface = IF_UNKNWN

    End Select

    If GetInterface = IF_UNKNWN Then
        cmd(0) = SCSI_INQUIRY
        cmd(4) = Len(inquiry)
        If ExecCMD(DrvID, cmd, 10, 0, False, SCSI_DIR_IN, VarPtr(inquiry), Len(inquiry)) Then
            If inquiry.rsv1(0) = 0 Then GetInterface = IF_ATAPI
        End If
    End If
End Function

'**************************************
'** Function: WriteParams
'** Returns: Boolean
'** Usage: sets the write parameter page of a CD-writer
'**************************************
Public Function WriteParams(ByVal DrvID As String, ByVal TestMode As Boolean, _
                             ByVal BURNProof As Boolean, ByVal TrackPause As Long, _
                             ByVal WriteType As Byte, ByVal TrackMode As Byte, _
                             ByVal DataBlockType As Byte, ByVal MultiSession As Boolean _
                           ) As Boolean

    Dim bufData(60) As Byte
    Dim PS As Byte
    Dim i As Integer

    'put the page into the buffer
    ModeSense10 DrvID, &H5, VarPtr(bufData(0)), UBound(bufData)

    bufData(1) = 58                         ' buffer length

    bufData(8) = 5                          ' Page Code
    bufData(9) = 50                         ' Page Length

    'WriteType, Test-Mode and Burn Proof
    bufData(10) = WriteType Or IIf(TestMode, &H10, 0) Or IIf(BURNProof, &H40, 0)

    bufData(11) = TrackMode Or IIf(MultiSession, &HC0, 0) 'Track-Mode, Multi-Session
    bufData(12) = DataBlockType                           'Data-Mode (Mode-1: 2048 Bytes User-Data)

    'Track-Pause-length: Min 150 sectors = 2 secs
    bufData(22) = RShift(TrackPause, 8) And &HFF ' Track-Pause-Length MSB
    bufData(23) = TrackPause And &HFF            ' Track-Pause-Length (150 = 2x75 = 2 Sekunden Pause) LSB

    'set the new WMP
    WriteParams = ModeSelect10(DrvID, VarPtr(bufData(0)), 60)
End Function

'**************************************
'** Function: ModeSelect10
'** Returns: Boolean
'** Usage: sets a mode page
'**************************************
Public Function ModeSelect10(ByVal DrvID As String, ByVal PtrBuffer As Long, _
                             ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_MODE_SEL10        ' MODE SELECT10 command [55h]
    cmd(1) = &H10                   ' PF = 1 (Page Format)
    cmd(7) = BufferLen \ &HFF       ' allocation length
    cmd(8) = BufferLen Mod &HFF     ' allocation length

    'execute
    ModeSelect10 = ExecCMD( _
                        DrvID, cmd, 10, 0, True, SCSI_DIR_OUT, PtrBuffer, BufferLen _
                   )
End Function

'**************************************
'** Function: ModeSense10
'** Returns: TRUE if it could read the mode page
'** Usage: reads mode pages to a buffer
'**************************************
Public Function ModeSense10(ByVal DrvID As String, ByVal MP As Byte, _
          ByVal PtrBuffer As Long, ByVal BufferLen As Long) As Boolean

    Dim cmd(9) As Byte
    cmd(0) = SCSI_MODE_SEN10    ' MODE SENSE 10 command [5Ah]
    cmd(2) = MP And &H3F        ' Mode Page
    cmd(8) = BufferLen          ' allocation length

    'execute
    ModeSense10 = ExecCMD( _
                    DrvID, cmd, 10, 0, False, SCSI_DIR_IN, PtrBuffer, BufferLen _
                  )
End Function

'**************************************
'** Function: InquiryDrive
'** Returns: String [undefined length]
'** Usage: requests information about drive like vendor, product and revision
'**************************************
Public Function InquiryDrive(ByVal DriveID As String, ByVal method As e_InquiryMethod) As String
    Dim q As Integer
    Dim cmd(5) As Byte
    Dim inquiryData As t_InqDat
    Dim InqBuf As String

    cmd(0) = SCSI_INQUIRY   ' Inquiry command [12h]
    cmd(4) = &H56           ' allocation length

    'execute the command
    If ExecCMD(DriveID, cmd, 6, 0, False, SCSI_DIR_IN, VarPtr(inquiryData), &H56) Then

        'return the drive's vendor
        If method = INQ_VENDOR Then
            For q = 0 To 7
                InqBuf = InqBuf & ChrW$(inquiryData.VID(q))
            Next q
            InquiryDrive = Trim$(InqBuf)

        'return the drive's name
        ElseIf method = INQ_NAME Then
            For q = 0 To 15
                InqBuf = InqBuf & ChrW$(inquiryData.PID(q))
            Next q
            InquiryDrive = Trim$(InqBuf)

        'return the drive's revision (= firmware version)
        ElseIf method = INQ_FIRMWARE Then
            For q = 0 To 3
                InqBuf = InqBuf & ChrW$(inquiryData.PVER(q))
            Next q
            InquiryDrive = Trim$(InqBuf)

        End If

    Else
        InquiryDrive = "Error"      'inquiry didn't work...
    End If
End Function
