Attribute VB_Name = "modSCSI"
Option Explicit

Public drives As New cDrives

'ASPI Function to get the ASPI Status and the number of installed host adapters
Public Declare Function GetASPI32SupportInfo Lib "SCSI" _
    Alias "GetASPI32SupportInfoEx" () As Long

'ASPI Function to execute SCSI commands or little ASPI helpers
Public Declare Function SendASPI32CommandEx Lib "SCSI" ( _
    hSRB As Any _
) As Long

Public Const SC_HA_INQUIRY As Byte = &H0       ' ASPI command to Inquiry a Host Adapter
Public Const SC_GET_DEV_TYPE As Byte = &H1     ' ASPI command to get the type of a installed device
Public Const SC_EXEC_SCSI_CMD As Byte = &H2    ' ASPI command to execute a SCSI command
Public Const SC_GET_DISK_INFO As Byte = &H6    ' ASPI command to get information about a disk

Public Const SS_PENDING As Byte = &H0          ' ASPI command status [pending]
Public Const SS_COMP As Byte = &H1             ' ASPI command status [completed]
Public Const SS_ABORTED As Byte = &H2          ' SRB aborted
Public Const SS_ABORT_FAIL As Byte = &H3       ' Unable to abort SRB
Public Const SS_ERR As Byte = &H4              ' SRB completed with error
Public Const SS_INVALID_CMD As Byte = &H80     ' Invalid ASPI command
Public Const SS_INVALID_HA As Byte = &H81      ' Invalid host adapter number
Public Const SS_NO_DEVICE As Byte = &H82       ' SCSI device not installed
Public Const SS_INVALID_SRB As Byte = &HE0     ' Invalid parameter set in SRB
Public Const SS_OLD_MANAGER As Byte = &HE1     ' ASPI manager doesn't support windows
Public Const SS_BUFFER_ALIGN As Byte = &HE1    ' Buffer not aligned (SS_OLD_MANAGER in Win32)
Public Const SS_ILLEGAL_MODE As Byte = &HE2    ' Unsupported Windows mode
Public Const SS_NO_ASPI As Byte = &HE3         ' No ASPI managers
Public Const SS_FAILED_INIT As Byte = &HE4     ' ASPI for windows failed init
Public Const SS_ASPI_IS_BUSY As Byte = &HE5    ' No resources available to execute command
Public Const SS_BUFFER_TOO_BIG As Byte = &HE6  ' Buffer size too big to handle
Public Const SS_MISMATCH_FILES As Byte = &HE7  ' The DLLs/EXEs of ASPI don't version check
Public Const SS_NO_ADAPTERS As Byte = &HE8     ' No host adapters located
Public Const SS_SHORT_RESOURCES As Byte = &HE9 ' Couldn't allocate resources  needed to init
Public Const SS_ASPI_IS_SHUTDOWN As Byte = &HEA ' Call came to ASPI after PROCESS_DETACH
Public Const SS_BAD_INSTALL As Byte = &HEB     ' The DLL or other components are installed wrong

Public Const HASTAT_OK As Byte = &H0           ' Host adapter did not detect an error.
Public Const HASTAT_TIMEOUT As Byte = &H9      ' Timed out while SRB was waiting to be processed.
Public Const HASTAT_CMD_TIMEOUT As Byte = &HB  ' While processing the SRB, adapter timed out.
Public Const HASTAT_BUS_RESET As Byte = &HE    ' A bus reset was detected.
Public Const HASTAT_PARITY_ERROR As Byte = &HF ' A parity error was detected.
Public Const HASTAT_SEL_TO As Byte = &H11      ' Selection Timeout.
Public Const HASTAT_DO_DU As Byte = &H12       ' Data overrun / data underrun.
Public Const HASTAT_BUS_FREE As Byte = &H13    ' Unexpected bus free.
Public Const HASTAT_PHASE_ERR As Byte = &H14   ' Target bus phase sequence failure.

Public Const STATUS_GOOD As Byte = &H0         ' Status Good.
Public Const STATUS_CHKCOND As Byte = &H2      ' Check Condition.
Public Const STATUS_CONDMET As Byte = &H4      ' Condition Met.
Public Const STATUS_BUSY As Byte = &H8         ' Busy.
Public Const STATUS_INTERM As Byte = &H10      ' Intermediate.
Public Const STATUS_INTCDMET As Byte = &H14    ' Intermediate-condition met.
Public Const STATUS_RESCONF As Byte = &H18     ' Reservation conflict.
Public Const STATUS_CMD_TERM As Byte = &H22    ' Command Terminated.
Public Const STATUS_QFULL As Byte = &H28       ' Queue full.

Public Const KEY_NOSENSE As Byte = &H0         ' No Sense.
Public Const KEY_RECERROR As Byte = &H1        ' Recovered Error.
Public Const KEY_NOTREADY As Byte = &H2        ' Not Ready.
Public Const KEY_MEDIUMERROR As Byte = &H3     ' Medium Error.
Public Const KEY_HARDERROR As Byte = &H4       ' Hardware Error.
Public Const KEY_ILLGLREQ As Byte = &H5        ' Illegal Request.
Public Const KEY_UNITATT As Byte = &H6         ' Unit Attention.
Public Const KEY_DATAPROT As Byte = &H7        ' Data Protection.
Public Const KEY_BLANKCHK As Byte = &H8        ' Blank Check.
Public Const KEY_VENDSPEC As Byte = &H9        ' Vendor Specific.
Public Const KEY_COPYABORT As Byte = &HA       ' Copy Aborted.
Public Const KEY_ABORTCMD As Byte = &HB        ' Aborted Command.
Public Const KEY_EQUAL As Byte = &HC           ' Equal (Search).
Public Const KEY_VOLOVRFLW As Byte = &HD       ' Volume Overflow.
Public Const KEY_MISCOMP As Byte = &HE         ' Miscompare (Search).
Public Const KEY_RSVD As Byte = &HF            ' Reserved.

Public Const SENSE_LEN As Byte = 14            ' Standard sense length
Public Const SRB_DIR_IN As Byte = &H8          ' get data from the drive
Public Const SRB_DIR_OUT As Byte = &H10        ' send data to the drive
Public Const SRB_POSTING As Byte = &H1         ' post a message when the command is completed
Public Const SRB_EVENT_NOTIFY As Byte = &H40   ' notify by a event when command is finished

Public Const INVALID_HANDLE_VALUE = -1     ' invalid file handle
Public Const OPEN_EXISTING = 3             ' Flag: Open existing files
Public Const GENERIC_READ = &H80000000     ' Flag: generic read rights
Public Const GENERIC_WRITE = &H40000000    ' Flag: generic write rights
Public Const FILE_READ_ACCESS = &H1        ' Flag: read access
Public Const FILE_WRITE_ACCESS = &H2       ' Flag: write access
Public Const FILE_SHARE_READ = &H1         ' Flag: Shared reading
Public Const FILE_SHARE_WRITE = &H2        ' Flag: Shared writing
Public Const FILE_ATTRIBUTE_NORMAL = &H80  ' Flag: ?

Public Const IOCTL_SCSI_PASS_THROUGH_DIRECT = &H4D014   ' DeviceIoControl NT command
                                                        ' to execute a SCSI command

Public Const IOCTL_SCSI_GET_ADDRESS = &H41018  ' DeviceIoControl NT command to get
                                               ' the SCSI address of a device

Public Const SCSI_IOCTL_DATA_OUT = 0                ' send data to the drive
Public Const SCSI_IOCTL_DATA_IN = 1                 ' recieve data from the drive
Public Const SCSI_IOCTL_DATA_UNSPECIFIED = 2        ' undefined data direction

'Windows function to create/open a file and get its handle
Public Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" ( _
    ByVal lpFileName As String, ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, lpSecurityAttributes As Any, _
    ByVal dwCreationDisposition As Long, ByVal dwFlagsAndAttributes As Long, _
    hTemplateFile As Long _
) As Long

'Windows function to execute hardware commands
Public Declare Function DeviceIoControl Lib "kernel32" ( _
    ByVal hDevice As Long, ByVal dwIoControlCode As Long, _
    ByRef lpInBuffer As Any, ByVal nInBufferSize As Long, _
    ByRef lpOutBuffer As Any, ByVal nOutBufferSize As Long, _
    ByRef lpBytesReturned As Long, ByRef lpOverlapped As OVERLAPPED _
) As Long

'Windows function to create an event
Public Declare Function CreateEvent Lib "kernel32" Alias "CreateEventA" ( _
        lpEventAttributes As SECURITY_ATTRIBUTES, _
        ByVal bManualReset As Long, _
        ByVal bInitialState As Long, _
        ByVal lpName As String _
) As Long

'Windows function to wait for an object, e.g. an event
Public Declare Function WaitForSingleObject Lib "kernel32" ( _
        ByVal hHandle As Long, _
        ByVal dwMilliseconds As Long _
) As Long

'Windows function to reset an event
Public Declare Function ResetEvent Lib "kernel32" ( _
        ByVal hEvent As Long _
) As Long

'WIndows function to close a specific handle
Public Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long _
) As Long

Public bUseASPI As Boolean  'Use ASPI or SPTI?
Public bW2K As Boolean      'Is Win2k?

'**************************************
'** Function: GetDriveHandle
'** Returns: Boolean
'** Usage: trys to get the handle for a drive
'**************************************
Public Function GetDriveHandle(ByVal drv As String, ByRef fh As Long) As Boolean
    Dim sd As SECURITY_ATTRIBUTES
    Dim flags As Long

    flags = GENERIC_READ
    If bW2K Then flags = flags Or GENERIC_WRITE

    'get the drive handle
    fh = CreateFile("\\.\" & drv & ":", flags, _
                    FILE_SHARE_READ Or FILE_SHARE_WRITE, sd, _
                    OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, ByVal 0)

'    fh = CreateFile("\\.\" & drv & ":", flags, FILE_SHARE_READ, _
'                        sd, OPEN_EXISTING, 0, 0&)

    'did it work or not?
    If fh = INVALID_HANDLE_VALUE Then Exit Function

    GetDriveHandle = True
End Function

'Proxy function for ASPICMD and SPTICMD
Public Function ExecCMD(ByVal DrvID As String, ByRef cmd() As Byte, _
                        ByVal CmdLen As Integer, ByVal flags As Long, _
                        ByVal WFE As Boolean, ByVal Dir As e_Direction, _
                        ByVal Pointer As Long, ByVal PointerLen As Long, _
                        Optional ByVal TimeOut As Long) As Boolean

    If bUseASPI Then

        ExecCMD = ASPICMD( _
                    DrvID, cmd, CmdLen, flags, WFE, _
                    Dir, Pointer, PointerLen _
                  )

    Else

        ExecCMD = SPTICMD( _
                    DrvID, cmd, CmdLen, WFE, Dir, _
                    Pointer, PointerLen, TimeOut _
                  )

    End If

End Function

'**************************************
'** Function: ExtractHAID
'** Returns: Integer
'** Usage: Extracts the Host Adapter [Extrkt = 0] or the target [Extrkt = 1] identifier
'**        from an ASPI drive identifier
'**************************************
Public Function ExtractHAID(ByVal DrvID As String, ByVal Extrkt As Integer) As Integer
    On Error Resume Next
    Dim splt() As String
    splt = Split(DrvID, "0000")
    ExtractHAID = CInt(IIf(Extrkt = 0, CInt(IIf(splt(0) = "", 0, splt(0))), _
                  CInt(IIf(splt(1) = "", 0, splt(1)))))
End Function

'**************************************
'** Function: ASPICMD
'** Returns: Boolean
'** Usage: Executes an SCSI command through the ASPI
'**************************************
Private Function ASPICMD(ByVal DrvID As String, ByRef cmd() As Byte, _
                         ByVal CmdLen As Integer, ByVal flags As Long, _
                         ByVal WFE As Boolean, ByVal Dir As e_Direction, _
                         ByVal Pointer As Long, ByVal PointerLen As Long, _
                         Optional ByVal TimeOut As Long = 5) As Boolean

    Dim hEvent As Long, i As Long
    Dim ExecIO As SRB_ExecuteIO
    Dim sd As SECURITY_ATTRIBUTES

    With sd
       .nLength = Len(sd)
       .lpSecurityDescriptor = 0
       .bInheritHandle = 0
    End With
    hEvent = CreateEvent(sd, True, False, 0)    ' create an event to wait for (if WFE = False)

    ExecIO.SRB_Cmd = SC_EXEC_SCSI_CMD           ' Execute SCSI command
    ExecIO.SRB_HaID = ExtractHAID(DrvID, 0)     ' get the host adapter
    ExecIO.SRB_Target = ExtractHAID(DrvID, 1)   ' get the target

    ExecIO.SRB_SenseLen = SENSE_LEN             ' Sense length

    If Dir = SCSI_DIR_IN Then
        ExecIO.SRB_Flags = SRB_DIR_IN           ' get data from the drive
    Else
        ExecIO.SRB_Flags = SRB_DIR_OUT          ' send data to the drive
    End If
    ExecIO.SRB_Flags = ExecIO.SRB_Flags Or SRB_EVENT_NOTIFY Or flags    ' set flags

    ExecIO.SRB_PostProc = hEvent                ' post the process to the created event
    ExecIO.SRB_BufPointer = Pointer             ' pointer to the buffer
    ExecIO.SRB_BufLen = PointerLen              ' buffer length

    For i = 0 To CmdLen - 1
        ExecIO.SRB_CDBByte(i) = cmd(i)          ' CDB (Command Descriptor Block)
    Next i
    ExecIO.SRB_CDBLen = CmdLen                  ' set the Command Length

    ResetEvent (hEvent)                         ' reset the created event

    SendASPI32CommandEx ExecIO                  ' execute the command

    If WFE = True Then
        While ExecIO.SRB_Status = SS_PENDING    ' if WFE, wait until the command is completed
            DoEvents
        Wend
    Else
        WaitForSingleObject hEvent, TimeOut * 1000 'if not, wait X seconds for the event
    End If
    CloseHandle (hEvent)                        ' close the handle to the event

    If (ExecIO.SRB_Status <> SS_COMP) Then      ' check the command status
        ASPICMD = False                         ' didn't work...
        Exit Function
    End If

    ASPICMD = True                              ' worked.
End Function

'*********************************
'** Function: SPTICMD
'** Returns: Boolean
'** Usage: executes a command over the SPTI
'**        (SCSI Pass-Through Interface).
'*********************************
Private Function SPTICMD(ByVal DrvID As String, ByRef cmd() As Byte, _
                         ByVal CmdLen As Integer, ByVal WFE As Boolean, _
                         ByVal Dir As e_Direction, ByVal Pointer As Long, _
                         ByVal PointerLen As Long, ByVal TimeOut As Long _
                        ) As Boolean

    Dim status As Long
    Dim hf As Long
    Dim pswb As SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER
    Dim OL As OVERLAPPED
    Dim length As Long, returned As Long
    Dim i As Long

    If Not GetDriveHandle(DrvID, hf) Then Exit Function ' Read-/Write-Acces to the drive

    If Dir = SCSI_DIR_IN Then
        pswb.spt.DataIn = SCSI_IOCTL_DATA_IN            ' read data
    Else
        pswb.spt.DataIn = SCSI_IOCTL_DATA_OUT           ' send data
    End If

    If TimeOut = 0 Then TimeOut = 5
    If WFE Then TimeOut = 99999

    pswb.spt.length = 44                                     ' length
    pswb.spt.TargetId = 0                                    ' Target ID
    pswb.spt.PathId = 0                                      ' Host Adapter
    pswb.spt.Lun = 0                                         ' Logical Unit Number
    pswb.spt.SenseInfoLength = 14                            ' Sense Info length
    pswb.spt.TimeOutValue = TimeOut                          ' Timeout Interval
    pswb.spt.DataBuffer = Pointer                            ' Buffer Pointer
    pswb.spt.DataTransferLength = PointerLen                 ' Datatransfer length
    pswb.spt.SenseInfoOffset = Len(pswb.spt) + Len(pswb.Fill) ' ucSenseBuf Offset

    For i = 0 To CmdLen - 1
        pswb.spt.Cdb(i) = cmd(i)                        ' feed it with the CDB
    Next i
    pswb.spt.CdbLength = CmdLen                         ' CDB length

    length = Len(pswb)
        status = DeviceIoControl(hf, _
                                 IOCTL_SCSI_PASS_THROUGH_DIRECT, _
                                 pswb, _
                                 length, _
                                 pswb, _
                                 length, _
                                 returned, _
                                 OL)                      ' execute the SRB

    If status = 1 And pswb.spt.ScsiStatus = 0 Then
        SPTICMD = True
    End If

    CloseHandle hf                                      ' close that drive handle!
End Function
