Attribute VB_Name = "modStructs"
Option Explicit


'Structure of HAInquiry (Host Adapter Inquiry) to get information
'about a host adapter
Public Type SRB_HAInquiry
    SRB_Cmd As Byte             ' ASPI command
    SRB_Status As Byte          ' SRB (SCSI-Request-Block) Status
    SRB_HaID As Byte            ' Host Adapter Identifier
    SRB_Flags As Byte           ' command flags
    SRB_Hdr_Rsvd As Long        ' reserved
    HA_Count As Byte            ' number of host adapters
    HA_Id As Byte               ' returned host adapter identifier
    HA_MgrId As String * 16     ' Manager Identifier (e.g. ASPI32)
    HA_Ident As String * 16     ' Host Adapter Identifier (e.g. ATAPI)
    HA_Unique(15) As Byte       ' unique host adapter parameters
    HA_Rsvd As Integer          ' reserved
    HA_Pad(19) As Byte          ' padding
End Type

'Structure of GetDevType (Get Device Type) to get information
'about a installed device
Public Type SRB_GetDevType
    SRB_Cmd As Byte             ' ASPI command
    SRB_Status As Byte          ' command status
    SRB_HaID As Byte            ' Host Adapter
    SRB_Flags As Byte           ' command flags
    SRB_Hdr_Rsvd As Long        ' reserved
    SRB_Target As Byte          ' Target Identifier
    SRB_Lun As Byte             ' Logical Unit Number (not needed)
    DEV_DeviceType As Byte      ' Device Type (5 = CD-Rom)
    DEV_Rsvd1 As Byte           ' reserved
    DEV_Pad(67) As Byte         ' padding
End Type

'Structure of ExecuteIO (Execute Input/Output) to send
'a SCSI/ATAPI command to a device and send/retrieve data
Public Type SRB_ExecuteIO
    SRB_Cmd As Byte             ' ASPI command
    SRB_Status As Byte          ' command status
    SRB_HaID As Byte            ' Host Adapter identifier
    SRB_Flags As Byte           ' command flags
    SRB_Hdr_Rsvd As Long        ' reserved
    SRB_Target As Byte          ' Target identifier
    SRB_Lun As Byte             ' Logical Unit Number (not needed)
    SRB_Rsvd1 As Integer        ' reserved
    SRB_BufLen As Long          ' Buffer Length
    SRB_BufPointer As Long      ' Pointer to the IO buffer
    SRB_SenseLen As Byte        ' Sense Length
    SRB_CDBLen As Byte          ' CDB (Command Descriptor Block) length
    SRB_HaStat As Byte          ' Host Adapter status
    SRB_TargStat As Byte        ' Target status
    SRB_PostProc As Long        ' ?
    SRB_Rsvd2(19) As Byte       ' reserved
    SRB_CDBByte(15) As Byte     ' CDB (Command Descriptor Block)
    SRB_SenseData(15) As Byte   ' Sense Data
End Type

'Structure of GetDiskInfo to get some information on an installed device
'doesn't work on WinNT/2000/XP because Int 13h isn't used anymore
Public Type SRB_GetDiskInfo
    SRB_Cmd As Byte             ' ASPI command
    SRB_Status As Byte          ' command status
    SRB_HaID As Byte            ' Host Adapter identifier
    SRB_Flags As Byte           ' command flags
    SRB_Hdr_Rsvd As Long        ' reserved
    SRB_Target As Byte          ' Target identifier
    SRB_Lun As Byte             ' Logical Unit Number (not needed)
    SRB_DriveFlags As Byte      ' Drive Flags
    SRB_Int13HDriveInfo As Byte ' + 65 = Chr$(drive char)
    SRB_Heads As Byte           ' ?
    SRB_Sectors As Byte         ' ?
    SRB_Rsvd1(9) As Byte        ' reserved
End Type

'Structure of the Inquiry command to get information on an installed CD-ROM device
Public Type t_InqDat
    PDT As Byte         ' Peripheral identifier/Peripheral device type
    PDQ As Byte         ' removable
    VER As Byte         ' MMC Version (for ATAPI 0)
    RDF As Byte         ' interface depenend field/response data format
    DLEN As Byte        ' additional length
    rsv1(1) As Byte     ' reserved
    Feat As Byte        ' ?
    VID(7) As Byte      ' vendor identifier
    PID(15) As Byte     ' product identifier
    PVER(3) As Byte     ' revision identifier (= firmware version)
    FWVER(20) As Byte   ' ?
End Type

'Structure of a raw TOC packet
Public Type t_RTOC_ENTRY
    sessionNr As Byte   ' current session number
    adrCtl As Byte      ' ADR/CTRL
    TNO As Byte         ' Track Number
    point As Byte       ' point identifies the use of the packet
    Min As Byte         ' minutes
    sec As Byte         ' seconds
    frm As Byte         ' frames
    zero As Byte        ' zero
    pmin As Byte        ' point minutes
    psec As Byte        ' point seconds
    pframe As Byte      ' point frames
End Type

'Structure ofthe raw TOC
Public Type t_RTOC_STRUCT
    dummy(3) As Byte                ' header
    packet(255) As t_RTOC_ENTRY     ' packets
End Type

'Structure of a simple TOC packet = track
Public Type t_STOC_TRACK
    rsvd1 As Byte       ' reserved
    ADR As Byte         ' ADR
    Track As Byte       ' Track
    rsvd2 As Byte       ' reserved
    addr(3) As Byte     ' address (h:m:s:f)
End Type

'Structure of the simple TOC
Public Type t_STOC_STRUCT
    TocLen(1) As Byte               ' data length
    FirstTrack As Byte              ' first existing track
    LastTrack As Byte               ' last existing track
    TocTrack(99) As t_STOC_TRACK    ' packets (tracks)
End Type

'Structure of the multimedia capabilities & mechanical status page
Public Type t_MMC
    PageCode As Byte                ' page code = 2Ah
    PageLen As Byte                 ' page length
    rsvd2(7) As Byte                ' reserved
    ReadSupported As Byte           ' read features (CD-R/RW/DVD-R/DVD-ROM/DVD-RAM)
    WriteSupported As Byte          ' write features (CD-R/RW-DVD-R/DVD-RAM)
    misc(3) As Byte                 ' miscellanious stuff
    MaxReadSpeed(1) As Byte         ' maximal read speed
    NumVolLevels(1) As Byte         ' number of volume levels
    BufferSize(1) As Byte           ' buffer size (usually 2048 or 8192 bytes)
    CurrReadSpeed(1) As Byte        ' current read speed
    rsvd As Byte                    ' reserved
    misc2 As Byte                   ' miscellanious stuff
    MaxWriteSpeed(1) As Byte        ' maximal write speed
    CurrWriteSpeed(1) As Byte       ' current write speed
    Rest(20) As Byte                ' others
End Type

'Structure of the data returned by ReadDiscInformation
Public Type t_RDI
    PageLen(1) As Byte                  ' page length
    states As Byte                      ' erasable/disc status/state of last session
    FirstTrack As Byte                  ' first track on disc
    NumSessionsLSB As Byte              ' MSB of number of sessions
    FirstTrackLastSessionLSB As Byte    ' LSB of first track in last session
    LastTrackLastSessionLSB As Byte     ' LSB of last track in last session
    misc As Byte                        ' misc. stuff
    DiscType As Byte                    ' Disc type (CD-ROM/CD-I/XA)
    NumSessionsMSB As Byte              ' MSB of number of sessions
    FirstTrackLastSessionMSB As Byte    ' MSB of first track in last session
    LastTrackLastSessionMSB As Byte     ' MSB of last track in last session
    DiscIdentification(3) As Byte       ' disc identification
    LastSessionLeadInStart(3) As Byte   ' Lead-In Start (h:m:s:f) of the last session
    LastPossibleLeadOutStart(3) As Byte ' Last possbile Lead-Out start (h:m:s:f)
    DBC(6) As Byte                      ' Disc Bar Code
End Type

'Structure of the data returned by ReadTrackInformation
Public Type t_RTI
    DataLen(1) As Byte                  ' data length
    TrackNumLSB As Byte                 ' LSB of the track number
    SessionNumLSB As Byte               ' LSB of the session number
    rsvd As Byte                        ' reserved
    TrackMode As Byte                   ' track mode
    DataMode As Byte                    ' data mode
    misc As Byte                        ' miscellanious stuff
    Track_Start(3) As Byte              ' track start address (h:m:s:f)
    Track_Next_Writable(3) As Byte      ' next writable address (h:m:s:f)
    Track_Free_Blocks(3) As Byte        ' free blocks (h:m:s:f)
    Track_Packet_Size(3) As Byte        ' fixed packet size (?)
    Track_Size(3) As Byte               ' track size (LBA)
    Track_Last_Recorded(3) As Byte      ' last recoreded address in the track
    TrackNumMSB As Byte                 ' MSB of the track number
    SessionNumMSB As Byte               ' MSB of the session number
    rsvd2(1) As Byte                    ' reserved
End Type

'Structure of the data returned by GetCapacity
Public Type t_ReadCap
    Blocks(3) As Byte       ' written sectors
    BlockLen(3) As Byte     ' sector length
End Type

'Structure of the Q Sub-Channel data header returned by ReadSubChannel
Public Type t_SubQHeader
    rsvd As Byte            ' reserved
    audio_stat As Byte      ' audio status (Play, Stop, ...)
    Data_Len(1) As Byte     ' data length
End Type

'Structure of the Q Sub-Channel information (Current Position) returned by ReadSubChannel
Public Type t_CDCurrPos
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    ADRCTRL As Byte         ' ADR/CTRL
    TrkNum As Byte          ' track number
    IndexNum As Byte        ' index number of the current track (0 = Pre-Gap, 1 = data)
    AbsCDAddr(3) As Byte    ' absolute address (position)
    TrkRelAddr(3) As Byte   ' track relative address (position within a track)
End Type

'Structure of the Q Sub-Channel information (ISRC) returned by ReadSubChannel
Public Type t_Q_ISRC
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    ADRCTRL As Byte         ' ADR/CTRL
    TrkNum As Byte          ' track number
    rsvd As Byte            ' reserved
    ISRC(15) As Byte        ' ISRC (International Standard Recording Code)
End Type

'Structure of the Q Sub-Channel information (MCN) returned by ReadSubChannel
Public Type t_Q_MCN
    header As t_SubQHeader  ' header (top)
    dataformat As Byte      ' data format of the requested information
    rsvd(2) As Byte         ' reserved
    MCN(15) As Byte         ' MCN (Media Catalog Number)
End Type

'Structure of the ATIP (Absolute Time In Pre-Groove)
Public Type t_ATIP
    Length(1) As Byte       ' data length
    rsvd1(1) As Byte        ' reserved
    ITWP As Byte            ' ?
    Uru As Byte             ' ?
    DiscType As Byte        ' disc type (CD-R/CD-RW)
    rsvd2 As Byte           ' reserved
    LeadIn_Min As Byte      ' lead-in start (minutes)
    LeadIn_Sec As Byte      ' lead-in start (seconds)
    LeadIn_Frm As Byte      ' lead-in start (frames)
    rsvd3 As Byte           ' reserved
    LeadOut_Min As Byte     ' lead-out start (minutes)
    LeadOut_Sec As Byte     ' lead-out start (seconds)
    LeadOut_Frm As Byte     ' lead-out start (frames)
    rsvd4 As Byte           ' reserved
End Type

'Structure of a CD-Text packet
Public Type t_CD_TEXT_PACKET
    idType As Byte          ' packet type
    idTrk As Byte           ' track number
    idSeq As Byte           ' sequence
    idFlg As Byte           ' flags
    txt(11) As Byte         ' text data (ASCII)
    crc(1) As Byte          ' CRC (Cyclic Redundancy Check)
End Type

'Structure of the CD-Text returned from the lead-in by ReadTOC
Public Type t_CD_TEXT
    dummy(3) As Byte                    ' header
    CDTEXT(255) As t_CD_TEXT_PACKET     ' CD-Text packets
End Type

'Structure of a standard wave header
Public Type t_WavHdr
    riff As String * 4          ' RIFF header
    len As Long                 ' length of the file
    WavFmt As String * 8        ' Wave Format
    HdrLen As Long              ' header length
    format As Integer           ' format
    NumChannels As Integer      ' number of channels
    SampleRate As Long          ' sample rate
    BytesPerSec As Long         ' bytes/second
    BlockAlign As Integer       ' block align
    BitsPerSample As Integer    ' bits per sample
    Data As String * 4          ' data chunk
    DataLen As Long             ' data length without the header length
End Type

'Structure of the data returned by IOCTL_SCSI_GET_ADDRESS by DeviceIoControl (only XP)
Public Type t_SCSI_ADDRESS
    Length As Long          ' data length
    PortNumber As Byte      ' port number (host adapter)
    PathId As Byte          ' ?
    TargetId As Byte        ' target identifier
    Lun As Byte             ' logical unit number
End Type

'Structure of the data returned by READ_DVD_STRUCTURE (format code 0)
Public Type t_DVD_Phys
    Length(1) As Byte       ' data length
    rsvd(1) As Byte         ' reserved
    '--------------
    booktype As Byte        ' Booktype (DVD-ROM, DVD-RAM, DVD-R, DVD+R, ...)
    discsize As Byte        ' Disc Size/Max. Rate
    LayerType As Byte       ' Num. Layers/Layer Type
    TrackDens As Byte       ' Track density
    zero As Byte            ' zero
    StartSector(2) As Byte  ' Physical start sector
    zero2 As Byte           ' zero
    EndSector(2) As Byte    ' Physical end sector
    zero3 As Byte           ' zero
    EndSectorLayer0 As Byte ' Physical end sector in layer 0
    bca As Byte             ' bca
    mspec(2030) As Byte     ' media specific
End Type

'Structure of the data returned by GetVersionEx
Public Type OSVERSIONINFOEX
    dwOSVersionInfoSize As Long     ' data length
    dwMajorVersion As Long          ' windows major version
    dwMinorVersion As Long          ' windows minor version
    dwBuildNumber As Long           ' windows build number
    dwPlatformId As Long            ' windows platform Identifier
    szCSDVersion As String * 128    ' windows version as a string
End Type

'Structure of the security attributes
Public Type SECURITY_ATTRIBUTES
    nLength As Long                 ' data length
    lpSecurityDescriptor As Long    ' security descriptor
    bInheritHandle As Long          ' inherit handle
End Type

''SPTI (SPTD) structure
'Public Type SCSI_PASS_THROUGH_DIRECT
'  Length As Integer                                 ' length of the request
'  ScsiStatus As Byte                                ' SCSI state
'  PathId As Byte                                    ' Host Adapter
'  TargetId As Byte                                  ' Target ID
'  Lun As Byte                                       ' Lun
'  CdbLength As Byte                                 ' CDB length
'  SenseInfoLength As Byte                           ' Sense Info length
'  DataIn As Byte                                    ' Data direction
'  DataTransferLength As Long                        ' data transfer length
'  TimeOutValue As Long                              ' Timeout interval
'  DataBuffer As Long                                ' Pointer to the databuffer
'  SenseInfoOffset As Long                           ' Startbyte of ucSenseBuf (48)
'  cdb(15) As Byte                                   ' Command Descriptor Block
'End Type
'
''SPTI+Sense Puffer structure
'Public Type SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER
'  spt As SCSI_PASS_THROUGH_DIRECT                   ' SPTI
'  Filler As Long                                    ' space filler
'  ucSenseBuf(31) As Byte                            ' Sense buffer
'  ucDataBuf(511) As Byte                            ' data buffer
'End Type

Public Type SCSI_PASS_THROUGH_DIRECT
    Length As Integer
    ScsiStatus As Byte
    PathId As Byte
    TargetId As Byte
    Lun As Byte
    CdbLength As Byte
    SenseInfoLength As Byte
    DataIn As Byte
    DataTransferLength As Long
    TimeOutValue As Long
    DataBuffer As Long
    SenseInfoOffset As Long
    Cdb(15) As Byte
    Fill(2) As Byte
End Type

Public Type SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER
    spt As SCSI_PASS_THROUGH_DIRECT
    Fill As Long
    SenseBuffer(31) As Byte
End Type

Public Type OVERLAPPED
    Internal As Long
    InternalHigh As Long
    offset As Long
    OffsetHigh As Long
    hEvent As Long
End Type
